/*
 * Copyright (C) 2006,2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: bcmst.c,v 1.30 2007/06/08 21:44:27 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <krb5/krb5.h>

#include <cert.h>
#include <certdb.h>
#include <certt.h>
#include <cryptohi.h>
#include <keyhi.h>
#include <pk11pub.h>
#include <secoid.h>

#include "aabag.h"
#include "certs.h"
#include "commont.h"
#include "bcmst.h"
#include "pkinitt.h"
#include "pkinit.h"

static void
dump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("od -t x1c", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static void
derdump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item ? item->len : 0);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n",
				item ? item->len : 0);
		}
		for (i = 0; (item != NULL) && (i < item->len); i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static unsigned char bcms_null_bytes[] = {SEC_ASN1_NULL, 0};
static SECItem bcms_null_item = {
	.data = &bcms_null_bytes[0],
	.len = 2,
};

/* ContentInfo: RFC 3852, 3, but with the constructed bit cleared on the
 * content itself.  We're using explicit tagging, so I think this is wrong. */
const SEC_ASN1Template
bcms_primitive_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

/* ContentInfo: RFC 3852, 3. */
const SEC_ASN1Template
bcms_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	/* ContentType == OBJECT IDENTIFIER */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

/* EncapsulatedContentInfo: RFC 3852, 5.2. */
const SEC_ASN1Template
bcms_encapsulated_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct content_info, content),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

/* SignerIdentifier: RFC 3852, 5.3. */
const SEC_ASN1Template
bcms_signer_identifier_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct signer_identifier, signer_type),
	.sub = NULL,
	.size = sizeof(struct signer_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_identifier,
			   signer_value.issuer_and_serial_number),
	.sub = &common_issuer_and_serial_number_template,
	.size = signer_type_issuer_and_serial_number,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = offsetof(struct signer_identifier,
			   signer_value.subject_key_identifier),
	.sub = &SEC_OctetStringTemplate,
	/* SubjectKeyIdentifier == OctetString */
	.size = signer_type_subject_key_identifier,
	},
	{ 0, 0, NULL, 0 },
};

/* Attribute: RFC 3852, 5.3. */
const SEC_ASN1Template
bcms_attribute_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct attribute),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct attribute, attr_type),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct attribute, attr_values),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem **),
	},
	{ 0, 0, NULL, 0 },
};

const SEC_ASN1Template
bcms_set_of_attributes_template[] = {
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = 0,
	.sub = bcms_attribute_template,
	.size = 0,
	},
};

/* SignedAttributes: RFC 3852, 5.3. */
const SEC_ASN1Template
bcms_signed_attributes_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0,
	.offset = 0,
	.sub = bcms_set_of_attributes_template,
	.size = 0,
	},
};

/* SignerInfo: RFC 3852, 12.1. */
const SEC_ASN1Template
bcms_signer_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct signer_info),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct signer_info, version),
	.sub = &SEC_IntegerTemplate,
	/* CMSVersion == Integer */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_info, sid),
	.sub = &bcms_signer_identifier_template,
	.size = sizeof(struct signer_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_info, digest_algorithm),
	.sub = &common_algorithm_identifier_template,
	/* DigestAlgorithmIdentifier == AlgorithmIdentifier */
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signer_info, signed_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signer_info, signature_algorithm),
	.sub = &common_algorithm_identifier_template,
	/* SignatureAlgorithmIdentifier == AlgorithmIdentifier */
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct signer_info, signature),
	.sub = &SEC_OctetStringTemplate,
	/* SignatureValue == OctetString */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signer_info, unsigned_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

/* SignedData: RFC 3852, 12.1 */
const SEC_ASN1Template
bcms_signed_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct signed_data),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct signed_data, version),
	.sub = &SEC_IntegerTemplate,
	/* CMSVersion == Integer */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct signed_data, digest_algorithms),
	.sub = common_algorithm_identifier_template,
	.size = 0,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct signed_data, encap_content_info),
	.sub = &bcms_encapsulated_content_info_template,
	.size = sizeof(struct content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signed_data, certificates),
	.sub = &SEC_SetOfAnyTemplate,
	.size = sizeof(SECItem**),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct signed_data, crls),
	.sub = &SEC_SetOfAnyTemplate,
	.size = sizeof(SECItem**),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct signed_data, signer_infos),
	.sub = bcms_signer_info_template,
	.size = 0,
	},
	{0, 0, NULL, 0},
};

/* ExternalPrincipalIdentifier: RFC 4556, Appendix A. */
const SEC_ASN1Template
bcms_external_principal_identifier_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct external_principal_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   der_subject_name),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   issuer_and_serial_number),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	/* This is supposed to be implicit, but we use an explicit tag here to
	 * avoid having to wrap and unwrap the raw data which NSS gives us. */
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2 | SEC_ASN1_EXPLICIT |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct external_principal_identifier,
			   subject_key_identifier),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
bcms_sequence_of_external_principal_identifier_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = bcms_external_principal_identifier_template,
	.size = sizeof(struct external_principal_identifier **),
	},
};

/* RecipientIdentifier: RFC 3852, 12.1. */
const SEC_ASN1Template
bcms_recipient_identifier_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_type),
	.sub = NULL,
	.size = sizeof(struct recipient_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_value.issuer_and_serial_number),
	.sub = &common_issuer_and_serial_number_template,
	.size = recipient_identifier_type_issuer_and_serial_number,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = offsetof(struct recipient_identifier,
			   recipient_identifier_value.subject_key_identifier),
	.sub = &SEC_OctetStringTemplate,
	.size = recipient_identifier_type_subject_key_identifier,
	},
	{ 0, 0, NULL, 0 },
};

/* EncryptedContentInfo: RFC 3852, 12.1. */
const SEC_ASN1Template
bcms_encrypted_content_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct encrypted_content_info),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct encrypted_content_info, content_type),
	.sub = &SEC_ObjectIDTemplate,
	/* ContentType == Object Identifier */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct encrypted_content_info,
			   content_encryption_algorithm),
	.sub = &common_algorithm_identifier_template,
	/* contentEncryptionAlgorithmIdentifier == AlgorithmIdentifier */
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct encrypted_content_info, encrypted_content),
	/* EncryptedContent == Octet String */
	.sub = SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0 },
};

/* KeyTransRecipientInfo: RFC 3852: 6.2.1. */
const SEC_ASN1Template
bcms_key_trans_recipient_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct key_trans_recipient_info),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct key_trans_recipient_info, version),
	.sub = &SEC_IntegerTemplate,
	/* CMSVersion == Integer */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct key_trans_recipient_info, rid),
	.sub = &bcms_recipient_identifier_template,
	.size = sizeof(struct recipient_identifier),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct key_trans_recipient_info,
			   key_encryption_algorithm),
	.sub = &common_algorithm_identifier_template,
	/* KeyEncryptionAlgorithmIdentifier == AlgorithmIdentifier */
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct key_trans_recipient_info, encrypted_key),
	.sub = &SEC_OctetStringTemplate,
	/* EncryptedKey == Octet String */
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* RecipientInfo: RFC 3852, 6.2. */
const SEC_ASN1Template
bcms_recipient_info_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct recipient_info, recipient_info_type),
	.sub = NULL,
	.size = sizeof(struct recipient_info),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct recipient_info, recipient_info_value.ktri),
	.sub = &bcms_key_trans_recipient_info_template,
	.size = recipient_info_type_key_trans,
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct recipient_info,
			   recipient_info_value.der_kari),
	.sub = &SEC_AnyTemplate,
	.size = recipient_info_type_key_agree,
	},
	{0, 0, NULL, 0},
};

/* EnvelopedData: RFC 3852, 12.1. */
const SEC_ASN1Template
bcms_enveloped_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct enveloped_data),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct enveloped_data, version),
	.sub = &SEC_IntegerTemplate,
	/* CMSVersion == Integer */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct enveloped_data, der_originator_info),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_SET_OF,
	.offset = offsetof(struct enveloped_data, recipient_infos),
	.sub = &bcms_recipient_info_template,
	.size = 0,
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct enveloped_data, encrypted_content_info),
	.sub = &bcms_encrypted_content_info_template,
	.size = sizeof(struct encrypted_content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct enveloped_data, der_unsigned_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* EncryptedData: RFC 3852, 8. */
const SEC_ASN1Template
bcms_encrypted_data_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct encrypted_data),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct encrypted_data, version),
	.sub = &SEC_IntegerTemplate,
	/* CMSVersion == Integer */
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct encrypted_data, encrypted_content_info),
	.sub = &bcms_encrypted_content_info_template,
	.size = sizeof(struct encrypted_content_info),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct encrypted_data, der_unsigned_attrs),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

struct attribute **
bcms_decode_set_of_attributes(struct module_context *mcontext,
			      PLArenaPool *pool, SECItem *item)
{
	struct attribute **attrs;

	attrs = NULL;
	if (SEC_ASN1DecodeItem(pool, &attrs, bcms_set_of_attributes_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding set of attributes.\n");
		return NULL;
	}
	return attrs;
}

struct attribute **
bcms_decode_signed_attributes(struct module_context *mcontext,
			      PLArenaPool *pool, SECItem *item)
{
	struct attribute **attrs;

	attrs = NULL;
	if (SEC_ASN1DecodeItem(pool, &attrs, bcms_signed_attributes_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding signed attributes.\n");
		return NULL;
	}
	return attrs;
}

struct external_principal_identifier **
bcms_decode_sequence_of_external_principal_identifier(struct module_context *mcontext,
						      PLArenaPool *pool,
						      SECItem *item)
{
	struct external_principal_identifier **epi;

	epi = NULL;
	if (SEC_ASN1DecodeItem(pool, &epi, bcms_sequence_of_external_principal_identifier_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding sequence of external principal "
			     "identifiers.\n");
		return NULL;
	}
	return epi;
}

struct content_info *
bcms_decode_content_info(struct module_context *mcontext,
			 PLArenaPool *pool, SECItem *item)
{
	struct content_info *ci;

	ci = PORT_ArenaZAlloc(pool, sizeof(*ci));
	if (ci == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ci, bcms_content_info_template,
			       item) != SECSuccess) {
		if (SEC_ASN1DecodeItem(pool, ci,
				       bcms_primitive_content_info_template,
				       item) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "Error decoding content-info.\n");
			return NULL;
		}
	}
	return ci;
}

SECItem *
bcms_encode_set_of_attributes(struct module_context *mcontext,
			      PLArenaPool *pool, struct attribute **attr)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, &attr,
			       bcms_set_of_attributes_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding set of attributes.\n");
	return NULL;
}

SECItem *
bcms_encode_signed_attributes(struct module_context *mcontext,
			      PLArenaPool *pool, struct attribute **attrs)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, &attrs,
			       bcms_signed_attributes_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding signed attributes.\n");
	return NULL;
}

SECItem *
bcms_encode_content_info(struct module_context *mcontext,
			 PLArenaPool *pool, struct content_info *ci)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ci,
			       bcms_content_info_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding content-info.\n");
	return NULL;
}

SECItem *
bcms_encode_primitive_content_info(struct module_context *mcontext,
				   PLArenaPool *pool, struct content_info *ci)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ci,
			       bcms_primitive_content_info_template) ==
	    &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1,
		     "Error encoding (constructed) content-info.\n");
	return NULL;
}

struct signed_data *
bcms_decode_signed_data(struct module_context *mcontext,
			PLArenaPool *pool, SECItem *item)
{
	struct signed_data *sd;

	sd = PORT_ArenaZAlloc(pool, sizeof(*sd));
	if (sd == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, sd, bcms_signed_data_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding signed-data.\n");
		return NULL;
	}
	return sd;
}

SECItem *
bcms_encode_signed_data(struct module_context *mcontext,
			PLArenaPool *pool, struct signed_data *sd)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, sd,
			       bcms_signed_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding signed-data.\n");
	return NULL;
}

struct enveloped_data *
bcms_decode_enveloped_data(struct module_context *mcontext,
			   PLArenaPool *pool, SECItem *item)
{
	struct enveloped_data *ed;

	ed = PORT_ArenaZAlloc(pool, sizeof(*ed));
	if (ed == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ed, bcms_enveloped_data_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding enveloped-data.\n");
		return NULL;
	}
	return ed;
}

SECItem *
bcms_encode_enveloped_data(struct module_context *mcontext,
			   PLArenaPool *pool, struct enveloped_data *ed)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ed,
			       bcms_enveloped_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding enveloped-data.\n");
	return NULL;
}

struct encrypted_data *
bcms_decode_encrypted_data(struct module_context *mcontext,
			   PLArenaPool *pool, SECItem *item)
{
	struct encrypted_data *ed;

	ed = PORT_ArenaZAlloc(pool, sizeof(*ed));
	if (ed == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ed, bcms_encrypted_data_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding encrypted-data.\n");
		return NULL;
	}
	return ed;
}

SECItem *
bcms_encode_encrypted_data(struct module_context *mcontext,
			   PLArenaPool *pool, struct encrypted_data *ed)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, ed,
			       bcms_encrypted_data_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 1, "Error encoding encrypted-data.\n");
	return NULL;
}

struct content_info *
bcms_build_content_info(PLArenaPool *pool, SECOidTag content_type_tag,
			SECItem *content_data)
{
	SECOidData *oid;
	oid = SECOID_FindOIDByTag(content_type_tag);
	if (oid != NULL) {
		return bcms_build_content_info_oid(pool, &oid->oid,
						   content_data);
	}
	return NULL;
}

struct content_info *
bcms_build_content_info_oid(PLArenaPool *pool, SECItem *content_type_oid,
			    SECItem *content_data)
{
	struct content_info *ret;
	SECOidData *oid;
	if (content_type_oid == NULL) {
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return NULL;
		}
		content_type_oid = &oid->oid;
	}
	ret = PORT_ArenaZAlloc(pool, sizeof(*ret));
	if (ret != NULL) {
		ret->content_type = *content_type_oid;
		ret->content = *content_data;
	}
	return NULL;
}

/* Turn a digest or signature tag into the corresponding digest tag. */
static SECOidTag
sanify_digtag(SECOidTag digtag)
{
	switch (digtag) {
	case SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION:
	case SEC_OID_MD5:
		digtag = SEC_OID_MD5;
		break;
	case SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION:
	case SEC_OID_ANSIX9_DSA_SIGNATURE_WITH_SHA1_DIGEST:
	case SEC_OID_SHA1:
		digtag = SEC_OID_SHA1;
		break;
	case SEC_OID_PKCS1_SHA256_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA256:
		digtag = SEC_OID_SHA256;
		break;
	case SEC_OID_PKCS1_SHA384_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA384:
		digtag = SEC_OID_SHA384;
		break;
	case SEC_OID_PKCS1_SHA512_WITH_RSA_ENCRYPTION:
	case SEC_OID_SHA512:
		digtag = SEC_OID_SHA512;
		break;
	default:
		break;
	}
	return digtag;
}

/* Map a digest/encrypt tag pair into a signature tag. */
static SECOidTag
find_sigoid(SECOidTag digtag, SECOidTag sigenctag)
{
	unsigned int i;
	const struct {
		SECOidTag digtag, enctag, sigtag;
	} tagmap[] = {
		{SEC_OID_MD5, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA1, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA256, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA256_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA384, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA384_WITH_RSA_ENCRYPTION},
		{SEC_OID_SHA512, SEC_OID_PKCS1_RSA_ENCRYPTION,
		 SEC_OID_PKCS1_SHA512_WITH_RSA_ENCRYPTION},
	};
	for (i = 0; i < sizeof(tagmap) / sizeof(tagmap[0]); i++) {
		if ((tagmap[i].digtag == digtag) &&
		    (tagmap[i].enctag == sigenctag)) {
			return tagmap[i].sigtag;
		}
	}
	return sigenctag;
}

/* Find the certificate in the list which matches the signer_identifier.  The
 * caller must free whatever certificate we return. */
static CERTCertificate *
find_cert_from_sid(struct module_context *mcontext,
		   CERTCertDBHandle *certdb, SECItem **certificates,
		   struct signer_identifier *sid)
{
	CERTCertificate *cert, **certs;
	struct issuer_and_serial_number *is;
	SECItem *spki;
	int i, n_certificates;

	/* First, check that we actually have a certificate list. */
	for (i = 0; (certificates != NULL) && (certificates[i] != NULL); i++) {
		continue;
	}
	if (i == 0) {
		pkinit_debug(mcontext, 1,
			     "Could not find certificate corresponding to "
			     "signer info in an empty certificate list.\n");
		return NULL;
	}
	n_certificates = i;
	certs = NULL;

	/* Parse the list of certificates so that we can examine them. */
	if (CERT_ImportCerts(certdb, 0,
			     n_certificates, certificates,
			     &certs,
			     PR_FALSE, PR_FALSE,
			     NULL) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error importing certificates.\n");
		return NULL;
	}

	/* Walk the list of certificates. */
	for (i = 0; i < n_certificates; i++) {
		cert = certs[i];
		switch (sid->signer_type) {
		case signer_type_issuer_and_serial_number:
			/* The signer is identified by issuer/serial. */
			is = &sid->signer_value.issuer_and_serial_number;
			if (SECITEM_ItemsAreEqual(&is->issuer,
						  &cert->derIssuer) &&
			    SECITEM_ItemsAreEqual(&is->serial,
						  &cert->serialNumber)) {
				cert = CERT_DupCertificate(cert);
				CERT_DestroyCertArray(certs, n_certificates);
				return cert;
			}
			break;
		case signer_type_subject_key_identifier:
			/* The signer is identified by the key identifier. */
			spki = &sid->signer_value.subject_key_identifier;
			if (SECITEM_ItemsAreEqual(spki, &cert->subjectKeyID)) {
				cert = CERT_DupCertificate(cert);
				CERT_DestroyCertArray(certs, n_certificates);
				return cert;
			}
			break;
		default:
			break;
		}
	}
	CERT_DestroyCertArray(certs, n_certificates);
	pkinit_debug(mcontext, 1,
		     "Could not find certificate in list which corresponded "
		     "to signer_info.\n");
	return NULL;
}

/* Verify ALL of the signatures on a signed_data, and if it has exactly one
 * signer, return the certificate of that signer. */
SECStatus
bcms_verify_signed_data(struct module_context *mcontext,
			PLArenaPool *pool,
			CERTCertDBHandle *certdb,
			struct signed_data *signed_data,
			CERTCertificate **lone_signer,
			time_t *signing_time)
{
	SECOidData *digoid, *sigoid;
	SECOidTag digtag, sigtag, attrtag;
	CERTCertificate *cert;
	SECKEYPublicKey *public;
	PK11Context *ctx;
	struct signer_info **signer_infos;
	struct algorithm_identifier *algorithm;
	struct attribute **attributes;
	SECItem *econtent, *econtent_type, *signed_content;
	SECItem digest, *signed_digest, *content_type;
	SECStatus status;
	int i, j;
	unsigned int digest_length;
	unsigned char digest_buf[64];
	PRBool saw_digest_attr, saw_content_type_attr, saw_signing_time_attr;
	PRBool matched_digest_attr, matched_content_type_attr;

	signer_infos = signed_data->signer_infos;
	status = SECSuccess;
	*lone_signer = NULL;
	cert = NULL;
	for (i = 0; (signer_infos != NULL) && (signer_infos[i] != NULL); i++) {
		/* Pick out the digest/encryption/signature types. */
		pkinit_debug(mcontext, 2, "Signature %d:\n", i);
		algorithm = &signer_infos[i]->signature_algorithm;
		sigoid = SECOID_FindOID(&algorithm->algorithm);
		algorithm = &signer_infos[i]->digest_algorithm;
		digoid = SECOID_FindOID(&algorithm->algorithm);
		if (digoid != NULL) {
			digtag = sanify_digtag(digoid->offset);
			if ((digtag > 0) && (digtag != digoid->offset)) {
				pkinit_debug(mcontext, 2,
					     "\tdigest (as provided): %s\n",
					     digoid->desc);
			}
			if (SECOID_FindOIDByTag(digtag)) {
				digoid = SECOID_FindOIDByTag(digtag);
			}
		}
		if (sigoid != NULL) {
			sigtag = find_sigoid(digoid->offset, sigoid->offset);
			if ((sigtag > 0) && (sigtag != sigoid->offset)) {
				pkinit_debug(mcontext, 2,
					     "\tsignature (as provided): %s\n",
					     sigoid->desc);
			}
			if (SECOID_FindOIDByTag(sigtag)) {
				sigoid = SECOID_FindOIDByTag(sigtag);
			}
		}
		if (digoid == NULL) {
			pkinit_debug(mcontext, 2, "\tunknown digest type.\n");
			break;
		}
		if (sigoid == NULL) {
			pkinit_debug(mcontext, 2, "\tunknown signature type.\n");
			break;
		}
		/* We won't use a cert left over from a previous iteration. */
		if (cert != NULL) {
			CERT_DestroyCertificate(cert);
		}
		/* Find the signing certificate in the list. */
		pkinit_debug(mcontext, 2, "\tdigest: %s\n", digoid->desc);
		pkinit_debug(mcontext, 2, "\tsignature: %s\n", sigoid->desc);
		cert = find_cert_from_sid(mcontext,
					  certdb, signed_data->certificates,
					  &signer_infos[i]->sid);
		digtag = digoid->offset;
		sigtag = sigoid->offset;

		/* Extract the verification key from the certificate. */
		public = NULL;
		if (cert != NULL) {
			pkinit_debug(mcontext, 2,
				     "\tFound signing certificate.\n");
			public = CERT_ExtractPublicKey(cert);
		} else {
			pkinit_debug(mcontext, 1,
				     "\tSigning certificate not included in "
				     "signed_data message.\n");
			/* No/invalid cert? Bail. */
			break;
		}
		econtent = NULL;
		econtent_type = NULL;
		if (public != NULL) {
			pkinit_debug(mcontext, 2, "\tPulled out public key.\n");
			econtent = &signed_data->encap_content_info.content;
			econtent_type = &signed_data->encap_content_info.content_type;
		} else {
			/* No/invalid public key? Bail. */
			pkinit_debug(mcontext, 1,
				     "\tUnable to extract public key from "
				     "signer certificate.\n");
			break;
		}
		if (econtent == NULL) {
			pkinit_debug(mcontext, 1,
				     "\tUnable to verify detached data.\n");
			SECKEY_DestroyPublicKey(public);
			public = NULL;
			continue;
		}
		/* Determine what was signed. */
		attributes = NULL;
		if (signer_infos[i]->signed_attrs.len > 0) {
			signed_content = &signer_infos[i]->signed_attrs;
			attributes = bcms_decode_signed_attributes(mcontext,
								   pool,
								   signed_content);
			if (attributes == NULL) {
				pkinit_debug(mcontext, 1,
					     "\tUnable to parse signed "
					     "attributes.\n");
				break;
			}
			/* Verify at least the content type and the digest.
			 * Record the signing time, if present. */
			saw_content_type_attr = PR_FALSE;
			saw_digest_attr = PR_FALSE;
			saw_signing_time_attr = PR_FALSE;
			matched_content_type_attr = PR_FALSE;
			matched_digest_attr = PR_FALSE;
			for (j = 0; attributes[j] != NULL; j++) {
				attrtag = SECOID_FindOIDTag(&attributes[j]->attr_type);
				if (attrtag <= 0) {
					continue;
				}
				switch (attrtag) {
				case SEC_OID_PKCS9_CONTENT_TYPE:
					saw_content_type_attr = PR_TRUE;
					matched_content_type_attr = PR_TRUE;
					/* Check for a match with the
					 * encapsulated content info. */
					if ((attributes[j]->attr_values[0] != NULL) &&
					    (attributes[j]->attr_values[1] == NULL)) {
						content_type = common_decode_objectid(mcontext, pool, attributes[j]->attr_values[0]);
						if (SECITEM_ItemsAreEqual(econtent_type,
									  content_type)) {
							matched_content_type_attr = PR_TRUE;
						}
					}
					break;
				case SEC_OID_PKCS9_MESSAGE_DIGEST:
					saw_digest_attr = PR_TRUE;
					matched_digest_attr = PR_FALSE;

					/* Recompute the digest. */
					ctx = PK11_CreateDigestContext(digtag);
					if (ctx == NULL) {
						pkinit_debug(mcontext, 0,
							     "\tError setting up hash.\n");
						break;
					}
					if (PK11_DigestBegin(ctx) != SECSuccess) {
						pkinit_debug(mcontext, 0,
							     "Invalid digest type.\n");
						PK11_DestroyContext(ctx, PR_TRUE);
						return SECFailure;
					}
					if (PK11_DigestOp(ctx, econtent->data, econtent->len) != SECSuccess) {
						pkinit_debug(mcontext, 0,
							     "\tError digesting signed data.\n");
						PK11_DestroyContext(ctx, PR_TRUE);
						break;
					}
					if (PK11_DigestFinal(ctx, digest_buf,
							     &digest_length,
							     sizeof(digest_buf)) != SECSuccess) {
						pkinit_debug(mcontext, 0,
							     "\tError digesting signed data.\n");
						PK11_DestroyContext(ctx, PR_TRUE);
						break;
					}
					if (digest_length > sizeof(digest_buf)) {
						pkinit_debug(mcontext, 0,
							     "\tError digesting signed data: digest too large.\n");
						PK11_DestroyContext(ctx, PR_TRUE);
						break;
					}
					digest.data = digest_buf;
					digest.len = digest_length;
					PK11_DestroyContext(ctx, PR_TRUE);

					if ((attributes[j]->attr_values[0] != NULL) &&
					    (attributes[j]->attr_values[1] == NULL)) {
						signed_digest = common_decode_octetstring(mcontext, pool, attributes[j]->attr_values[0]);
						if ((signed_digest != NULL) &&
						    (SECITEM_ItemsAreEqual(&digest,
									   signed_digest))) {
							matched_digest_attr = PR_TRUE;
						}
					}
					break;
				case SEC_OID_PKCS9_SIGNING_TIME:
					saw_signing_time_attr = PR_TRUE;
					if ((signing_time != NULL) &&
					    (attributes[j]->attr_values[0] != NULL) &&
					    (attributes[j]->attr_values[1] == NULL)) {
						*signing_time = pkinit_time_from_signing_time(mcontext, pool, attributes[j]->attr_values[0]);
					}
					break;
				default:
					break;
				}
			}
			if (!saw_content_type_attr) {
				pkinit_debug(mcontext, 1,
					     "\tSigned attributes list did not "
					     "include a required content "
					     "type attribute.\n");
				break;
			}
			if (!matched_content_type_attr) {
				pkinit_debug(mcontext, 1,
					     "\tSigned content type doesn't "
					     "match the encapsulated content "
					     "type.\n");
				break;
			}
			if (!saw_digest_attr) {
				pkinit_debug(mcontext, 1,
					     "\tSigned attributes list did not "
					     "include a required digest "
					     "attribute.\n");
				break;
			}
			if (!matched_digest_attr) {
				pkinit_debug(mcontext, 1,
					     "\tSigned digest doesn't "
					     "match the data.\n");
				break;
			}
			/* For the signature verification, we'll need to
			 * replace the implicit context-specific tag with the
			 * constructed SET tag. */
			signed_content = SECITEM_ArenaDupItem(pool,
							      signed_content);
			signed_content->data[0] = SEC_ASN1_SET |
						  SEC_ASN1_CONSTRUCTED;
		} else {
			signed_content = econtent;
		}
		/* Verify the signature over whatever was signed. */
		if (signed_content != NULL) {
			status = VFY_VerifyData(signed_content->data,
						signed_content->len,
						public,
						&signer_infos[i]->signature,
						sigoid->offset,
						NULL);
			if (status == SECSuccess) {
				pkinit_debug(mcontext, 2,
					     "\tSignature successfully "
					     "verified.\n");
			} else {
				pkinit_debug(mcontext, 1,
					     "\tSignature check failed over "
					     "%d bytes of data.\n",
					     signed_content->len);
				break;
			}
		} else {
			/* No data to verify the signature? Bail. */
			pkinit_debug(mcontext, 1, "\tNo signed data.\n");
			break;
		}
		SECKEY_DestroyPublicKey(public);
		public = NULL;
	}
	/* Make sure we freed the public key, in case we broke out of the loop
	 * early. */
	if (public != NULL) {
		SECKEY_DestroyPublicKey(public);
		public = NULL;
	}
	/* If we had no signatures, we've failed. */
	if (i == 0) {
		pkinit_debug(mcontext, 1,
			     "No signatures successfully verified.\n");
		return SECFailure;
	}
	/* If we had only one signer, return a copy of the signer's
	 * certificate. */
	if ((i == 1) && (lone_signer != NULL)) {
		*lone_signer = cert;
	} else {
		if (cert != NULL) {
			CERT_DestroyCertificate(cert);
		}
	}
	/* If we had failed any check (meaning we didn't run out of signatures
	 * to verify), we've failed. */
	if (signer_infos[i] == NULL) {
		return SECSuccess;
	}
	/* Return the status of the last-attempted signature verification. */
	return status;
}

/* Make a copy of the list of certificates, ensuring that the passed-in value
 * is in there somewhere. */
static SECItem **
bcms_make_certificate_list(struct module_context *mcontext,
			   PLArenaPool *pool, SECItem **list,
			   CERTCertificate *cert)
{
	PRBool found;
	SECItem **nlist;
	int i;

	found = PR_FALSE;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		if (SECITEM_ItemsAreEqual(list[i], &cert->derCert)) {
			found = PR_TRUE;
			break;
		}
	}

	if (found) {
		nlist = list;
	} else {
		/* i contains the length of the current list */
		nlist = PORT_ArenaZAlloc(pool, sizeof(SECItem*) * (i + 2));
		if (nlist == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		/* add our entry to the end of the list */
		nlist[i] = SECITEM_ArenaDupItem(pool, &cert->derCert);
		/* copy the rest of the list */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	}

	return nlist;
}

/* Add a signature to the signed_data structure. */
static SECStatus
bcms_add_signer_to_signed_data(struct module_context *mcontext,
			       PLArenaPool *pool,
			       struct signed_data *sig,
			       SECOidTag digtag,
			       SECOidTag sigtag,
			       PK11SlotInfo *preferred_slot,
			       CERTCertDBHandle *certdb,
			       CERTCertificate *cert,
			       struct aa_bag *bag,
			       PK11PasswordFunc pwcb,
			       struct pkinit_pwcb_args *pwcb_args)
{
	SECOidData *digoid, *sigoid, *oid;
	SECKEYPrivateKey *privkey;
	struct signer_info **signer_infos, *signer_info;
	SECItem **certificates;
	struct algorithm_identifier **digest_algorithms;
	struct attribute content_type_attr, digest_attr, signing_time_attr;
	struct attribute *attributes[4];
	SECItem *content_types[2], *digest_values[2], *signing_times[2];
	SECStatus status;
	SECItem *sign_content, *econtent;
	SECItem *null_item, digest_item, signing_time_item;
	PK11Context *ctx;
	int signer_count, i;
	unsigned char digest_buf[64];
	unsigned int digest_length, attribute;

	/* Double-check that we actually have something to sign. */
	if (sig->encap_content_info.content.len == 0) {
		pkinit_debug(mcontext, 0, "No content to sign.\n");
		return SECFailure;
	}
	econtent = &sig->encap_content_info.content;

	/* Find the private key which corresponds to the certificate. */
	privkey = cert_find_key_for_cert(mcontext,
					 preferred_slot, cert, bag,
					 pwcb_args);
	if (privkey == NULL) {
		pkinit_debug(mcontext, 0,
			     "Unable to locate private key "
			     "corresponding to signer "
			     "certificate \"%s\".\n",
			     cert->subjectName);
		return SECFailure;
	}

	/* Create the new signer_info structure and set its version. */
	signer_info = PORT_ArenaZAlloc(pool, sizeof(*signer_info));
	if (SEC_ASN1EncodeInteger(pool, &signer_info->version, 1) !=
	    &signer_info->version) {
		pkinit_debug(mcontext, 0, "Error encoding version.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}

	/* Set up the signer identifier. */
	signer_info->sid.signer_type =
		signer_type_issuer_and_serial_number;
	signer_info->sid.signer_value.issuer_and_serial_number.issuer =
		cert->derIssuer;
	signer_info->sid.signer_value.issuer_and_serial_number.serial =
		cert->serialNumber;

	/* Set up the digest algorithm. */
	digoid = SECOID_FindOIDByTag(digtag);
	if (digoid == NULL) {
		pkinit_debug(mcontext, 0, "Invalid digest type.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	signer_info->digest_algorithm.algorithm = digoid->oid;

	/* An MD5 digest requires NULL parameters.  For everyting else, while
	 * NULL must be accepted, parameters are better omitted. */
	null_item = SECITEM_ArenaDupItem(pool, &bcms_null_item);
	switch (digoid->offset) {
	case SEC_OID_MD5:
		signer_info->digest_algorithm.parameters = null_item;
		break;
	default:
		signer_info->digest_algorithm.parameters = NULL;
		break;
	}

	/* Digest the data. */
	ctx = PK11_CreateDigestContext(digoid->offset);
	if (ctx == NULL) {
		pkinit_debug(mcontext, 0, "Invalid digest type.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	if (PK11_DigestBegin(ctx) != SECSuccess) {
		pkinit_debug(mcontext, 0, "Invalid digest type.\n");
		PK11_DestroyContext(ctx, PR_TRUE);
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	if (PK11_DigestOp(ctx, econtent->data, econtent->len) != SECSuccess) {
		pkinit_debug(mcontext, 0, "Error digesting data for %s.\n",
			     digoid->desc);
		PK11_DestroyContext(ctx, PR_TRUE);
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	if (PK11_DigestFinal(ctx, digest_buf, &digest_length,
			     sizeof(digest_buf)) != SECSuccess) {
		pkinit_debug(mcontext, 0, "Error digesting data for %s.\n",
			     digoid->desc);
		PK11_DestroyContext(ctx, PR_TRUE);
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	if (digest_length > sizeof(digest_buf)) {
		pkinit_debug(mcontext, 0, "Error digesting data for %s.\n",
			     digoid->desc);
		PK11_DestroyContext(ctx, PR_TRUE);
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	PK11_DestroyContext(ctx, PR_TRUE);
	memset(&digest_item, 0, sizeof(digest_item));
	digest_item.data = digest_buf;
	digest_item.len = digest_length;

	/* Set up the authenticated attributes. */
	content_types[0] = common_encode_objectid(mcontext, pool,
						  &sig->encap_content_info.content_type);
	content_types[1] = NULL;
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS9_CONTENT_TYPE);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error in NSS: no knowledge "
			     "of PKCS#9 Content Type.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	content_type_attr.attr_type = oid->oid;
	content_type_attr.attr_values = content_types;

	digest_values[0] = common_encode_octetstring(mcontext, pool,
						     &digest_item);
	digest_values[1] = NULL;
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS9_MESSAGE_DIGEST);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error in NSS: no knowledge "
			     "of PKCS#9 Message Digest.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	digest_attr.attr_type = oid->oid;
	digest_attr.attr_values = digest_values;

	if (pkinit_signing_time_from_time(mcontext, pool, NULL,
					  &signing_time_item) == SECSuccess) {
		signing_times[0] = &signing_time_item;
		signing_times[1] = NULL;
		pkinit_debug(mcontext, 2, "Including signing time.\n");
	} else {
		signing_times[0] = NULL;
		pkinit_debug(mcontext, 2, "Unable to including signing "
			     "time.\n");
	}
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS9_SIGNING_TIME);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error in NSS.: no knowledge "
			     "of PKCS#9 Signing Time\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	}
	signing_time_attr.attr_type = oid->oid;
	signing_time_attr.attr_values = signing_times;

	attribute = 0;
	attributes[attribute++] = &content_type_attr;
	if (signing_times[0] != NULL) {
		attributes[attribute++] = &signing_time_attr;
	}
	attributes[attribute++] = &digest_attr;
	attributes[attribute++] = NULL;
	sign_content = bcms_encode_set_of_attributes(mcontext, pool,
						     attributes);

	/* Assign the value, then do the tag switcheroo so that the encoding
	 * comes out right. */
	if (sign_content != NULL) {
		signer_info->signed_attrs = *sign_content;
		sign_content = SECITEM_ArenaDupItem(pool, sign_content);
		signer_info->signed_attrs.data[0] = SEC_ASN1_CONTEXT_SPECIFIC |
						    0 |
						    SEC_ASN1_CONSTRUCTED;
	} else {
		memset(&signer_info->signed_attrs, 0,
		       sizeof(signer_info->signed_attrs));
		sign_content = econtent;
	}

	/* Set up the signing algorithm. */
	sigtag = find_sigoid(sanify_digtag(digtag), sigtag);
	sigoid = SECOID_FindOIDByTag(sigtag);
	if (sigoid == NULL) {
		pkinit_debug(mcontext, 0,
			     "Internal error: attempted to sign with unknown "
			     "signature type.\n");
		SECKEY_DestroyPrivateKey(privkey);
		return SECFailure;
	} else {
		pkinit_debug(mcontext, 2,
			     "Generating %s signature over %d bytes.\n",
			     sigoid->desc, sign_content->len);
	}
	signer_info->signature_algorithm.algorithm = sigoid->oid;
	signer_info->signature_algorithm.parameters = null_item;

	/* Create a PKCS#1 signature using the selected digest. */
	status = SEC_SignData(&signer_info->signature,
			      sign_content->data,
			      sign_content->len,
			      privkey,
			      sigoid->offset);
	SECKEY_DestroyPrivateKey(privkey);
	privkey = NULL;

	if (status != SECSuccess) {
		pkinit_debug(mcontext, 0, "Error signing data.\n");
		return SECFailure;
	}

	/* No unauthenticated attributes. */
	memset(&signer_info->unsigned_attrs, 0,
	       sizeof(signer_info->unsigned_attrs));

	/* Append our signer_info to the existing list.  Start by counting the
	 * existing list's size. */
	signer_infos = sig->signer_infos;
	for (signer_count = 0;
	     (signer_infos != NULL) && (signer_infos[signer_count] != NULL);
	     signer_count++) {
		continue;
	}
	signer_infos = PORT_ArenaZAlloc(pool,
					sizeof(struct signer_info*) *
					(signer_count + 2));
	if (signer_infos == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory\n");
		return SECFailure;
	}

	/* Save the old elements of the list, and then append the new one. */
	for (i = 0; i < signer_count; i++) {
		signer_infos[i] = sig->signer_infos[i];
	}
	signer_infos[i] = signer_info;

	/* Save the digest algorithm in the signed_data's list. */
	digest_algorithms = common_make_algorithm_identifier_list(mcontext,
								  pool,
								  sig->digest_algorithms,
								  &signer_info->digest_algorithm);
	if (digest_algorithms == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory\n");
		return SECFailure;
	}

	/* Save the certificate to the signed_data's list. */
	certificates = bcms_make_certificate_list(mcontext, pool,
						  sig->certificates, cert);
	if (certificates == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory\n");
		return SECFailure;
	}

	/* Save the new lists. */
	sig->digest_algorithms = digest_algorithms;
	sig->signer_infos = signer_infos;
	sig->certificates = certificates;

	return status;
}

/* Add a certificiate to the signed_data structure. */
static SECStatus
bcms_add_cert_to_signed_data(struct module_context *mcontext,
			     PLArenaPool *pool,
			     struct signed_data *sig,
			     CERTCertificate *cert)
{
	SECItem **certificates;
	if (cert != NULL) {
		certificates = bcms_make_certificate_list(mcontext, pool,
							  sig->certificates,
							  cert);
		if (certificates == NULL) {
			return SECFailure;
		}
		sig->certificates = certificates;
	}
	return SECSuccess;
}

/* Add a certificiate's chain (minus itself and its root) to the signed_data
 * structure. */
static SECStatus
bcms_add_cert_chain_to_signed_data(struct module_context *mcontext,
				   PLArenaPool *pool,
				   struct signed_data *sig,
				   CERTCertificate *cert,
				   struct aa_bag *bag)
{
	CERTCertificate *issuer, *next_link;

	issuer = NULL;
	pkinit_debug(mcontext, 2, "Walking certifying chain for \"%s\".\n",
		     cert->subjectName);
	issuer = cert_find_cert_issuer(cert, 0, 0, bag);
	while ((issuer != NULL) && (!issuer->isRoot)) {
		pkinit_debug(mcontext, 2, "Adding chain link \"%s\" "
			     "(trust is %02x,%02x,%02x).\n",
			     issuer->subjectName,
			     issuer->subjectName,
			     issuer->trust ?
			     issuer->trust->sslFlags : 0,
			     issuer->trust ?
			     issuer->trust->emailFlags : 0,
			     issuer->trust ?
			     issuer->trust->objectSigningFlags : 0);
		if (bcms_add_cert_to_signed_data(mcontext, pool, sig,
						 issuer) != SECSuccess) {
			CERT_DestroyCertificate(issuer);
			return SECFailure;
		}
		next_link = cert_find_cert_issuer(issuer, 0, 0, bag);
		CERT_DestroyCertificate(issuer);
		issuer = next_link;
	}
	if (issuer == NULL) {
		pkinit_debug(mcontext, 2,
			     "No more links in \"%s\"'s cert chain found.\n",
			     cert->subjectName);
		return SECFailure;
	} else {
		if (issuer->isRoot) {
			pkinit_debug(mcontext, 2, "Stopping at root \"%s\" "
				     "(trust is %03x).\n",
				     issuer->subjectName,
				     issuer->trust ?
				     issuer->trust->sslFlags : 0);
		}
		CERT_DestroyCertificate(issuer);
	}
	return SECSuccess;
}

/* Populate a new signed_data structure with the given encapsulated content
 * type and data. */
static SECStatus
bcms_populate_signed_data(struct module_context *mcontext,
			  PLArenaPool *pool,
			  struct signed_data *signed_data,
			  const SECItem *encapsulated_oid,
			  const SECItem *encapsulated_data)
{
	/* Build the signed-data payload, version first. */
	if (SEC_ASN1EncodeInteger(pool, &signed_data->version, 1) !=
	    &signed_data->version) {
		pkinit_debug(mcontext, 0,
			     "Error encoding signed-data version.\n");
		return SECFailure;
	}

	/* Initialize as empty the list of digest algorithms. */
	signed_data->digest_algorithms = NULL;

	/* Build the encapsulated content-info. */
	signed_data->encap_content_info.content_type = *encapsulated_oid;
	signed_data->encap_content_info.content = *encapsulated_data;

	/* Initially, we have no signer-infos, so no certificates and stuff. */
	signed_data->certificates = NULL;
	signed_data->crls = NULL;
	signed_data->signer_infos = NULL;

	return SECSuccess;
}

/* Build a new encoded signed_data. */
SECItem *
bcms_create_signed_data(struct module_context *mcontext,
			PLArenaPool *pool,
			const SECItem *econtent_type,
			SECItem *data,
			CERTCertDBHandle *certdb,
			struct aa_bag *bag,
			CERTCertificate *signer,
			SECOidTag digest_algorithm,
			SECOidTag signature_algorithm,
			PK11SlotInfo *preferred_slot,
			PK11PasswordFunc pwcb,
			struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus status;
	SECOidData *oid;
	struct signed_data *signed_data;

	if (econtent_type == NULL) {
		/* Retrieve the OID for the default of "data". */
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return NULL;
		}
		econtent_type = &oid->oid;
	}

	/* Build the new structure. */
	signed_data = PORT_ArenaZAlloc(pool, sizeof(*signed_data));
	if (signed_data == NULL) {
		return NULL;
	}
	if (bcms_populate_signed_data(mcontext, pool, signed_data,
				      econtent_type, data) != SECSuccess) {
		return NULL;
	}

	/* Add the signature. */
	status = bcms_add_signer_to_signed_data(mcontext, pool,
						signed_data,
						digest_algorithm,
						signature_algorithm,
						preferred_slot,
						certdb,
						signer,
						bag,
						pwcb,
						pwcb_args);
	if (status != SECSuccess) {
		pkinit_debug(mcontext, 0, "Error signing request.\n");
		return NULL;
	}

	/* Add any certs which we need to add beyond our own. */
	bcms_add_cert_chain_to_signed_data(mcontext, pool, signed_data, signer,
					   bag);

	/* Encode the end result. */
	return bcms_encode_signed_data(mcontext, pool, signed_data);
}

/* Make a copy of the list of recipients, ensuring that the passed-in value
 * is in there somewhere. */
static struct recipient_info **
bcms_make_recipient_infos_list(struct module_context *mcontext,
			       PLArenaPool *pool, struct recipient_info **list,
			       struct recipient_info *recipient)
{
	PRBool found;
	struct recipient_info **nlist;
	struct key_trans_recipient_info *a, *b;
	struct algorithm_identifier *alga, *algb;
	int i;

	found = PR_FALSE;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		if (list[i]->recipient_info_type !=
		    recipient->recipient_info_type) {
			continue;
		}
		switch (list[i]->recipient_info_type) {
		case recipient_info_type_key_trans:
			a = &list[i]->recipient_info_value.ktri;
			b = &recipient->recipient_info_value.ktri;
			if (!SECITEM_ItemsAreEqual(&a->version, &b->version)) {
				continue;
			}
			if (a->rid.recipient_identifier_type !=
			    b->rid.recipient_identifier_type) {
				continue;
			}
			alga = &a->key_encryption_algorithm;
			algb = &b->key_encryption_algorithm;
			if (!SECITEM_ItemsAreEqual(&alga->algorithm,
						   &algb->algorithm)) {
				continue;
			}
			if (!SECITEM_ItemsAreEqual(alga->parameters,
						   algb->parameters)) {
				continue;
			}
			switch (a->rid.recipient_identifier_type) {
			case recipient_identifier_type_issuer_and_serial_number:
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.issuer_and_serial_number.issuer,
							   &b->rid.recipient_identifier_value.issuer_and_serial_number.issuer)) {
					continue;
				}
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.issuer_and_serial_number.serial,
							   &b->rid.recipient_identifier_value.issuer_and_serial_number.serial)) {
					continue;
				}
				break;
			case recipient_identifier_type_subject_key_identifier:
				if (!SECITEM_ItemsAreEqual(&a->rid.recipient_identifier_value.subject_key_identifier,
							   &b->rid.recipient_identifier_value.subject_key_identifier)) {
					continue;
				}
				break;
			default:
				break;
			}
			if (!SECITEM_ItemsAreEqual(&a->encrypted_key,
						   &b->encrypted_key)) {
				continue;
			}
			found = PR_TRUE;
			break;
		case recipient_info_type_key_agree:
			if (SECITEM_ItemsAreEqual(&list[i]->recipient_info_value.der_kari,
						  &recipient->recipient_info_value.der_kari)) {
				found = PR_TRUE;
				break;
			}
			break;
		case recipient_info_type_kek_recipient:
			if (SECITEM_ItemsAreEqual(&list[i]->recipient_info_value.der_kekri,
						  &recipient->recipient_info_value.der_kekri)) {
				found = PR_TRUE;
				break;
			}
			break;
		default:
			break;
		}
	}

	if (found) {
		nlist = list;
	} else {
		/* i contains the length of the current list */
		nlist = PORT_ArenaZAlloc(pool,
					 sizeof(struct recipient_info *) *
					 (i + 2));
		if (nlist == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		/* add our entry to the end of the list */
		nlist[i] = recipient;
		/* copy the rest of the list */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	}

	return nlist;
}

/* Like CERT_FindCertByIssuerAndSN(), but explicitly search all slots. */
struct issuer_and_sn_data {
	struct module_context *mcontext;
	CERTIssuerAndSN *isn;
	CERTCertificate *cert;
};

/* If the isuer and serial number of the given certificate matches the ones
 * passed in with the data, save a copy of the certificate. */
static SECStatus
compare_issuer_and_sn(CERTCertificate *cert, SECItem *unused_der_cert,
		      void *cb_data)
{
	struct issuer_and_sn_data *data;

	data = cb_data;
	if (data->cert != NULL) {
		return SECSuccess;
	}
	if (SECITEM_ItemsAreEqual(&cert->derIssuer,
				  &data->isn->derIssuer) &&
	    SECITEM_ItemsAreEqual(&cert->serialNumber,
				  &data->isn->serialNumber)) {
		data->cert = CERT_DupCertificate(cert);
	}
	return SECSuccess;
}

/* Search the cache of loaded certs, and then go out to all tokens. */
static CERTCertificate *
find_cert_by_issuer_and_sn(struct module_context *mcontext,
			   CERTCertDBHandle *certdb,
			   CERTIssuerAndSN *isn,
			   struct pkinit_pwcb_args *pwcb_args)
{
	struct issuer_and_sn_data data;

	data.cert = CERT_FindCertByIssuerAndSN(certdb, isn);
	if (data.cert != NULL) {
		return data.cert;
	}

	data.mcontext = mcontext;
	data.isn = isn;

	PK11_TraverseSlotCerts(compare_issuer_and_sn, &data, pwcb_args);

	return data.cert;
}

/* Find the certificate in the database which matches the recipient_identifier.
 * The returned certificate must be destroyed by the caller. */
static CERTCertificate *
find_cert_from_rid(struct module_context *mcontext,
		   CERTCertDBHandle *certdb, struct recipient_identifier *rid,
		   struct pkinit_pwcb_args *pwcb_args)
{
	struct issuer_and_serial_number *is;
	SECItem *spki;
	CERTIssuerAndSN isn;

	switch (rid->recipient_identifier_type) {
	case recipient_identifier_type_issuer_and_serial_number:
		/* The recipient is identified by issuer/serial. */
		is = &rid->recipient_identifier_value.issuer_and_serial_number;
		memset(&isn, 0, sizeof(isn));
		isn.derIssuer = is->issuer;
		isn.serialNumber = is->serial;
		return find_cert_by_issuer_and_sn(mcontext, certdb,
						  &isn, pwcb_args);
		break;
	case recipient_identifier_type_subject_key_identifier:
		/* The recipient is identified by the key identifier. */
		spki = &rid->recipient_identifier_value.subject_key_identifier;
		return CERT_FindCertBySubjectKeyID(certdb, spki);
		break;
	default:
		break;
	}
	return NULL;
}

static SECItem *
bcms_recover_enveloped_data_key(struct module_context *mcontext,
				PLArenaPool *pool,
				PK11SlotInfo *preferred_slot,
				CERTCertDBHandle *certdb,
				struct enveloped_data *env,
				struct aa_bag *bag,
				PK11PasswordFunc pwcb,
				struct pkinit_pwcb_args *pwcb_args)
{
	int i;
	struct recipient_info *recipient_infos;
	struct key_trans_recipient_info *ktri;
	CERTCertificate *cert;
	SECKEYPrivateKey *privkey;
	SECOidData *oid;
	SECItem *key;
	SECStatus status;

	key = NULL;
	for (i = 0;
	     (env->recipient_infos != NULL) &&
	     (env->recipient_infos[i] != NULL) &&
	     (key == NULL);
	     i++) {
		recipient_infos = env->recipient_infos[i];
		/* we're looking for a transported key, nothing else will do */
		if (recipient_infos->recipient_info_type !=
		    recipient_info_type_key_trans) {
			pkinit_debug(mcontext, 1,
				     "recipient-info %d is not for key "
				     "transport.\n", i);
			continue;
		}
		/* we're looking for a transported key, nothing else will do */
		ktri = &recipient_infos->recipient_info_value.ktri;
		oid = SECOID_FindOID(&ktri->key_encryption_algorithm.algorithm);
		if (oid == NULL) {
			pkinit_debug(mcontext, 1,
				     "recipient-info %d uses an unrecognized "
				     "key encryption algorithm.\n", i);
			continue;
		} else {
			pkinit_debug(mcontext, 2,
				     "recipient-info %d uses \"%s\" to encrypt "
				     "the key\n", i, oid->desc);
		}
		/* ensure we know how to use that cipher */
		switch (oid->offset) {
		case SEC_OID_PKCS1_RSA_ENCRYPTION:
			break;
		default:
			pkinit_debug(mcontext, 1,
				     "recipient-info %d uses an unsupported "
				     "key encryption algorithm.\n", i);
			continue;
			break;
		}
		/* try to find a matching cert/key pair */
		cert = find_cert_from_rid(mcontext, certdb, &ktri->rid,
					  pwcb_args);
		if (cert == NULL) {
			pkinit_debug(mcontext, 1,
				     "Couldn't find certificate for "
				     "recipient.\n");
			continue;
		}
		/* see if we have the matching private key */
		privkey = cert_find_key_for_cert(mcontext,
						 preferred_slot, cert, bag,
						 pwcb_args);
		if (privkey == NULL) {
			pkinit_debug(mcontext, 0,
				     "Unable to locate private key "
				     "corresponding to recipient "
				     "certificate \"%s\".\n",
				     cert->subjectName);
			CERT_DestroyCertificate(cert);
			continue;
		}
		CERT_DestroyCertificate(cert);
		/* try to decrypt the key data */
		switch (oid->offset) {
		case SEC_OID_PKCS1_RSA_ENCRYPTION:
			key = SECITEM_AllocItem(pool, NULL,
						ktri->encrypted_key.len);
			if (key != NULL) {
				status = PK11_PrivDecryptPKCS1(privkey,
							       key->data,
							       &key->len,
							       key->len,
							       ktri->encrypted_key.data,
							       ktri->encrypted_key.len);
				if (status != SECSuccess) {
					pkinit_debug(mcontext, 1,
						     "Error decrypting "
						     "envelope key.\n");
					key = NULL;
				} else {
					pkinit_debug(mcontext, 2,
						     "Decrypted enveloped "
						     "data key (%u bits).\n",
						     key->len * 8);
				}
			} else {
				pkinit_debug(mcontext, 0, "Out of memory.\n");
			}
			break;
		default:
			break;
		}
		SECKEY_DestroyPrivateKey(privkey);
	}
	return key;
}

static SECItem *
bcms_add_padding(PLArenaPool *pool, SECItem *unpadded_data, int block_size)
{
	unsigned char *p;
	SECItem *ret;
	unsigned char pad_len;

	pad_len = (block_size - (unpadded_data->len % block_size));
	ret = SECITEM_AllocItem(pool, NULL, unpadded_data->len + pad_len);
	if (ret != NULL) {
		if (unpadded_data->len > 0) {
			memcpy(ret->data, unpadded_data->data,
			       unpadded_data->len);
		}
		p = ret->data + unpadded_data->len;
		while (p < ret->data + unpadded_data->len + pad_len) {
			*p = pad_len;
			p++;
		}
	}
	return ret;
}

static SECItem *
bcms_remove_padding(PLArenaPool *pool, SECItem *padded_data,
		    unsigned int block_size)
{
	unsigned char *p;
	unsigned char pad_len;
	SECItem *ret;

	if ((block_size > 1) && (padded_data->len >= block_size)) {
		p = padded_data->data + padded_data->len - 1;
		pad_len = *p;
		if (pad_len <= block_size) {
			ret = SECITEM_AllocItem(pool, NULL,
						padded_data->len - pad_len);
			if (ret != NULL) {
				memcpy(ret->data, padded_data->data,
				       padded_data->len - pad_len);
				return ret;
			}
		}
	}
	return NULL;
}

/* Recover the data from an enveloped_data structure. */
SECItem *
bcms_extract_enveloped_data(struct module_context *mcontext,
			    PLArenaPool *pool,
			    CERTCertDBHandle *certdb,
			    struct enveloped_data *env,
			    PK11SlotInfo *preferred_slot,
			    struct aa_bag *bag,
			    PK11PasswordFunc pwcb,
			    struct pkinit_pwcb_args *pwcb_args)
{
	SECOidData *oid;
	SECItem *key;
	struct encrypted_content_info *eci;
	struct algorithm_identifier *enc_alg;
	struct rc2_cbc_parameters rc2_params;
	unsigned long rc2_version;
	SECItem aes_params;
	SECItem des3_params;
	CK_RC2_CBC_PARAMS rc2_cbc_params;
	CK_MECHANISM mech;
	PK11SlotInfo *slot;
	PK11SymKey *symkey;
	PK11Context *ctx;
	SECItem params, *plaintext;
	int outlen;
	unsigned int ublock_size;
	int block_size;

	eci = &env->encrypted_content_info;
	enc_alg = &eci->content_encryption_algorithm;

	/* Allocate the decryption buffer. */
	plaintext = SECITEM_AllocItem(pool, NULL, eci->encrypted_content.len);
	if (plaintext == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}

	/* Recover the payload encryption key. */
	key = bcms_recover_enveloped_data_key(mcontext, pool,
					      preferred_slot, certdb,
					      env, bag, pwcb, pwcb_args);
	if (key == NULL) {
		pkinit_debug(mcontext, 1,
			     "Error recovering temporary key.\n");
		return NULL;
	}

	/* Spot-check that we'll be able to decrypt the payload, and then parse
	 * out the algorithm-specific parameters. */
	oid = SECOID_FindOID(&enc_alg->algorithm);
	if (oid == NULL) {
		pkinit_debug(mcontext, 1,
			     "Unrecognized encryption algorithm.\n");
		return NULL;
	}
	pkinit_debug(mcontext, 2, "Recovered %d-bit CEK for use with %s.\n",
		     key->len * 8, oid->desc);
	memset(&mech, 0, sizeof(mech));
	switch (oid->offset) {
	case SEC_OID_RC2_CBC:
		block_size = 8;
		ublock_size = 8;
		if (enc_alg->parameters == NULL) {
			pkinit_debug(mcontext, 1,
				     "Expected RC2-CBC parameters, "
				     "got none.\n");
			return NULL;
		}
		if (SEC_ASN1DecodeItem(pool, &rc2_params,
				       common_rc2_cbc_parameters_template,
				       enc_alg->parameters) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "Error parsing RC2-CBC parameters.\n");
			return NULL;
		}
		if (SEC_ASN1DecodeInteger(&rc2_params.version,
					  &rc2_version) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "Error decoding RC2-CBC \"version\".\n");
			return NULL;
		}
		if (rc2_params.iv.len != ublock_size) {
			pkinit_debug(mcontext, 1,
				     "RC2-CBC IV is of unexpected size "
				     "(expected %d bits, got %d).\n",
				     ublock_size * 8, rc2_params.iv.len * 8);
			return NULL;
		}

		rc2_cbc_params.ulEffectiveBits = rc2_version;
		/* FIXME: map the RC2 version to a bit size.  Right now we just
		 * use the keying data as-is without verifying that there's a
		 * match here. */
		rc2_cbc_params.ulEffectiveBits = key->len * 8;
		memcpy(&rc2_cbc_params.iv[0], rc2_params.iv.data, block_size);
		mech.mechanism = CKM_RC2_CBC;
		mech.pParameter = &rc2_cbc_params;
		mech.ulParameterLen = sizeof(&rc2_cbc_params);
		break;
	case SEC_OID_DES_EDE3_CBC:
		block_size = 8;
		ublock_size = 8;
		if (enc_alg->parameters == NULL) {
			pkinit_debug(mcontext, 1,
				     "Expected DES3 parameters, got none.\n");
			return NULL;
		}
		if (SEC_ASN1DecodeItem(pool, &des3_params,
				       SEC_OctetStringTemplate,
				       enc_alg->parameters) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "Error parsing DES3 parameters.\n");
			return NULL;
		}
		if (des3_params.len != ublock_size) {
			pkinit_debug(mcontext, 1,
				     "RC2-CBC IV is of unexpected size "
				     "(expected %d bits, got %d).\n",
				     ublock_size * 8, des3_params.len * 8);
			return NULL;
		}
		mech.mechanism = CKM_DES3_CBC;
		mech.pParameter = des3_params.data;
		mech.ulParameterLen = des3_params.len;
		break;
	case SEC_OID_AES_128_CBC:
	case SEC_OID_AES_192_CBC:
	case SEC_OID_AES_256_CBC:
		block_size = 16;
		ublock_size = 16;
		if (enc_alg->parameters == NULL) {
			pkinit_debug(mcontext, 1,
				     "Expected AES parameters, got none.\n");
			return NULL;
		}
		if (SEC_ASN1DecodeItem(pool, &aes_params,
				       SEC_OctetStringTemplate,
				       enc_alg->parameters) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "Error parsing AES parameters.\n");
			return NULL;
		}
		if (aes_params.len != ublock_size) {
			pkinit_debug(mcontext, 1,
				     "AES-CBC IV is of unexpected size "
				     "(expected %d bits, got %d).\n",
				     ublock_size * 8, aes_params.len * 8);
			return NULL;
		}
		mech.mechanism = CKM_AES_CBC;
		mech.pParameter = aes_params.data;
		mech.ulParameterLen = aes_params.len;
		break;
	default:
		pkinit_debug(mcontext, 1,
			     "Unsupported encryption algorithm \"%s\".\n",
			     oid->desc);
		return NULL;
	}

	/* Use the PKCS11 interface to decrypt the data. */
	slot = PK11_GetBestSlot(mech.mechanism, pwcb_args);
	if (slot == NULL) {
		pkinit_debug(mcontext, 0, "Can't find slot for decryption.\n");
		return NULL;
	}
	symkey = PK11_ImportSymKey(slot, mech.mechanism, PK11_OriginUnwrap,
				   CKA_DECRYPT, key, pwcb_args);
	if (symkey == NULL) {
		PK11_FreeSlot(slot);
		pkinit_debug(mcontext, 0,
			     "Can't import temporary decryption key.\n");
		return NULL;
	}
	params.data = mech.pParameter;
	params.len = mech.ulParameterLen;
	ctx = PK11_CreateContextBySymKey(mech.mechanism, CKA_DECRYPT,
					 symkey, &params);
	if (ctx == NULL) {
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		pkinit_debug(mcontext, 0,
			     "Can't initialize decryption context.\n");
		return NULL;
	}

	if ((PK11_CipherOp(ctx, plaintext->data, &outlen, plaintext->len,
			   eci->encrypted_content.data,
			   eci->encrypted_content.len) != SECSuccess) ||
	    (PK11_Finalize(ctx) != SECSuccess)) {
		PK11_DestroyContext(ctx, PR_TRUE);
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		pkinit_debug(mcontext, 0,
			     "Error decrypting encapsulated content.\n");
		PK11_DestroyContext(ctx, PR_TRUE);
		return NULL;
	}
	PK11_FreeSymKey(symkey);
	PK11_DestroyContext(ctx, PR_TRUE);
	PK11_FreeSlot(slot);

	plaintext->len = outlen;

	/* Make a copy of the plaintext to the caller. */
	return bcms_remove_padding(pool, plaintext, ublock_size);
}

/* Add a new recipient_infos for the passed-in certificate to the list of
 * recipients in the enveloped-data structure. */
static SECStatus
bcms_add_recipient_to_enveloped_data(struct module_context *mcontext,
				     PLArenaPool *pool,
				     struct enveloped_data *env,
				     CERTCertificate *recipient,
				     SECItem *content_enc_key)
{
	struct recipient_info *rec;
	struct key_trans_recipient_info *trans;
	struct issuer_and_serial_number *isn;
	SECItem *enc;
	SECOidData *oid;
	SECKEYPublicKey *key;

	rec = PORT_ArenaZAlloc(pool, sizeof(*rec));
	if (rec == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}

	/* Mark that this is a key-transport recipient-infos. */
	rec->recipient_info_type = recipient_info_type_key_trans;

	/* Set the recipient version (1), indicating that it's issuer-serial
	 * identification. */
	trans = &rec->recipient_info_value.ktri;
	if (SEC_ASN1EncodeInteger(pool, &trans->version, 1) !=
	    &trans->version) {
		pkinit_debug(mcontext, 0, "Internal error encoding version.\n");
		return SECFailure;
	}

	/* Set the recipient's issuer-serial pair. */
	trans->rid.recipient_identifier_type =
		recipient_identifier_type_issuer_and_serial_number;
	isn = &trans->rid.recipient_identifier_value.issuer_and_serial_number;
	isn->issuer = recipient->derIssuer;
	isn->serial = recipient->serialNumber;

	/* Recover the recipient's public key. */
	key = CERT_ExtractPublicKey(recipient);
	if (key == NULL) {
		pkinit_debug(mcontext, 1,
			     "Error extracting public key from recipient "
			     "certificate.\n");
		return SECFailure;
	}
	if (key->keyType != rsaKey) {
		pkinit_debug(mcontext, 1,
			     "Unsupported client public key type.\n");
		SECKEY_DestroyPublicKey(key);
		return SECFailure;
	}

	/* Describe how we'll be encrypting the content-encryption key. */
	oid = SECOID_FindOIDByTag(SEC_OID_PKCS1_RSA_ENCRYPTION);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error.\n");
		SECKEY_DestroyPublicKey(key);
		return SECFailure;
	}
	trans->key_encryption_algorithm.algorithm = oid->oid;
	trans->key_encryption_algorithm.parameters = &bcms_null_item;

	/* Encrypt the content-encryption key. */
	pkinit_debug(mcontext, 2,
		     "Encrypting %d bytes of CEK data for \"%s\".\n",
		     content_enc_key->len, recipient->subjectName);
	enc = SECITEM_AllocItem(pool, NULL, key->u.rsa.modulus.len);
	if (PK11_PubEncryptPKCS1(key, enc->data,
				 content_enc_key->data, content_enc_key->len,
				 NULL) != SECSuccess) {
		pkinit_debug(mcontext, 0,
			     "Error encrypting content-encryption key.\n");
		SECKEY_DestroyPublicKey(key);
		return SECFailure;
	}
	trans->encrypted_key = *enc;

	/* Don't need the public key any more. */
	SECKEY_DestroyPublicKey(key);
	key = NULL;

	/* Finally, add the recipient info to the list. */
	env->recipient_infos = bcms_make_recipient_infos_list(mcontext, pool,
							      env->recipient_infos,
							      rec);
	return SECSuccess;
}

static SECStatus
bcms_populate_enveloped_data(struct module_context *mcontext,
			     PLArenaPool *pool,
			     struct enveloped_data *enveloped_data,
			     const SECItem *encrypted_content_type)
{
	struct encrypted_content_info *encrypted_info;
	SECOidData *oid;

	/* Set the version number. */
	if (SEC_ASN1EncodeInteger(pool, &enveloped_data->version, 0) !=
	    &enveloped_data->version) {
		pkinit_debug(mcontext, 0,
			     "Internal error while encoding enveloped-data "
			     "version.\n");
		return SECFailure;
	}

	/* We have no sender identification. */
	memset(&enveloped_data->der_originator_info, 0,
	       sizeof(enveloped_data->der_originator_info));

	/* Initialize as empty the list of recipients. */
	enveloped_data->recipient_infos = NULL;

	/* We have no unprotected attributes. */
	memset(&enveloped_data->der_unsigned_attrs, 0,
	       sizeof(enveloped_data->der_unsigned_attrs));

	/* Fill in the content type of the encrypted content-info. */
	encrypted_info = &enveloped_data->encrypted_content_info;
	if (encrypted_content_type == NULL) {
		/* Retrieve the OID for the default of "data". */
		oid = SECOID_FindOIDByTag(SEC_OID_PKCS7_DATA);
		if (oid == NULL) {
			return SECFailure;
		}
		encrypted_content_type = &oid->oid;
	}
	encrypted_info->content_type = *encrypted_content_type;

	return SECSuccess;
}

static SECStatus
bcms_encrypt_with_key(struct module_context *mcontext,
		      CK_MECHANISM *mech,
		      PK11Origin key_origin,
		      SECItem *key,
		      SECItem *padded_plaintext,
		      SECItem *ciphertext)
{
	PK11SlotInfo *slot;
	PK11Context *ctx;
	PK11SymKey *symkey;
	SECItem params;
	int outlen;

	/* Actually encrypt the payload, using the content encryption key. */
	slot = PK11_GetBestSlot(mech->mechanism, NULL);
	if (slot == NULL) {
		pkinit_debug(mcontext, 0,
			     "Can't find suitable slot for encrypting data.\n");
		return SECFailure;
	}
	symkey = PK11_ImportSymKey(slot, mech->mechanism, key_origin,
				   CKA_ENCRYPT, key, NULL);
	if (symkey == NULL) {
		pkinit_debug(mcontext, 0,
			     "Can't import content encryption key.\n");
		PK11_FreeSlot(slot);
		return SECFailure;
	}
	params.data = mech->pParameter;
	params.len = mech->ulParameterLen;
	ctx = PK11_CreateContextBySymKey(mech->mechanism, CKA_ENCRYPT,
					 symkey, &params);
	if (ctx == NULL) {
		pkinit_debug(mcontext, 0,
			     "Can't initialize encryption context.\n");
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		return SECFailure;
	}
	if ((PK11_CipherOp(ctx,
			   ciphertext->data,
			   &outlen,
			   ciphertext->len,
			   padded_plaintext->data,
			   padded_plaintext->len) != SECSuccess) ||
	    (PK11_Finalize(ctx) != SECSuccess)) {
		PK11_DestroyContext(ctx, PR_TRUE);
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		pkinit_debug(mcontext, 0,
			     "Error encrypting encapsulated content.\n");
		PK11_DestroyContext(ctx, PR_TRUE);
		return SECFailure;
	}
	PK11_DestroyContext(ctx, PR_TRUE);
	PK11_FreeSymKey(symkey);
	PK11_FreeSlot(slot);

	ciphertext->len = outlen;
	return SECSuccess;
}

/* Build a new encoded enveloped_data. */
SECItem *
bcms_create_enveloped_data(struct module_context *mcontext,
			   PLArenaPool *pool,
			   CERTCertDBHandle *certdb,
			   krb5_context kcontext,
			   SECItem *content_type,
			   SECOidTag preferred_cek_algorithm,
			   SECItem *data,
			   SECItem *recipient_cert)
{
	struct enveloped_data *enveloped_data;
	struct encrypted_content_info *encrypted_info;
	struct algorithm_identifier *algorithm;
	CERTCertificate **recipients;
	SECItem content_enc_key, *ciphertext, *plaintext;
	CK_MECHANISM mech;
	int block_size;
	unsigned int ublock_size;

	pkinit_debug(mcontext, 2, "Generating enveloped-data.\n");

	/* Create and populate the enveloped data structure. */
	enveloped_data = PORT_ArenaZAlloc(pool, sizeof(*enveloped_data));
	if (enveloped_data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}

	/* Build the enveloped_data structure. */
	if (SEC_ASN1EncodeInteger(pool, &enveloped_data->version, 0) !=
	    &enveloped_data->version) {
		pkinit_debug(mcontext, 0, "Internal error encoding version.\n");
		return NULL;
	}

	if (bcms_populate_enveloped_data(mcontext, pool, enveloped_data,
					 content_type) != SECSuccess) {
		return NULL;
	}

	/* Note the content encryption algorithm. */
	encrypted_info = &enveloped_data->encrypted_content_info;
	algorithm = &encrypted_info->content_encryption_algorithm;
	if (common_generate_content_encryption_key(mcontext, pool,
						   kcontext,
						   preferred_cek_algorithm,
						   algorithm,
						   &content_enc_key,
						   &mech,
						   &block_size)) {
		pkinit_debug(mcontext, 0,
			     "Error generating content encryption key.\n");
		return NULL;
	}
	ublock_size = block_size;

	/* Pad out the data to get our plaintext. */
	plaintext = bcms_add_padding(pool, data, ublock_size);
	if (plaintext == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}

	/* Allocate room for the destination buffer.  We're just guessing that
	 * this is the right size. */
	ciphertext = SECITEM_AllocItem(pool, NULL,
				       plaintext->len + (block_size * 2));
	if (ciphertext == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}

	/* Actually encrypt the payload, using the content encryption key. */
	pkinit_debug(mcontext, 2,
		     "Encrypting %d bytes of data (%d with padding) "
		     "with CEK.\n",
		     data->len, plaintext->len);
	if (bcms_encrypt_with_key(mcontext,
				  &mech,
				  PK11_OriginGenerated,
				  &content_enc_key,
				  plaintext,
				  ciphertext) != SECSuccess) {
		pkinit_debug(mcontext, 0,
			     "Error encrypting encapsulated content.\n");
		return NULL;
	}
	encrypted_info->encrypted_content = *ciphertext;

	/* Wrap up a copy of the CEK for the client. */
	recipients = NULL;
	if (CERT_ImportCerts(certdb, 0,
			     1, &recipient_cert,
			     &recipients,
			     PR_FALSE, PR_FALSE,
			     NULL) != SECSuccess) {
		pkinit_debug(mcontext, 0,
			     "Error parsing client certificate.\n");
		return NULL;
	}
	if (bcms_add_recipient_to_enveloped_data(mcontext, pool,
						 enveloped_data,
						 recipients[0],
						 &content_enc_key) != SECSuccess) {
		CERT_DestroyCertArray(recipients, 1);
		pkinit_debug(mcontext, 0,
			     "Error encrypting CEK for recipient.\n");
		return NULL;
	}
	CERT_DestroyCertArray(recipients, 1);

	/* Encode the end result. */
	return bcms_encode_enveloped_data(mcontext, pool, enveloped_data);
}

/* Add a certificate to the list, but only if it's not already in there. */
struct external_principal_identifier **
bcms_make_external_principal_identifier_list(struct module_context *mcontext,
					     PLArenaPool *pool,
					     struct external_principal_identifier **list,
					     CERTCertificate *cert)
{
	PRBool found;
	struct external_principal_identifier **nlist, *epi;
	struct issuer_and_serial_number issuer_and_serial;
	SECItem *isn, *item;
	int i;

	found = PR_FALSE;
	isn = NULL;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		epi = list[i];
		if (!SECITEM_ItemsAreEqual(&epi->der_subject_name,
					   &cert->derSubject)) {
			/* different names */
			continue;
		}
		if (isn == NULL) {
			issuer_and_serial.issuer = cert->derIssuer;
			issuer_and_serial.serial = cert->serialNumber;
			isn = common_encode_issuer_and_serial_number(mcontext,
								     pool,
								     &issuer_and_serial);
		}
		if (!SECITEM_ItemsAreEqual(isn,
					   &epi->issuer_and_serial_number)) {
			/* issuer/serial differ */
			continue;
		}
		if ((epi->subject_key_identifier.len > 0) &&
		    (cert->subjectKeyID.len > 0) &&
		    !SECITEM_ItemsAreEqual(&epi->subject_key_identifier,
					   &cert->subjectKeyID)) {
			/* different subject key identifier */
			continue;
		}
		found = PR_TRUE;
		break;
	}

	if (found) {
		nlist = list;
	} else {
		/* i contains the length of the current list */
		nlist = PORT_ArenaZAlloc(pool, sizeof(list[0]) * (i + 2));
		if (nlist == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		/* add our entry to the end of the list */
		epi = PORT_ArenaZAlloc(pool, sizeof(*epi));
		nlist[i] = epi;
		/* copy the subject name */
		item = SECITEM_ArenaDupItem(pool, &cert->derSubject);
		if (item != NULL) {
			epi->der_subject_name = *item;
		}
		/* copy the issuer/serial, creating it if need be */
		if (isn == NULL) {
			issuer_and_serial.issuer = cert->derIssuer;
			issuer_and_serial.serial = cert->serialNumber;
			isn = common_encode_issuer_and_serial_number(mcontext,
								     pool,
								     &issuer_and_serial);
		}
		if (isn != NULL) {
			epi->issuer_and_serial_number = *isn;
		}
		/* copy the subject key ID, if it's there */
		if (cert->subjectKeyID.len > 0) {
			item = SECITEM_ArenaDupItem(pool, &cert->subjectKeyID);
			if (item != NULL) {
				epi->subject_key_identifier = *item;
			}
		}
		/* copy the rest of the list */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	}

	return nlist;
}

static struct external_principal_identifier **
bcms_get_trusted_certifier_list(struct module_context *mcontext,
				PLArenaPool *pool,
				int need_trust,
				struct pkinit_pwcb_args *pwcb_args)
{
	PK11SlotList *slotlist;
	PK11SlotListElement *le;
	PK11SlotInfo *slot;
	CERTCertList *certs;
	CERTCertListNode *ln;
	CERTCertTrust trust;
	int flags, count;
	struct external_principal_identifier **epi;

	epi = NULL;
	slotlist = PK11_GetAllTokens(CKM_INVALID_MECHANISM,
				     PR_FALSE, PR_TRUE, pwcb_args);
	if (slotlist != NULL) {
		for (le = slotlist->head; le != NULL; le = le->next) {
			slot = le->slot;
			if (!PK11_IsFriendly(slot) || PK11_NeedLogin(slot)) {
				pkinit_debug(mcontext, 2,
					     "Slot \"%s\" locked.\n",
					     PK11_GetTokenName(slot));
				continue;
			} else {
				pkinit_debug(mcontext, 2,
					     "Looking for trusted roots in "
					     "\"%s\".\n",
					     PK11_GetTokenName(slot));
			}
			certs = PK11_ListCertsInSlot(slot);
			count = 0;
			if (certs != NULL) {
				for (ln = CERT_LIST_HEAD(certs);
				     !CERT_LIST_EMPTY(certs) &&
				     !CERT_LIST_END(ln, certs);
				     ln = CERT_LIST_NEXT(ln)) {
					if (ln->cert->isRoot) {
						count++;
						if (CERT_GetCertTrust(ln->cert,
								      &trust) == SECSuccess) {
							flags = trust.sslFlags;
							if ((flags & need_trust) == need_trust) {
								pkinit_debug(mcontext, 2, "Found trusted certificate for \"%s\".\n",
									     ln->cert->subjectName);
								epi = bcms_make_external_principal_identifier_list(mcontext, pool, epi, ln->cert);
							} else {
								pkinit_debug(mcontext, 2, "Found untrusted certificate for \"%s\".\n",
									     ln->cert->subjectName);
							}
						} else {
							pkinit_debug(mcontext, 0, "Error getting trust for \"%s\".\n",
								     ln->cert->subjectName);
						}
					} else {
						pkinit_debug(mcontext, 2, "Found non-root certificate for \"%s\".\n",
							     ln->cert->subjectName);
					}
				}
				CERT_DestroyCertList(certs);
			}
			if (count == 0) {
				pkinit_debug(mcontext, 2,
					     "No trusted roots in \"%s\".\n",
					     PK11_GetTokenName(slot));
			}
		}
		PK11_FreeSlotList(slotlist);
	}
	for (count = 0; (epi != NULL) && (epi[count] != NULL); count++) {
		continue;
	}
	pkinit_debug(mcontext, 2, "Found %d trusted roots in all.\n", count);
	return epi;
}

struct external_principal_identifier **
bcms_get_trusted_client_certifier_list(struct module_context *mcontext,
				       PLArenaPool *pool,
				       struct pkinit_pwcb_args *pwcb_args)
{
	return bcms_get_trusted_certifier_list(mcontext, pool,
					       CERTDB_TRUSTED_CLIENT_CA,
					       pwcb_args);
}

struct external_principal_identifier **
bcms_get_trusted_kdc_certifier_list(struct module_context *mcontext,
				    PLArenaPool *pool,
				    struct pkinit_pwcb_args *pwcb_args)
{
	return bcms_get_trusted_certifier_list(mcontext, pool,
					       CERTDB_TRUSTED_CA,
					       pwcb_args);
}
