/*
 * Copyright (C) 2006,2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: certs.c,v 1.37 2007/06/08 21:44:27 nalin Exp $"

#include "../config.h"

#include <stdio.h>
#include <time.h>

#include <krb5/krb5.h>

#include <nss.h>
#include <cert.h>
#include <certdb.h>
#include <secerr.h>
#include <secmodt.h>

#include "commont.h"
#include "bcmst.h"
#include "pkinitt.h"
#include "certs.h"
#include "pkinit.h"

#ifdef USE_PAL_BACKPORT
#include "backport-errors.h"
#endif

#define APPDEFAULT_LIST_SEPARATORS " \t,"

static unsigned char oid_ms_sc_login_data[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02, 0x02};
static const SECOidData
oid_ms_sc_login = {
	.oid = {
		.data = oid_ms_sc_login_data,
		.len = 10,
	},
	.offset = 0,
	.desc = "Microsoft Windows SmartCard Logon",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_ms_upn_name_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02, 0x03};
static const SECOidData
oid_ms_upn_name = {
	.oid = {
		.data = oid_ms_upn_name_bytes,
		.len = 10,
	},
	.offset = 0,
	.desc = "Microsoft Windows User Principal Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char
oid_ms_guid_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x19, 0x01};
static const SECOidData
oid_ms_guid = {
	.oid = {
		.data = oid_ms_guid_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Microsoft Globally-Unique Identifier",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char
oid_ms_certificate_template_name_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01, 0x82, 0x37, 0x14, 0x02};
static const SECOidData
oid_ms_certificate_template_name = {
	.oid = {
		.data = oid_ms_certificate_template_name_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Microsoft Certificate Template Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

/* pkinit-SAN 1.3.6.1.5.2.2 */
static unsigned char oid_pkinit_san_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x02};
static const SECOidData
oid_pkinit_san = {
	.oid = {
		.data = oid_pkinit_san_bytes,
		.len = 6,
	},
	.offset = 0,
	.desc = "PKINIT Subject Alternate Name",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_auth_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x01};
static const SECOidData
oid_pkinit_auth_data = {
	.oid = {
		.data = oid_pkinit_auth_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Authentication Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_rkey_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x03};
static const SECOidData
oid_pkinit_rkey_data = {
	.oid = {
		.data = oid_pkinit_rkey_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Reply Key Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_dhkey_data_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x02};
static const SECOidData
oid_pkinit_dhkey_data = {
	.oid = {
		.data = oid_pkinit_dhkey_data_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Diffie-Hellman Reply Key Data",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_key_purpose_client_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x04};
static const SECOidData
oid_pkinit_key_purpose_client = {
	.oid = {
		.data = oid_pkinit_key_purpose_client_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Key Purpose: Client",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_pkinit_key_purpose_kdc_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x03, 0x05};
static const SECOidData
oid_pkinit_key_purpose_kdc = {
	.oid = {
		.data = oid_pkinit_key_purpose_kdc_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "PKINIT Key Purpose: KDC",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_dh_public_number_bytes[] = {0x2a, 0x86, 0x48, 0xce, 0x3e, 0x02, 0x01};
static const SECOidData
oid_dh_public_number = {
	.oid = {
		.data = oid_dh_public_number_bytes,
		.len = 7,
	},
	.offset = 0,
	.desc = "Diffie-Hellman Public Number",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

static unsigned char oid_dh_key_agreement_bytes[] = {0x2a, 0x86, 0x48, 0x86, 0xf6, 0x33, 0x01, 0x03, 0x01};
static const SECOidData
oid_dh_key_agreement = {
	.oid = {
		.data = oid_dh_key_agreement_bytes,
		.len = 9,
	},
	.offset = 0,
	.desc = "Diffie-Hellman Key Agreement",
	.mechanism = 0,
	.supportedExtension = UNSUPPORTED_CERT_EXTENSION,
};

/* Template for an otherName.  If it's a MS-UPN, then the data is an explicit
 * UTF8String. If it's a PKINIT-SAN, it's a structure. */
static SEC_ASN1Template
pkinit_name_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = pkinit_kerberos_principal_name_template,
	.size = sizeof(struct kerberos_principal_name),
	},
};
static SEC_ASN1Template
ms_upn_name_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT |
		SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = SEC_UTF8StringTemplate,
	.size = sizeof(SECItem),
	},
};
static SEC_ASN1Template
ms_guid_template[] = {
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_EXPLICIT | SEC_ASN1_CONSTRUCTED,
	.offset = 0,
	.sub = SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
};
static SEC_ASN1Template
other_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct other_name),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct other_name, oid),
	.sub = NULL,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED,
	.offset = offsetof(struct other_name, data),
	.sub = SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* Template for decoding the choice of subjectAltName. */
static SEC_ASN1Template
san_choice_template[] = {
	{
	.kind = SEC_ASN1_CHOICE,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_type), /* id */
	.sub = NULL,
	.size = sizeof(struct subject_alt_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.other_name),
	.sub = other_name_template,
	.size = subject_alt_name_other_name, /* identifier to save for this choice */
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.rfc822_name),
	.sub = &SEC_IA5StringTemplate,
	.size = subject_alt_name_rfc822_name, /* identifier to save for this choice */
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 2,
	.offset = offsetof(struct subject_alt_name,
			   subject_alt_name_value.dnsname),
	.sub = &SEC_IA5StringTemplate,
	.size = subject_alt_name_dnsname, /* identifier to save for this choice */
	},
	{0, 0, NULL, 0},
};

/* Template for unpacking the subjectAltName data sequence into an
 * array of other_name structures. */
static SEC_ASN1Template
san_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = &san_choice_template,
	.size = 0,
	},
};

/* Template for unpacking an EKU list. */
static SEC_ASN1Template
eku_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = SEC_ObjectIDTemplate,
	.size = 0,
	},
};

const SECItem *
cert_get_oid_pkinit_auth_data(void)
{
	return &oid_pkinit_auth_data.oid;
}
const SECOidData *
cert_get_oid_pkinit_auth_data_oid(void)
{
	return &oid_pkinit_auth_data;
}
const SECItem *
cert_get_oid_pkinit_rkey_data(void)
{
	return &oid_pkinit_rkey_data.oid;
}
const SECOidData *
cert_get_oid_pkinit_rkey_data_oid(void)
{
	return &oid_pkinit_rkey_data;
}
const SECItem *
cert_get_oid_pkinit_dhkey_data(void)
{
	return &oid_pkinit_dhkey_data.oid;
}
const SECItem *
cert_get_oid_dh_public_number(void)
{
	return &oid_dh_public_number.oid;
}
const SECItem *
cert_get_oid_dh_key_agreement(void)
{
	return &oid_dh_key_agreement.oid;
}

/* Check if the certificate contains the desired OID as an EKU value. */
static SECStatus
cert_eku_matches_oid(struct module_context *mcontext,
		     CERTCertificate *cert, const SECItem *oid, PRBool *matches)
{
	SECItem eku_value;
	const SECItem **oids;
	SECOidData *eku;
	int i;

	eku = SECOID_FindOIDByTag(SEC_OID_X509_EXT_KEY_USAGE);
	if (eku == NULL) {
		pkinit_debug(mcontext, 0, "Internal error: no knowledge of the "
			     "X.509 extendedKeyUsage type.\n");
		return SECFailure;
	}
	if (CERT_FindCertExtension(cert, eku->offset,
				   &eku_value) == SECSuccess) {
		oids = NULL;
		*matches = PR_FALSE;
		if (SEC_ASN1DecodeItem(cert->arena, &oids, eku_template,
				       &eku_value) == SECSuccess) {
			for (i = 0; (oids != NULL) && (oids[i] != NULL); i++) {
				if (SECITEM_ItemsAreEqual(oids[i], oid)) {
					pkinit_debug(mcontext, 2,
						     "Found desired EKU.\n");
					*matches = PR_TRUE;
					break;
				}
			}
			return SECSuccess;
		} else {
			pkinit_debug(mcontext, 1,
				     "Error parsing extendedKeyUsage "
				     "extension.\n");
			return SECFailure;
		}
	}

	pkinit_debug(mcontext, 1,
		     "Certificate has no extendedKeyUsage extension.\n");
	return SECFailure;
}

static void
derdump(SECItem *item)
{
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		fprintf(stderr, "Dumping %d bytes.\n", item->len);
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
}

/* Check if the certificate contains the tagged OID as an EKU value. */
static SECStatus
cert_eku_matches_tag(struct module_context *mcontext,
		     CERTCertificate *cert, SECOidTag eku, PRBool *matches)
{
	SECOidData *oid;
	oid = SECOID_FindOIDByTag(eku);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error: no knowldege of the "
			     "extendedKeyUsage extension.\n");
		return SECFailure;
	}
	return cert_eku_matches_oid(mcontext, cert, &oid->oid, matches);
}

/* Compare a SECItem and a krb5_data item. */
static PRBool
cert_secitem_equals_krb5_data(SECItem *item, krb5_data *data)
{
	return (item->len == data->length) &&
	       (memcmp(item->data, data->data, item->len) == 0);
}
static PRBool
cert_secitem_list_equals_krb5_data_list(SECItem **items, krb5_data *list,
					int length)
{
	int i;
	for (i = 0; i < length; i++) {
		if ((items == NULL) || (items[i] == NULL)) {
			return PR_FALSE;
		}
	}
	if (items[i] != NULL) {
		return PR_FALSE;
	}
	for (i = 0; i < length; i++) {
		if (!cert_secitem_equals_krb5_data(items[i], &list[i])) {
			return PR_FALSE;
		}
	}
	return PR_TRUE;
}

/* Retrieve an extension based on its OID. */
static SECStatus
cert_find_extension_oid(CERTCertificate *cert,
			const SECItem *oid, SECItem *value)
{
	int i;
	for (i = 0;
	     (cert->extensions != NULL) && (cert->extensions[i] != NULL);
	     i++) {
		if (SECITEM_ItemsAreEqual(&cert->extensions[i]->id, oid)) {
			if (value != NULL) {
				*value = cert->extensions[i]->value;
			}
			return SECSuccess;
		}
	}
	return SECFailure;
}

/* Check if the certificate certificateTemplateName matches the given string. */
static SECStatus
cert_template_name_matches(struct module_context *mcontext,
			   CERTCertificate *cert, const char *ascii_value,
			   PRBool *matches)
{
	SECItem template_name, name;
	unsigned int i;
	if (cert_find_extension_oid(cert,
				    &oid_ms_certificate_template_name.oid,
				    &template_name) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No certificateTemplateName extension.\n");
		return SECFailure;
	}
	if (SEC_ASN1DecodeItem(cert->arena, &name,
			       SEC_BMPStringTemplate,
			       &template_name) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error decoding certificateTemplateName.\n");
		return SECFailure;
	}
	/* Yes, this is cheap. But short of doing character set conversion,
	 * it'll do for now. */
	for (i = 0; ascii_value[i] != '\0'; i++) {
		/* Are we out of certificateTemplateName value? */
		if ((i * 2) >= name.len) {
			pkinit_debug(mcontext, 2,
				     "certificateTemplateName was shorter than "
				     "we expected.\n");
			break;
		}
		/* Match one character. */
		if ((((char*)name.data)[i * 2] != '\0') ||
		    (((char*)name.data)[i * 2 + 1] != ascii_value[i])) {
			break;
		}
	}
	if ((i > 0) && (ascii_value[i] == '\0') && ((i * 2) == name.len)) {
		*matches = PR_TRUE;
		pkinit_debug(mcontext, 2,
			     "certificateTemplateName matched \"%s\"\n",
			     ascii_value);
	} else {
		pkinit_debug(mcontext, 2,
			     "certificateTemplateName did not match \"%s\"\n",
			     ascii_value);
		pkinit_debug(mcontext, 2,
			     "certificateTemplateName value was \"%.*s\"\n",
			     name.len, name.data);
	}
	return SECSuccess;
}

static PRBool
check_item_in_list(const SECItem *item, const char *list)
{
	const char *p, *q;
	p = list;
	p += strspn(p, APPDEFAULT_LIST_SEPARATORS);
	while ((p != NULL) && (*p != '\0')) {
		q = p + strcspn(p, APPDEFAULT_LIST_SEPARATORS);
		if ((item->len == q - p) &&
	    	    (memcmp(item->data, p, q - p) == 0)) {
			return PR_TRUE;
		}
		p = q + strspn(q, APPDEFAULT_LIST_SEPARATORS);
	}
	return PR_FALSE;
}

/* Check if the certificate has a subjectAltName value matching the DNS name. */
static SECStatus
cert_san_matches_dns_for_realm(struct module_context *mcontext,
			       CERTCertificate *cert,
			       krb5_context kcontext, krb5_principal server,
			       PRBool *matches)
{
	struct subject_alt_name **names;
	SECItem san_value, name;
	SECOidData *oid;
	CERTName *subject;
	CERTRDN *rdn;
	CERTAVA *ava;
	char *trusted_servers;
	int i, j;

	/* Read the list of trusted servers. */
	krb5_appdefault_string(kcontext, "pkinit",
			       krb5_princ_realm(kcontext, server),
			       "trusted_servers", "", &trusted_servers);

	/* Pick apart the name.  Windows accepts a host name in the CN part of
	 * the subject name, so we should do the same. */
	subject = &cert->subject;
	for (i = 0;
	     (subject->rdns != NULL) && (subject->rdns[i] != NULL);
	     i++) {
		rdn = subject->rdns[i];
		for (j = 0;
		     (rdn->avas != NULL) && (rdn->avas[j] != NULL);
		     j++) {
			ava = rdn->avas[j];
			oid = SECOID_FindOID(&ava->type);
			if ((oid != NULL) &&
			    (oid->offset == SEC_OID_AVA_COMMON_NAME)) {
				if (check_item_in_list(&ava->value,
						       trusted_servers)) {
					pkinit_debug(mcontext, 2,
						     "Found CN=%.*s in "
						     "subject name.\n",
						     ava->value.len,
						     ava->value.data);
					*matches = PR_TRUE;
				}
			}
		}
	}

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No subjectAltName extension in cert while "
			     "checking for matching DNS name.\n");
		return SECFailure;
	}

	/* Split up the subjectAltName sequence. */
	names = NULL;
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error parsing subjectAltName extension.\n");
		return SECFailure;
	}

	/* Iterate over all of the values. */
	*matches = PR_FALSE;
	for (i = 0; (names != NULL) && (names[i] != NULL) && !(*matches); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_dnsname:
			name = names[i]->subject_alt_name_value.dnsname;
			pkinit_debug(mcontext, 2, "SAN %d is DNS: \"%.*s\".\n",
				     i + 1, name.len, name.data);
			if (check_item_in_list(&name, trusted_servers)) {
				*matches = PR_TRUE;
			}
			break;
		default:
			pkinit_debug(mcontext, 2,
				     "SAN %d is type %d, not an otherName.\n",
				     i + 1, names[i]->subject_alt_name_type);
			break;
		}
	}

	pkinit_debug(mcontext, 2, "%d subjectAltName values found\n", i);
	return SECSuccess;
}

/* Check if the certificate subjectAltName UPN value matches the principal. */
static SECStatus
cert_san_matches_upn(struct module_context *mcontext, CERTCertificate *cert,
		     krb5_context kcontext, krb5_principal principal,
		     PRBool *matches)
{
	struct subject_alt_name **names;
	SECItem san_value, unparsed_name, ms_upn_name;
	char *unparsed;
	int i;

	i = 0;
	names = NULL;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No subjectAltName extension in cert while "
			     "checking for matching UPN.\n");
		return SECFailure;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error parsing subjectAltName extension.\n");
		return SECFailure;
	}

	/* Create the UTF8String representation of the principal name. */
	if (krb5_unparse_name(kcontext, principal, &unparsed) != 0) {
		pkinit_debug(mcontext, 2,
			     "Error unparsing krb5_principal.\n");
		return SECFailure;
	}
	unparsed_name.data = (unsigned char *) unparsed;
	unparsed_name.len = strlen(unparsed);

	/* Iterate over all of the values. */
	*matches = PR_FALSE;
	for (i = 0; (names != NULL) && (names[i] != NULL) && !(*matches); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a Microsoft UPN value. */
			if (SECITEM_ItemsAreEqual(&names[i]->subject_alt_name_value.other_name.oid,
						  &oid_ms_upn_name.oid)) {
				pkinit_debug(mcontext, 2,
					     "SAN %d is a UPN.\n", i + 1);
				/* If we can decode the UTF-8 string. */
				if (SEC_ASN1DecodeItem(cert->arena,
						       &ms_upn_name,
						       ms_upn_name_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					/* And it matches, then we're okay. */
					if (SECITEM_ItemsAreEqual(&ms_upn_name,
								  &unparsed_name)) {
						pkinit_debug(mcontext, 2,
							     "UPN Matched.\n");
						*matches = PR_TRUE;
					} else {
						pkinit_debug(mcontext, 2,
							     "\"%.*s\" != "
							     "\"%.*s\"\n",
							     ms_upn_name.len,
							     ms_upn_name.data,
							     unparsed_name.len,
							     unparsed_name.data);
					}
				} else {
					pkinit_debug(mcontext, 2,
						     "Error decoding UPN.\n");
				}
			} else {
				pkinit_debug(mcontext, 2,
					     "SAN %d is not a UPN.\n", i + 1);
			}
			break;
		default:
			pkinit_debug(mcontext, 2,
				     "SAN %d is type %d, not an otherName.\n",
				     i + 1, names[i]->subject_alt_name_type);
			break;
		};
	}
	krb5_free_unparsed_name(kcontext, unparsed);
	pkinit_debug(mcontext, 2, "%d subjectAltName values found\n", i);
	return SECSuccess;
}

/* Check if the certificate subjectAltName matches the principal. */
static SECStatus
cert_san_matches_principal(struct module_context *mcontext,
			   CERTCertificate *cert,
			   krb5_principal principal,
			   PRBool *matches)
{
	struct subject_alt_name **names;
	SECItem san_value;
	struct kerberos_principal_name kerberos_principal_name;
	int i;

	i = 0;
	names = NULL;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No subjectAltName extension in cert while "
			     "checking for matching principal.\n");
		return SECFailure;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error parsing subjectAltName extension.\n");
		return SECFailure;
	}

	/* Iterate over all of the values. */
	*matches = PR_FALSE;
	for (i = 0; (names != NULL) && (names[i] != NULL) && !(*matches); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a PKINIT SAN value. */
			if (SECITEM_ItemsAreEqual(&names[i]->subject_alt_name_value.other_name.oid,
						  &oid_pkinit_san.oid)) {
				pkinit_debug(mcontext, 2,
					     "SAN %d is a Kerberos name.\n",
					     i + 1);
				/* If we can decode the PKINIT SAN. */
				memset(&kerberos_principal_name, 0,
				       sizeof(kerberos_principal_name));
				if (SEC_ASN1DecodeItem(cert->arena,
						       &kerberos_principal_name,
						       pkinit_name_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					/* And it matches, then we're okay. */
					if (cert_secitem_equals_krb5_data(&kerberos_principal_name.realm.name, krb5_princ_realm(kcontext, principal)) &&
					    cert_secitem_list_equals_krb5_data_list(kerberos_principal_name.principal_name.name_string, principal->data, krb5_princ_size(kcontext, principal))) {
						pkinit_debug(mcontext, 2,
							     "Principal matched.\n");
						*matches = PR_TRUE;
					} else {
						pkinit_debug(mcontext, 2,
							     "Principal did not match.\n");
					}
				} else {
					pkinit_debug(mcontext, 2,
						     "Error decoding principal "
						     "name.\n");
				}
			} else {
				pkinit_debug(mcontext, 2,
					     "SAN %d is not a Kerberos name.\n",
					     i + 1);
			}
			break;
		default:
			pkinit_debug(mcontext, 2,
				     "SAN %d is type %d, not an otherName.\n",
				     i + 1, names[i]->subject_alt_name_type);
			break;
		};
	}
	pkinit_debug(mcontext, 2, "%d subjectAltName values found.\n", i);
	return SECSuccess;
}

static PRBool
cert_matches_epi(struct module_context *mcontext,
		 PLArenaPool *pool,
		 CERTCertificate *candidate,
		 struct external_principal_identifier *epi)
{
	struct issuer_and_serial_number *isn;
	if ((candidate->derSubject.len > 0) &&
	    (epi->der_subject_name.len > 0) &&
	    !SECITEM_ItemsAreEqual(&candidate->derSubject,
				   &epi->der_subject_name)) {
		return PR_FALSE;
	}
	if ((candidate->subjectKeyID.len > 0) &&
	    (epi->subject_key_identifier.len > 0) &&
	    !SECITEM_ItemsAreEqual(&candidate->subjectKeyID,
				   &epi->subject_key_identifier)) {
		return PR_FALSE;
	}
	if (epi->issuer_and_serial_number.len > 0) {
		isn = common_decode_issuer_and_serial_number(mcontext, pool,
							     &epi->issuer_and_serial_number);
		if (isn != NULL) {
			if (!SECITEM_ItemsAreEqual(&candidate->derIssuer,
						   &isn->issuer)) {
				return PR_FALSE;
			}
			if (!SECITEM_ItemsAreEqual(&candidate->serialNumber,
						   &isn->serial)) {
				return PR_FALSE;
			}
		}
	}
	return PR_TRUE;
}

CERTCertificate *
cert_find_cert_issuer(CERTCertificate *cert, int64 valid_time,
		      SECCertUsage usage, struct aa_bag *bag)
{
	CERTCertificate *ret;
	CERTCertList *cert_list;
	CERTCertListNode *node;
	ret = CERT_FindCertIssuer(cert, valid_time, usage);
	if (ret != NULL) {
		return ret;
	}
	ret = NULL;
	cert_list = aa_bag_find_cert_by_subject(bag,
						&cert->derIssuer,
						NULL);
	if (cert_list != NULL) {
		for (node = CERT_LIST_HEAD(cert_list);
		     !CERT_LIST_EMPTY(cert_list) &&
		     !CERT_LIST_END(node, cert_list);
		     node = CERT_LIST_NEXT(node)) {
			if (CERT_CheckCertValidTimes(node->cert,
						     valid_time ?
						     valid_time :
						     PR_Now(),
						     PR_TRUE) ==
			    secCertTimeValid) {
				ret = CERT_DupCertificate(node->cert);
				break;
			}
		}
		CERT_DestroyCertList(cert_list);
	}
	return ret;
}

static PRBool
cert_have_key_for_cert(struct module_context *mcontext,
		       PLArenaPool *pool, PK11SlotInfo *slot,
		       CERTCertificate *cert, struct aa_bag *bag,
		       struct pkinit_pwcb_args *pwcb_args)
{
	SECKEYPrivateKey *privkey;
	privkey = cert_find_key_for_cert(mcontext, slot, cert, bag, pwcb_args);
	if (privkey != NULL) {
		SECKEY_DestroyPrivateKey(privkey);
		return PR_TRUE;
	}
	return PR_FALSE;
}

static PRBool
cert_is_preferred(struct module_context *mcontext, PLArenaPool *pool,
		  CERTCertificate *candidate, struct aa_bag *bag,
		  struct external_principal_identifier **restricted_ca_list,
		  struct external_principal_identifier **banned_cert_list,
		  const char **additional_dn_list,
		  krb5_context kcontext, krb5_principal principal,
		  PRInt32 cert_flags,
		  struct pkinit_pwcb_args *pwcb_args)
{
	PRBool match, saw_restricted_ca, saw_banned_cert;
	CERTCertificate *issuer, *cert;
	unsigned int i;

	if (candidate == NULL) {
		return PR_FALSE;
	}
	if ((cert_flags & CERT_NAME_MASK) == 0) {
		return PR_FALSE;
	}
	/* Only consider certificates for which we can construct an entire
	 * certificate chain. */
	issuer = CERT_DupCertificate(candidate);
	saw_restricted_ca = PR_FALSE;
	saw_banned_cert = PR_FALSE;
	/* Check that this isn't a cert which is banned outright. */
	for (i = 0;
	     (banned_cert_list != NULL) &&
	     (banned_cert_list[i] != NULL);
	     i++) {
		if (cert_matches_epi(mcontext, pool, candidate,
				     banned_cert_list[i])) {
			pkinit_debug(mcontext, 2,
				     "Certificate \"%s\" is not to be used.\n",
				     issuer->subjectName);
			saw_banned_cert = PR_TRUE;
			break;
		}
	}
	/* Now walk the certifying chain, checking for restricted and banned CA
	 * certificates. */
	while (issuer != NULL) {
		cert = issuer;
		if ((krb5_princ_size(kcontext, principal) > 0) &&
		    (krb5_princ_component(kcontext, principal,
		   			  0)->length == KRB5_TGS_NAME_SIZE) &&
		    (memcmp(krb5_princ_component(kcontext, principal, 0)->data,
		   				 KRB5_TGS_NAME,
						 KRB5_TGS_NAME_SIZE) == 0)) {
			issuer = cert_find_cert_issuer(cert, 0,
						       certUsageSSLServer,
						       bag);
		} else {
			issuer = cert_find_cert_issuer(cert, 0,
						       certUsageSSLClient,
						       bag);
		}
		CERT_DestroyCertificate(cert);
		cert = NULL;
		if (issuer != NULL) {
			for (i = 0;
			     (restricted_ca_list != NULL) &&
			     (restricted_ca_list[i] != NULL);
			     i++) {
				if (cert_matches_epi(mcontext, pool, issuer,
						     restricted_ca_list[i])) {
					saw_restricted_ca = PR_TRUE;
				}
			}
			for (i = 0;
			     (banned_cert_list != NULL) &&
			     (banned_cert_list[i] != NULL);
			     i++) {
				if (cert_matches_epi(mcontext, pool, issuer,
						     banned_cert_list[i])) {
					pkinit_debug(mcontext, 2,
						     "Issuer \"%s\" is not to "
						     "be used.\n",
						     issuer->subjectName);
					saw_banned_cert = PR_TRUE;
				}
			}
			if (issuer->isRoot) {
				break;
			}
		} else {
			pkinit_debug(mcontext, 2,
				     "No next link in certifying chain.\n");
		}
	}
	/* Catch incomplete chains. */
	if (issuer == NULL) {
		/* Trust chain doesn't lead to a known root. */
		return PR_FALSE;
	} else {
		CERT_DestroyCertificate(issuer);
		issuer = NULL;
	}
	/* Banned cert somewhere = don't use. */
	if (saw_banned_cert) {
		pkinit_debug(mcontext, 2, "Marking certificate as not-okay.\n");
		return PR_FALSE;
	}
	/* If we had a list of valid CAs, and we didn't see one, don't use. */
	if ((restricted_ca_list != NULL) && !saw_restricted_ca) {
		pkinit_debug(mcontext, 2,
			     "Certificate was not issued by a given CA.\n");
		return PR_FALSE;
	}
	/* The PKINIT certificate will have our principal name encoded as a
	 * krb5PrincipalName. */
	if (cert_flags & CERT_MATCHES_KRB_SAN) {
		cert_flags &= ~CERT_MATCHES_KRB_SAN;
		if ((cert_san_matches_principal(mcontext, candidate,
						principal,
						&match) != SECSuccess) ||
		    !match) {
			if (!(cert_flags & CERT_NAME_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		} else {
			/* No need to check other name types. */
			pkinit_debug(mcontext, 2,
				     "Kerberos subjectAltName Matched.\n");
			cert_flags &= ~CERT_NAME_MASK;
		}
	}
	/* The Windows certificate will have our principal name encoded as a
	 * userPrincipalName, and *may* have the Client Authentication and Smart
	 * Card Logon EKUs set. */
	if (cert_flags & CERT_MATCHES_UPN_SAN) {
		cert_flags &= ~CERT_MATCHES_UPN_SAN;
		if ((cert_san_matches_upn(mcontext, candidate,
					  kcontext, principal,
					  &match) != SECSuccess) ||
		    !match) {
			if (!(cert_flags & CERT_NAME_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		} else {
			/* No need to check other name types. */
			pkinit_debug(mcontext, 2, "UPN Matched.\n");
			cert_flags &= ~CERT_NAME_MASK;
		}
	}
	/* A Domain Controller certificate may only have a trusted server's DNS
	 * name encoded in it. */
	if (cert_flags & CERT_MATCHES_TRUSTED_SERVERS) {
		cert_flags &= ~CERT_MATCHES_TRUSTED_SERVERS;
		if ((cert_san_matches_dns_for_realm(mcontext, candidate,
						    kcontext, principal,
						    &match) != SECSuccess) ||
		    !match) {
			if (!(cert_flags & CERT_NAME_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		} else {
			/* No need to check other name types. */
			pkinit_debug(mcontext, 2, "DNS name matched.\n");
			cert_flags &= ~CERT_NAME_MASK;
		}
	}
	/* The certificate may not have a principal name included as a SAN
	 * value, so check the (potentially very long) list for a match. */
	if (cert_flags & CERT_MATCHES_DN_LIST) {
		cert_flags &= ~CERT_MATCHES_DN_LIST;
		for (i = 0;
		     (additional_dn_list != NULL) &&
		     (additional_dn_list[i] != NULL);
		     i++) {
			if ((candidate->subjectName != NULL) &&
			    (strcmp(candidate->subjectName,
				    additional_dn_list[i]) == 0)) {
				break;
			}
		}
		if ((additional_dn_list == NULL) ||
		    (additional_dn_list[i] == NULL)) {
			if (!(cert_flags & CERT_NAME_MASK)) {
				/* Return the error if we have no more name
				 * types to try matching. */
				return PR_FALSE;
			}
		} else {
			/* No need to check other name types. */
			pkinit_debug(mcontext, 2,
				     "subjectAltName matched mapping.\n");
			cert_flags &= ~CERT_NAME_MASK;
		}
	}
	/* Now check for the EKU values which were specified: normal client
	 * authentication. */
	if (cert_flags & CERT_MATCHES_CLIENT_AUTH_EKU) {
		if ((cert_eku_matches_tag(mcontext, candidate,
					  SEC_OID_EXT_KEY_USAGE_CLIENT_AUTH,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* Microsoft Smartcard Logon */
	if (cert_flags & CERT_MATCHES_MS_SMARTCARD_EKU) {
		if ((cert_eku_matches_oid(mcontext, candidate,
					  &oid_ms_sc_login.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* Server Authentication */
	if (cert_flags & CERT_MATCHES_SERVER_AUTH_EKU) {
		if ((cert_eku_matches_tag(mcontext, candidate,
					  SEC_OID_EXT_KEY_USAGE_SERVER_AUTH,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* PKINIT Client */
	if (cert_flags & CERT_MATCHES_PKINIT_CLIENT_EKU) {
		if ((cert_eku_matches_oid(mcontext, candidate,
					  &oid_pkinit_key_purpose_client.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	/* PKINIT KDC */
	if (cert_flags & CERT_MATCHES_PKINIT_KDC_EKU) {
		if ((cert_eku_matches_oid(mcontext, candidate,
					  &oid_pkinit_key_purpose_kdc.oid,
					  &match) != SECSuccess) ||
		    !match) {
			return PR_FALSE;
		}
	}
	return PR_TRUE;
}

static CERTCertificate *
cert_find_preferred_cert_using_slot_or_bag(struct module_context *mcontext,
					   PLArenaPool *pool,
					   PK11SlotInfo *slot,
					   struct aa_bag *bag,
					   struct pkinit_pwcb_args *pwcb_args,
					   struct external_principal_identifier **restricted_ca_list,
					   struct external_principal_identifier **banned_cert_list,
					   const char **additional_dn_list,
					   krb5_context kcontext,
					   krb5_principal principal,
					   PRInt32 cert_flags)
{
	CERTCertificate *cert, *ret = NULL;
	CERTCertList *cert_list;
	CERTCertListNode *node;

	cert_list = NULL;
	if (slot != NULL) {
		pkinit_debug(mcontext, 2,
			     "Scanning slot \"%s\" for certificates.\n",
			     PK11_GetSlotName(slot));
		if (PK11_Authenticate(slot, PR_TRUE, pwcb_args) != SECSuccess) {
			pkinit_debug(mcontext, 1,
				     "PK11_Authenticate() failed for "
				     "slot \"%s\".\n",
				     PK11_GetSlotName(slot));
			return NULL;
		}
		cert_list = PK11_ListCertsInSlot(slot);
		if ((cert_list == NULL) || CERT_LIST_EMPTY(cert_list)) {
			pkinit_debug(mcontext, 2,
				     "No certificates found.\n");
			return NULL;
		}
	} else {
		pkinit_debug(mcontext, 2,
			     "Scanning text files for certificates.\n");
		cert_list = aa_bag_list_loaded_certs(bag);
		if ((cert_list == NULL) || CERT_LIST_EMPTY(cert_list)) {
			pkinit_debug(mcontext, 2,
				     "No certificates found.\n");
			return NULL;
		}
	}

	if (cert_list != NULL) {
		for (node = CERT_LIST_HEAD(cert_list);
		     !CERT_LIST_EMPTY(cert_list) &&
		     !CERT_LIST_END(node, cert_list);
		     node = CERT_LIST_NEXT(node)) {

			cert = node->cert;
			if (cert == NULL) {
				continue;
			}

			if (cert_have_key_for_cert(mcontext, pool,
						   slot, cert, bag,
						   pwcb_args) != PR_TRUE) {
				pkinit_debug(mcontext, 2, "Skipping \"%s\": no "
					     "matching private key loaded.\n",
					     cert->subjectName);
				continue;
			}

			pkinit_debug(mcontext, 2, "Checking \"%s\".\n",
				     cert->subjectName);
			if (cert_is_preferred(mcontext, pool, cert, bag,
					      restricted_ca_list,
					      banned_cert_list,
					      additional_dn_list,
					      kcontext, principal,
					      cert_flags,
					      pwcb_args)) {
				pkinit_debug(mcontext, 2,
					     "Certificate \"%s\" looks okay.\n",
					     cert->subjectName);
				/* CERT_DestroyCertList will free 'cert', so we
				 * need to make a copy of it. */
				ret = CERT_DupCertificate(cert);
				if (!ret) {
					pkinit_debug(mcontext, 2,
						     "%s: failed to allocate "
						     "memory to hold a "
						     "certificate\n",
						     __FUNCTION__);
					break;
				}
				break;
			} else {
				pkinit_debug(mcontext, 2,
					     "Certificate \"%s\" doesn't "
					     "look okay.\n",
					     cert->subjectName);
			}
		}
	}

	CERT_DestroyCertList(cert_list);
	return ret;
}

/* Search, first the preferred slot, then the supplied bag, and then
 * everywhere, for a certificate which matches the named principal. */
CERTCertificate *
cert_find_preferred_cert(struct module_context *mcontext,
			 PLArenaPool *pool,
			 PK11SlotInfo *preferred_slot,
			 struct aa_bag *bag,
			 struct pkinit_pwcb_args *pwcb_args,
			 struct external_principal_identifier **restricted_ca_list,
			 struct external_principal_identifier **banned_cert_list,
			 const char **additional_dn_list,
			 krb5_context kcontext,
			 krb5_principal principal,
			 PRInt32 cert_flags)
{
	CERTCertificate *cert;
	PK11SlotList *slot_list;
	PK11SlotListElement *le;

	if (preferred_slot) {
		cert = cert_find_preferred_cert_using_slot_or_bag(mcontext,
								  pool,
								  preferred_slot,
								  bag,
								  pwcb_args,
								  restricted_ca_list,
								  banned_cert_list,
								  additional_dn_list,
								  kcontext,
								  principal,
								  cert_flags);
		if (cert != NULL) {
			return cert;
		}
	}

	/* Now search the bag. */
	if (bag != NULL) {
		cert = cert_find_preferred_cert_using_slot_or_bag(mcontext,
								  pool,
								  NULL,
								  bag,
								  pwcb_args,
								  restricted_ca_list,
								  banned_cert_list,
								  additional_dn_list,
								  kcontext,
								  principal,
								  cert_flags);
		if (cert != NULL) {
			return cert;
		}
	}

	/* Failing that, search all tokens. */
	slot_list = PK11_GetAllTokens(CKM_INVALID_MECHANISM,
				      PR_FALSE, PR_FALSE, NULL);
	if (slot_list == NULL) {
		pkinit_debug(mcontext, 2,
			     "PK11_GetAllTokens returned no results.\n");
		return NULL;
	}

	/* First pass: only try those slots that are already logged-in */
	for (le = slot_list->head; le != NULL; le = le->next) {
		if (!PK11_IsLoggedIn(le->slot, pwcb_args) &&
		    PK11_NeedLogin(le->slot)) {
			pkinit_debug(mcontext, 2,
				     "No need to log in to %s.\n",
				     PK11_GetSlotName(le->slot));
			continue;
		}

		cert = cert_find_preferred_cert_using_slot_or_bag(mcontext,
								  pool,
								  le->slot, bag,
								  pwcb_args,
								  restricted_ca_list,
								  banned_cert_list,
								  additional_dn_list,
								  kcontext,
								  principal,
								  cert_flags);
		if (cert != NULL) {
			PK11_FreeSlotList(slot_list);
			return cert;
		}
	}

	/* Second pass: check certificates on slots which require login */
	for (le = slot_list->head; le != NULL; le = le->next) {
		if (PK11_IsLoggedIn(le->slot, NULL) ||
		    !PK11_NeedLogin(le->slot)) {
			pkinit_debug(mcontext, 2,
				     "Need to log in to %s.\n",
				     PK11_GetSlotName(le->slot));
			continue;
		}

		cert = cert_find_preferred_cert_using_slot_or_bag(mcontext,
								  pool,
								  le->slot, bag,
								  pwcb_args,
								  restricted_ca_list,
								  banned_cert_list,
								  additional_dn_list,
								  kcontext,
								  principal,
								  cert_flags);
		if (cert != NULL) {
			PK11_FreeSlotList(slot_list);
			return cert;
		}
	}

	PK11_FreeSlotList(slot_list);
	return NULL;
}

/* Try to determine if the certificate is a "CA" certificate. */
static SECStatus
cert_certificate_get_is_ca(struct module_context *mcontext,
			   CERTCertificate *cert, PRBool *is_ca)
{
	CERTBasicConstraints basic_constraints;
	SECItem value;

	if (CERT_FindCertExtension(cert, SEC_OID_X509_BASIC_CONSTRAINTS,
				   &value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No basic constraints value.\n");
		return SECFailure;
	}
	if (CERT_DecodeBasicConstraintValue(&basic_constraints,
					    &value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error decoding basic constraints value.\n");
		return SECFailure;
	}
	*is_ca = basic_constraints.isCA;
	return SECSuccess;
}

/* Check if the raw GUID shows up in the string. */
static PRBool
guid_matches(struct module_context *mcontext,
	     const char *guid, SECItem *raw_guid)
{
	unsigned char bytes[16];
	int i;
	long l;
	const char *p;
	char *q;
	/* We only check properly-formatted GUIDs. */
	if (raw_guid->len != 16) {
		pkinit_debug(mcontext, 2,
			     "GUID claims to have length %d, bogus.\n",
			     raw_guid->len);
		return PR_FALSE;
	}
	/* Now walk the string and extract enough bytes to make a full GUID. */
	p = guid;
	i = 0;
	while (*p != '\0') {
		l = strtol(p, &q, 16);
		if ((q == NULL) || ((q - p) > 2)) {
			/* Malformed. */
			break;
		}
		pkinit_debug(mcontext, 2,
			     " trusted_guid[%d] = 0x%02x\n", i, l & 0xff);
		bytes[i++] = (l & 0xff);
		p = q + strcspn(q, "0123456789ABCDEFabcdef");
		if (i == 16) {
			if (memcmp(bytes, raw_guid->data, 16) == 0) {
				pkinit_debug(mcontext, 2,
					     "The trusted_guid value "
					     "matches.\n");
				return PR_TRUE;
			}
			i = 0;
			if (*p != '\0') {
				pkinit_debug(mcontext, 2,
					     "More data, will keep trying.\n");
			}
		}
	}
	if (i != 0) {
		pkinit_debug(mcontext, 2, "Discarding partially-listed GUID.\n");
	}

	/* Done! */
	return PR_FALSE;
}

/* Pull out the GUID from the specified certificate. */
SECItem **
cert_get_guids(struct module_context *mcontext, CERTCertificate *cert)
{
	struct subject_alt_name **names;
	SECItem san_value, guid;
	SECItem **values;
	int i, j;

	/* Find the subjectAltName extension. */
	if (CERT_FindCertExtension(cert, SEC_OID_X509_SUBJECT_ALT_NAME,
				   &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "No subjectAltName extension in cert while "
			     "checking for GUID.\n");
		return NULL;
	}

	/* Split up the subjectAltName sequence. */
	if (SEC_ASN1DecodeItem(cert->arena, &names,
			       san_template, &san_value) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "Error parsing subjectAltName extension.\n");
		return NULL;
	}

	/* Iterate over all of the values. */
	for (i = 0; (names != NULL) && (names[i] != NULL); i++) {
		continue;
	}
	values = PORT_ArenaZAlloc(cert->arena, sizeof(SECItem *) * (i + 1));
	if (values == NULL) {
		return NULL;
	}
	for (i = 0, j = 0; (names != NULL) && (names[i] != NULL); i++) {
		switch (names[i]->subject_alt_name_type) {
		case subject_alt_name_other_name:
			/* Check for a GUID. */
			if (SECITEM_ItemsAreEqual(&oid_ms_guid.oid,
						  &names[i]->subject_alt_name_value.other_name.oid)) {
				/* Try to decode and store the value. */
				if (SEC_ASN1DecodeItem(cert->arena,
						       &guid,
						       ms_guid_template,
						       &names[i]->subject_alt_name_value.other_name.data) == SECSuccess) {
					values[j++] = SECITEM_ArenaDupItem(cert->arena, &guid);
				}
			}
			break;
		}
	}
	if (j > 0) {
		return values;
	} else {
		return NULL;
	}
}

/* Check if the specified certificate contains a GUID which is listed as one of
 * those which are trusted to hand out TGTs for the server's realm. */
static SECStatus
cert_san_has_guid_for_realm(struct module_context *mcontext,
			    CERTCertificate *kdc,
			    krb5_context kcontext,
			    krb5_principal server,
			    PRBool *is_trusted)
{
	struct subject_alt_name **names;
	SECItem **guids;
	char *trusted_guid;
	int i;

	i = 0;
	names = NULL;

	/* Find out if we have a list of trusted GUIDs. */
	krb5_appdefault_string(kcontext, "pkinit",
			       krb5_princ_realm(kcontext, server),
			       "trusted_guid", "", &trusted_guid);
	if ((trusted_guid == NULL) || (strcmp(trusted_guid, "") == 0)) {
		pkinit_debug(mcontext, 2, "No configured trusted_guid.\n");
		return SECFailure;
	}

	/* Retrieve the GUID values. */
	guids = cert_get_guids(mcontext, kdc);
	for (i = 0; (guids != NULL) && (guids[i] != NULL); i++) {
		/* Compare it to the listed values. */
		if (guid_matches(mcontext, trusted_guid, guids[i])) {
			pkinit_debug(mcontext, 2, "Found matching GUID.\n");
			*is_trusted = PR_TRUE;
		}
	}
	return SECSuccess;
}

/* Determine if the certificate is a valid domain controller certificate for
 * the realm. */
PRBool
cert_is_domain_controller(struct module_context *mcontext,
			  CERTCertificate *signer,
			  krb5_context kcontext,
			  krb5_principal tgt_principal)
{
	PRBool matches;
	if (((cert_template_name_matches(mcontext, signer,
					 "DomainController",
					 &matches) != SECSuccess) ||
	     !matches) &&
	    ((cert_template_name_matches(mcontext, signer,
	   				 "DomainControllerAuthentication",
					 &matches) != SECSuccess) ||
	     !matches)) {
		pkinit_debug(mcontext, 2, "Certificate doesn't belong to a "
			     "domain controller.\n");
		return PR_FALSE;
	}
	if ((cert_san_has_guid_for_realm(mcontext, signer,
					 kcontext,
					 tgt_principal,
					 &matches) != SECSuccess) ||
	    !matches) {
		pkinit_debug(mcontext, 2,
			     "Certificate GUID is not trusted for realm.\n");
		return PR_FALSE;
	}
	return PR_TRUE;
}

/* Check if we should trust a given certificate which was used by a KDC. */
krb5_error_code
cert_validate_kdc_certificate(struct module_context *mcontext,
			      PLArenaPool *pool,
			      CERTCertDBHandle *certdb,
			      struct aa_bag *bag,
			      CERTCertificate *kdc,
			      SECItem **additional_certificates,
			      struct pkinit_pwcb_args *pwcb_args,
			      const char **additional_dn_list,
			      PRInt32 cert_flags,
			      krb5_context kcontext, krb5_principal server)
{
	SECCertificateUsage usages;
	CERTCertificate *issuer, *cert, **other_certs;
	CERTVerifyLog verify_log;
	CERTVerifyLogNode *verify_node;
	SECStatus verify_status;
	PRBool is_ca, is_dc, is_trusted, matches;
	unsigned int i;
	int n_other_certs;

	/* Import the full set of certs, which we expect came from the KDC. */
	for (n_other_certs = 0;
	     (additional_certificates != NULL) &&
	     (additional_certificates[n_other_certs] != NULL);
	     n_other_certs++) {
		continue;
	}
	if (CERT_ImportCerts(certdb, 0,
			     n_other_certs, additional_certificates,
			     &other_certs,
			     PR_FALSE, PR_FALSE,
			     "kdc temporary") != SECSuccess) {
		n_other_certs = 0;
		other_certs = NULL;
	}

	/* Only consider certificates for which we can construct an entire
	 * certificate chain. */
	issuer = CERT_DupCertificate(kdc);
	while (issuer != NULL) {
		cert = issuer;
		issuer = cert_find_cert_issuer(cert, 0,
					       certUsageSSLServer, bag);
		CERT_DestroyCertificate(cert);
		cert = NULL;
		if ((issuer != NULL) && (issuer->isRoot)) {
			break;
		}
	}
	if (issuer == NULL) {
		pkinit_debug(mcontext, 1,
			     "Couldn't construct validation chain for KDC "
			     "certificate (\"%s\" issued by \"%s\").\n",
			     kdc->subjectName, kdc->issuerName);
		CERT_DestroyCertArray(other_certs, n_other_certs);
		return KRB5KDC_ERR_CANT_VERIFY_CERTIFICATE;
	} else {
		CERT_DestroyCertificate(issuer);
		issuer = NULL;
	}

	/* Do the default validity check. */
	memset(&verify_log, 0, sizeof(verify_log));
	verify_log.arena = pool;
	verify_status = CERT_VerifyCertificate(certdb, kdc, PR_TRUE,
					       certificateUsageSSLServer,
					       PR_Now(),
					       pwcb_args,
					       &verify_log, &usages);
	if (verify_status != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "KDC certificate (\"%s\" issued by \"%s\") "
			     "failed simple validation check.\n",
			     kdc->subjectName, kdc->issuerName);
		for (verify_node = verify_log.head;
		     verify_node != NULL;
		     verify_node = verify_node->next) {
			if (verify_node->error != 0) {
				switch (verify_node->error) {
				case SEC_ERROR_EXPIRED_CERTIFICATE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     "certificate expired");
					break;
				case SEC_ERROR_UNKNOWN_SIGNER:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     "unknown signer");
					break;
				case SEC_ERROR_UNTRUSTED_ISSUER:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\" "
						     "(%02x,%02x,%02x): %s\n",
						     verify_node->cert->subjectName,
						     verify_node->cert->trust ?
						     verify_node->cert->trust->sslFlags : 0,
						     verify_node->cert->trust ?
						     verify_node->cert->trust->emailFlags : 0,
						     verify_node->cert->trust ?
						     verify_node->cert->trust->objectSigningFlags : 0,
						     "untrusted issuer");
				case SEC_ERROR_INADEQUATE_KEY_USAGE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     "inadequate key usage");
					break;
				case SEC_ERROR_INADEQUATE_CERT_TYPE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     "inadequate certificate type.");
					break;
				case SEC_ERROR_REVOKED_CERTIFICATE:
				case SEC_ERROR_REVOKED_CERTIFICATE_CRL:
				case SEC_ERROR_REVOKED_CERTIFICATE_OCSP:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     "certificate revoked");
					break;
				default:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     verify_node->cert->subjectName,
						     PR_ErrorToString(verify_node->error,
								      PR_LANGUAGE_I_DEFAULT));
					break;
				}
			}
		}
		CERT_DestroyCertArray(other_certs, n_other_certs);
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}
	CERT_DestroyCertArray(other_certs, n_other_certs);

	/* Check that it's a terminal node. */
	is_ca = PR_TRUE;
	if ((cert_certificate_get_is_ca(mcontext, kdc,
					&is_ca) == SECSuccess) && is_ca) {
		pkinit_debug(mcontext, 1,
			     "KDC certificate shouldn't also be a CA "
			     "certificate.\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Check for an externally-provided mapping. */
	for (i = 0;
	     (cert_flags & CERT_MATCHES_DN_LIST) &&
	     (additional_dn_list != NULL) &&
	     (additional_dn_list[i] != NULL);
	     i++) {
		if ((kdc->subjectName != NULL) &&
		    (strcmp(kdc->subjectName, additional_dn_list[i]) == 0)) {
			pkinit_debug(mcontext, 1,
				     "KDC certificate matched mapping.\n");
			/* Just check that it can be used to sign things -- if
			 * we don't know which convention the certificate was
			 * issued for, then we can't check for any other EKU
			 * values and still expect to work correctly for all
			 * cases. */
			if (CERT_CheckCertUsage(kdc,
						KU_DIGITAL_SIGNATURE) != SECSuccess) {
				/* Oops, it needs to be able to sign stuff. */
				pkinit_debug(mcontext, 1,
					     "KDC certificate can't be used "
					     "for signing.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
		}
		return 0;
	}

	/* If it has the realm's TGS name as its principal name, and the
	 * certificate can be used for server authentication, then that's good
	 * enough for us. */
	if ((cert_flags & CERT_MATCHES_KRB_SAN) &&
	    (cert_san_matches_principal(mcontext, kdc, server,
	   				&matches) == SECSuccess) &&
	    matches) {
		pkinit_debug(mcontext, 2,
			     "KDC certificate has matching PKINIT SAN.\n");
		/* PKINIT name matched. Check if we need to check EKU stuffs. */
		if (cert_eku_matches_oid(mcontext, kdc,
					 &oid_pkinit_key_purpose_kdc.oid,
					 &matches) == SECSuccess) {
			if (!matches) {
				/* Oops, it has EKU, but not the one we want. */
				pkinit_debug(mcontext, 1,
					     "KDC certificate can't be used "
					     "for KDC authentication.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
			if (CERT_CheckCertUsage(kdc,
						KU_DIGITAL_SIGNATURE) != SECSuccess) {
				/* Oops, it needs to be able to sign stuff. */
				pkinit_debug(mcontext, 1,
					     "KDC certificate can't be used "
					     "for signing.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
		}
		return 0;
	}

	/* We've got a Windows-style certificate name? */
	if ((cert_flags & CERT_MATCHES_UPN_SAN) &&
	    (cert_san_matches_upn(mcontext, kdc, kcontext, server,
				  &matches) == SECSuccess) &&
	    matches) {
		if (cert_eku_matches_tag(mcontext, kdc,
					 SEC_OID_EXT_KEY_USAGE_SERVER_AUTH,
					 &matches) == SECSuccess) {
			if (!matches) {
				/* Oops, it has EKU, but not the one we want. */
				pkinit_debug(mcontext, 1,
					     "KDC certificate can't be used "
					     "for SSL server "
					     "authentication.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
			if (CERT_CheckCertUsage(kdc,
						KU_DIGITAL_SIGNATURE) != SECSuccess) {
				/* Oops, it needs to be able to sign stuff. */
				pkinit_debug(mcontext, 1,
					     "KDC certificate can't be used "
					     "for signing.\n");
				return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
			}
		}
		pkinit_debug(mcontext, 2,
			     "KDC certificate has matching UPN SAN.\n");
		return 0;
	}

	/* Try it the new way. */
	if ((cert_san_matches_dns_for_realm(mcontext, kdc,
					    kcontext, server,
					    &matches) == SECSuccess) &&
	    matches == PR_TRUE) {
		if (((cert_eku_matches_tag(mcontext, kdc,
					   SEC_OID_EXT_KEY_USAGE_SERVER_AUTH,
					   &matches) == SECSuccess) &&
		      matches) ||
		    ((cert_eku_matches_oid(mcontext, kdc,
					   &oid_pkinit_key_purpose_kdc.oid,
					   &matches) == SECSuccess) &&
		      matches)) {
			pkinit_debug(mcontext, 2,
				     "KDC certificate has DNS name which is "
				     "trusted for realm, and has correct EKU "
				     "value.\n");
			return 0;
		}
	}

	/* Otherwise do the best we can to see if it's a valid Windows
	 * domain controller certificate.  First, check if the
	 * certificateTemplateName has a good value. */
	is_dc = PR_FALSE;
	if (((cert_template_name_matches(mcontext, kdc, "DomainController",
					 &is_dc) != SECSuccess) ||
	     !is_dc) &&
	    ((cert_template_name_matches(mcontext, kdc,
	   				 "DomainControllerAuthentication",
					 &is_dc) != SECSuccess) ||
	     !is_dc)) {
		pkinit_debug(mcontext, 1, "KDC certificate doesn't belong to a "
			     "domain controller.\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Check that it has a GUID, and that the GUID is in the list
	 * which we trust to sign TGTs for the given realm. */
	is_trusted = PR_FALSE;
	if (cert_san_has_guid_for_realm(mcontext, kdc, kcontext, server,
					&is_trusted) != SECSuccess) {
		pkinit_debug(mcontext, 2,
			     "KDC certificate has no GUID extension.\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}
	if (!is_trusted) {
		pkinit_debug(mcontext, 1,
			     "KDC certificate GUID is not trusted by "
			     "client configuration.\n");
		return KRB5KDC_ERR_KDC_NOT_TRUSTED;
	}

	/* Done. */
	pkinit_debug(mcontext, 2, "KDC certificate successfully validated.\n");
	return 0;
}

krb5_error_code
cert_validate_client_certificate(struct module_context *mcontext,
				 PLArenaPool *pool,
				 CERTCertDBHandle *certdb,
				 struct aa_bag *bag,
				 CERTCertificate *client_cert,
				 SECItem **additional_certificates,
			         struct pkinit_pwcb_args *pwcb_args,
				 const char **additional_dn_list,
				 PRInt32 cert_flags,
				 krb5_context kcontext, krb5_principal client,
				 struct external_principal_identifier ***error_epis)
{
	SECCertificateUsage usages;
	CERTCertificate *issuer, *cert, **other_certs;
	CERTVerifyLog verify_log;
	CERTVerifyLogNode *vnode;
	struct external_principal_identifier **epi;
	SECStatus verify_status;
	int n_other_certs;
	PRBool is_ca, matches;

	*error_epis = NULL;

	/* Import the full set of certs, which we expect came from a client. */
	for (n_other_certs = 0;
	     (additional_certificates != NULL) &&
	     (additional_certificates[n_other_certs] != NULL);
	     n_other_certs++) {
		continue;
	}
	if (CERT_ImportCerts(certdb, 0,
			     n_other_certs, additional_certificates,
			     &other_certs,
			     PR_FALSE, PR_FALSE,
			     "client temporary") != SECSuccess) {
		n_other_certs = 0;
		other_certs = NULL;
	}

	/* Only consider certificates for which we can construct an entire
	 * certificate chain. */
	issuer = CERT_DupCertificate(client_cert);
	while (issuer != NULL) {
		cert = issuer;
		issuer = cert_find_cert_issuer(cert, 0,
					       certUsageSSLServer, bag);
		CERT_DestroyCertificate(cert);
		cert = NULL;
		if ((issuer != NULL) && (issuer->isRoot)) {
			break;
		}
	}
	if (issuer == NULL) {
		pkinit_debug(mcontext, 1,
			     "Couldn't construct validation chain for client "
			     "certificate (\"%s\" issued by \"%s\").\n",
			     client_cert->subjectName, client_cert->issuerName);
		CERT_DestroyCertArray(other_certs, n_other_certs);
		return KRB5KDC_ERR_CANT_VERIFY_CERTIFICATE;
	} else {
		CERT_DestroyCertificate(issuer);
		issuer = NULL;
	}

	/* Do the default validity check. */
	memset(&verify_log, 0, sizeof(verify_log));
	verify_log.arena = pool;
	verify_status = CERT_VerifyCertificate(certdb, client_cert, PR_TRUE,
					       certificateUsageSSLClient,
					       PR_Now(),
					       pwcb_args,
					       &verify_log, &usages);
	if (verify_status != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Client certificate (\"%s\" issued by \"%s\") "
			     "failed simple validation check.\n",
			     client_cert->subjectName, client_cert->issuerName);
		for (vnode = verify_log.head;
		     vnode != NULL;
		     vnode = vnode->next) {
			if (vnode->error != 0) {
				switch (vnode->error) {
				case SEC_ERROR_EXPIRED_CERTIFICATE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     "certificate expired");
					break;
				case SEC_ERROR_UNKNOWN_SIGNER:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     "unknown signer");
					break;
				case SEC_ERROR_UNTRUSTED_ISSUER:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\" "
						     "(%02x,%02x,%02x): %s\n",
						     vnode->cert->subjectName,
						     vnode->cert->trust ?
						     vnode->cert->trust->sslFlags : 0,
						     vnode->cert->trust ?
						     vnode->cert->trust->emailFlags : 0,
						     vnode->cert->trust ?
						     vnode->cert->trust->objectSigningFlags : 0,
						     "untrusted issuer");
					break;
				case SEC_ERROR_INADEQUATE_KEY_USAGE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     "inadequate key usage");
					break;
				case SEC_ERROR_INADEQUATE_CERT_TYPE:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     "inadequate certificate "
						     "type");
					break;
				case SEC_ERROR_REVOKED_CERTIFICATE:
				case SEC_ERROR_REVOKED_CERTIFICATE_CRL:
				case SEC_ERROR_REVOKED_CERTIFICATE_OCSP:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     "certificate revoked");
					break;
				default:
					pkinit_debug(mcontext, 1,
						     "Error at \"%s\": %s\n",
						     vnode->cert->subjectName,
						     PR_ErrorToString(vnode->error,
								      PR_LANGUAGE_I_DEFAULT));
					break;
				}
			}
		}
		CERT_DestroyCertArray(other_certs, n_other_certs);
		switch (verify_status) {
		case SEC_ERROR_BAD_SIGNATURE:
			/* encode the invalid certificate list */
			epi = NULL;
			for (vnode = verify_log.head;
			     vnode != NULL;
			     vnode = vnode->next) {
				switch (vnode->error) {
				case SEC_ERROR_BAD_SIGNATURE:
					epi = bcms_make_external_principal_identifier_list(mcontext, pool, epi, vnode->cert);
					break;
				default:
					break;
				}
			}
			*error_epis = epi;
			return KRB5KDC_ERR_INVALID_CERTIFICATE;
			break;
		case SEC_ERROR_OCSP_UNKNOWN_RESPONSE_TYPE:
		case SEC_ERROR_OCSP_BAD_HTTP_RESPONSE:
		case SEC_ERROR_OCSP_MALFORMED_REQUEST:
		case SEC_ERROR_OCSP_SERVER_ERROR:
		case SEC_ERROR_OCSP_TRY_SERVER_LATER:             
		case SEC_ERROR_OCSP_REQUEST_NEEDS_SIG:            
		case SEC_ERROR_OCSP_UNAUTHORIZED_REQUEST:
		case SEC_ERROR_OCSP_UNKNOWN_RESPONSE_STATUS:
		case SEC_ERROR_OCSP_UNKNOWN_CERT:
		case SEC_ERROR_OCSP_NOT_ENABLED:
		case SEC_ERROR_OCSP_NO_DEFAULT_RESPONDER:
		case SEC_ERROR_OCSP_MALFORMED_RESPONSE:
		case SEC_ERROR_OCSP_UNAUTHORIZED_RESPONSE:
		case SEC_ERROR_OCSP_FUTURE_RESPONSE:
		case SEC_ERROR_OCSP_OLD_RESPONSE:
		case SEC_ERROR_OCSP_INVALID_SIGNING_CERT:
			/* encode the invalid certificate list */
			epi = NULL;
			for (vnode = verify_log.head;
			     vnode != NULL;
			     vnode = vnode->next) {
				switch (vnode->error) {
				case SEC_ERROR_OCSP_UNKNOWN_RESPONSE_TYPE:
				case SEC_ERROR_OCSP_BAD_HTTP_RESPONSE:
				case SEC_ERROR_OCSP_MALFORMED_REQUEST:
				case SEC_ERROR_OCSP_SERVER_ERROR:
				case SEC_ERROR_OCSP_TRY_SERVER_LATER:             
				case SEC_ERROR_OCSP_REQUEST_NEEDS_SIG:            
				case SEC_ERROR_OCSP_UNAUTHORIZED_REQUEST:
				case SEC_ERROR_OCSP_UNKNOWN_RESPONSE_STATUS:
				case SEC_ERROR_OCSP_UNKNOWN_CERT:
				case SEC_ERROR_OCSP_NOT_ENABLED:
				case SEC_ERROR_OCSP_NO_DEFAULT_RESPONDER:
				case SEC_ERROR_OCSP_MALFORMED_RESPONSE:
				case SEC_ERROR_OCSP_UNAUTHORIZED_RESPONSE:
				case SEC_ERROR_OCSP_FUTURE_RESPONSE:
				case SEC_ERROR_OCSP_OLD_RESPONSE:
				case SEC_ERROR_OCSP_INVALID_SIGNING_CERT:
					epi = bcms_make_external_principal_identifier_list(mcontext, pool, epi, vnode->cert);
					break;
				default:
					break;
				}
			}
			*error_epis = epi;
			return KRB5KDC_ERR_REVOCATION_STATUS_UNKNOWN;
			break;
		case SEC_ERROR_REVOKED_CERTIFICATE:
		case SEC_ERROR_REVOKED_CERTIFICATE_CRL:
		case SEC_ERROR_REVOKED_CERTIFICATE_OCSP:
			/* encode the invalid certificate list */
			epi = NULL;
			for (vnode = verify_log.head;
			     vnode != NULL;
			     vnode = vnode->next) {
				switch (vnode->error) {
				case SEC_ERROR_REVOKED_CERTIFICATE:
				case SEC_ERROR_REVOKED_CERTIFICATE_CRL:
				case SEC_ERROR_REVOKED_CERTIFICATE_OCSP:
					epi = bcms_make_external_principal_identifier_list(mcontext, pool, epi, vnode->cert);
					break;
				default:
					break;
				}
			}
			*error_epis = epi;
			return KRB5KDC_ERR_REVOKED_CERTIFICATE;
			break;
		case SEC_ERROR_EXPIRED_CERTIFICATE:
		case SEC_ERROR_UNKNOWN_SIGNER:
		case SEC_ERROR_UNTRUSTED_ISSUER:
		case SEC_ERROR_UNTRUSTED_CERT:
		default:
			return KRB5KDC_ERR_CLIENT_NOT_TRUSTED;
		}
	}
	CERT_DestroyCertArray(other_certs, n_other_certs);

	/* Check that it's a terminal node. */
	if ((cert_certificate_get_is_ca(mcontext, client_cert,
					&is_ca) == SECSuccess) &&
	    is_ca) {
		pkinit_debug(mcontext, 1,
			     "Client certificate shouldn't also be a CA "
			     "certificate.\n");
		return KRB5KDC_ERR_CLIENT_NOT_TRUSTED;
	}

	/* Check that it matches the principal name. */
	if (!cert_is_preferred(mcontext, pool, client_cert, bag,
			       NULL, NULL, additional_dn_list,
			       kcontext, client, cert_flags, pwcb_args)) {
		pkinit_debug(mcontext, 1,
			     "Client certificate for \"%s\" didn't match "
			     "client name.\n", client_cert->subjectName);
		return KRB5KDC_ERR_CLIENT_NAME_MISMATCH;
	}

	/* Check that it matches the desired EKU. */
	if ((cert_eku_matches_oid(mcontext, client_cert,
				  &oid_pkinit_key_purpose_client.oid,
				  &matches) == SECSuccess) &&
	    !matches &&
	    (cert_eku_matches_oid(mcontext, client_cert,
				  &oid_ms_sc_login.oid,
				  &matches) == SECSuccess) &&
	    !matches) {
		/* Oops, it has EKU, but not the one we want. */
		pkinit_debug(mcontext, 1, "Client certificate can't be used "
			     "for PKINIT authentication.\n");
		return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
	}
	if (CERT_CheckCertUsage(client_cert,
				KU_DIGITAL_SIGNATURE) != SECSuccess) {
		/* Oops, it needs to be able to sign stuff. */
		pkinit_debug(mcontext, 1,
			     "Client certificate can't be used for signing.\n");
		return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
	}

	pkinit_debug(mcontext, 2, "Client certificate has matching SAN.\n");
	return 0;
}

krb5_error_code
cert_verify_cert_for_encryption(struct module_context *mcontext,
				CERTCertDBHandle *certdb,
				struct aa_bag *unused_bag,
				SECItem *client)
{
	CERTCertificate **client_certs;

	/* Verify that the client's certificate is allowed to be used for
	 * encryption. */
	client_certs = NULL;
	if ((CERT_ImportCerts(certdb, 0,
			      1, &client,
			      &client_certs,
			      PR_FALSE, PR_FALSE,
			      "pkinit client") != SECSuccess) ||
	    (client_certs == NULL) ||
	    (client_certs[0] == NULL)) {
		pkinit_debug(mcontext, 1,
			     "Error parsing client certificate.\n");
		return KRB5KRB_ERR_GENERIC;
	}
	if (CERT_CheckCertUsage(client_certs[0],
				KU_KEY_ENCIPHERMENT) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Client key is not intended for encryption.\n");
		CERT_DestroyCertArray(client_certs, 1);
		return KRB5KDC_ERR_INCONSISTENT_KEY_PURPOSE;
	}
	CERT_DestroyCertArray(client_certs, 1);
	return 0;
}

/* Locate the key which corresponds to the provided certificate's public key. */
SECKEYPrivateKey *
cert_find_key_for_cert(struct module_context *mcontext,
		       PK11SlotInfo *preferred_slot,
		       CERTCertificate *cert,
		       struct aa_bag *bag,
		       struct pkinit_pwcb_args *pwcb_args)
{
	SECKEYPrivateKeyList *klist;
	SECKEYPrivateKeyListNode *klistnode;
	SECKEYPrivateKey *privkey;
	SECKEYPublicKey *pubkey;
	SECItem *encoded_pubkey, *candidate;

	/* First, search the preferred slot. */
	privkey = NULL;
	if (preferred_slot != NULL) {
		privkey = PK11_FindPrivateKeyFromCert(preferred_slot, cert,
						      pwcb_args);
	}
	/* Then search all slots. */
	if (privkey == NULL) {
		privkey = PK11_FindKeyByAnyCert(cert, pwcb_args);
	}
	/* Look in the bag. */
	if (privkey == NULL) {
		pubkey = NULL;
		encoded_pubkey = NULL;
		klist = aa_bag_list_loaded_keys(bag);
		if (klist != NULL) {
			if (encoded_pubkey == NULL) {
				pubkey = CERT_ExtractPublicKey(cert);
				encoded_pubkey = PK11_DEREncodePublicKey(pubkey);
				SECKEY_DestroyPublicKey(pubkey);
			}
			for (klistnode = PRIVKEY_LIST_HEAD(klist);
			     !PRIVKEY_LIST_END(klistnode, klist);
			     klistnode = PRIVKEY_LIST_NEXT(klistnode)) {
				pkinit_debug(mcontext, 3, "Found a privkey.\n");
				pubkey = SECKEY_ConvertToPublicKey(klistnode->key);
				candidate = PK11_DEREncodePublicKey(pubkey);
				SECKEY_DestroyPublicKey(pubkey);
				if (SECITEM_ItemsAreEqual(candidate, encoded_pubkey)) {
					pkinit_debug(mcontext, 3, "Found a MATCHING privkey.\n");
					privkey = klistnode->key;
				}
				if (privkey != NULL) {
					break;
				}
			}
		}
	}
	return privkey;
}
