/*
 * Copyright (C) 2006,2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: commont.c,v 1.19 2007/05/22 21:20:42 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <krb5/krb5.h>

#include <cert.h>
#include <certdb.h>
#include <certt.h>

#include "certs.h"
#include "commont.h"

static void
derdump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static void
dump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("od -t x1c", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}

/* GeneralString template.  Don't know why NSS doesn't provide a
 * SEC_GeneralStringTemplate, when providing one for each primitive type
 * seems to have been the intent. */
const SEC_ASN1Template
common_general_string_template[] = {
	{
	.kind = SEC_ASN1_GENERAL_STRING,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(SECItem),
	},
};

/* AlgorithmIdentifier: RFC 2459, 4.1.1.2. */
const SEC_ASN1Template
common_algorithm_identifier_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_OBJECT_ID,
	.offset = offsetof(struct algorithm_identifier, algorithm),
	.sub = &SEC_ObjectIDTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_OPTIONAL |
		SEC_ASN1_POINTER |
		SEC_ASN1_ANY,
	.offset = offsetof(struct algorithm_identifier, parameters),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem*),
	},
	{0, 0, NULL, 0},
};

const SEC_ASN1Template
common_sequence_of_algorithm_identifier_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = common_algorithm_identifier_template,
	.size = 0,
	},
};

/* ValidationParms: RFC 2459, 7.3.2. */
const SEC_ASN1Template
common_validation_parms_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct validation_parms),
	},
	{
	.kind = SEC_ASN1_BIT_STRING,
	.offset = offsetof(struct validation_parms, seed),
	.sub = &SEC_BitStringTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct validation_parms, pgen_counter),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0},
};

/* DH parameters: draft-ietf-cat-kerberos-pk-init-08.txt, 3.1.2.2. */
const SEC_ASN1Template
common_dh_parameters_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct dh_parameters),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct dh_parameters, p),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct dh_parameters, g),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct dh_parameters, private_value_length),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0},
};

/* DomainParameters: RFC 2459, 7.3.2. */
const SEC_ASN1Template
common_domain_parameters_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct domain_parameters),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct domain_parameters, p),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct domain_parameters, g),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct domain_parameters, q),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER | SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct domain_parameters, j),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INLINE |
		SEC_ASN1_POINTER |
		SEC_ASN1_OPTIONAL,
	.offset = offsetof(struct domain_parameters, validation_parms),
	.sub = &common_validation_parms_template,
	.size = sizeof(struct validation_parms *),
	},
	{ 0, 0, NULL, 0},
};

/* IssuerAndSerialNumber: RFC 3852, 10.2.4. */
const SEC_ASN1Template
common_issuer_and_serial_number_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct issuer_and_serial_number),
	},
	{
	.kind = SEC_ASN1_ANY,
	.offset = offsetof(struct issuer_and_serial_number, issuer),
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct issuer_and_serial_number, serial),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{ 0, 0, NULL, 0},
};

/* SubjectPublicKeyInfo: RFC 2459, appendix A.1. */
const SEC_ASN1Template
common_subject_public_key_info_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct subject_public_key_info),
	},
	{
	.kind = SEC_ASN1_INLINE,
	.offset = offsetof(struct subject_public_key_info, algorithm),
	.sub = common_algorithm_identifier_template,
	.size = sizeof(struct algorithm_identifier),
	},
	{
	.kind = SEC_ASN1_BIT_STRING,
	.offset = offsetof(struct subject_public_key_info, subject_public_key),
	.sub = &SEC_BitStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* RC2CBCParameter: RFC 2630, 12.4.2. */
const SEC_ASN1Template
common_rc2_cbc_parameters_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct rc2_cbc_parameters),
	},
	{
	.kind = SEC_ASN1_INTEGER,
	.offset = offsetof(struct rc2_cbc_parameters, version),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct rc2_cbc_parameters, iv),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
	{0, 0, NULL, 0},
};

/* CBCParameter == IV: RFC 2630, 12.4.1. */
const SEC_ASN1Template
common_des3_cbc_parameters_template[] = {
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct des3_cbc_parameters, iv),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
};

/* CBCParameter == AES-IV: RFC 3565, 4.1. */
const SEC_ASN1Template
common_aes_cbc_parameters_template[] = {
	{
	.kind = SEC_ASN1_OCTET_STRING,
	.offset = offsetof(struct aes_cbc_parameters, iv),
	.sub = &SEC_OctetStringTemplate,
	.size = sizeof(SECItem),
	},
};

const SEC_ASN1Template
common_sequence_of_any_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = &SEC_AnyTemplate,
	.size = sizeof(SECItem **),
	},
};

/* Create a SECItem to hold the contents of the krb5_data item. */
SECItem *
common_secitem_from_krb5_data(struct module_context *mcontext,
			      PLArenaPool *pool, krb5_data *data)
{
	SECItem *ret;
	if (data != NULL) {
		ret = SECITEM_AllocItem(pool, NULL, data->length);
		if (ret == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory.\n");
			return ret;
		}
		memcpy(ret->data, data->data, data->length);
		return ret;
	}
	return NULL;
}

krb5_data *
common_krb5_data_from_secitem(struct module_context *mcontext,
			      krb5_context kcontext, SECItem *item)
{
	krb5_data datum, *ret;
	if (item != NULL) {
		datum.data = (char *) item->data;
		datum.length = item->len;
		ret = NULL;
		if (krb5_copy_data(kcontext, &datum, &ret) != 0) {
			pkinit_debug(mcontext, 0, "Out of memory.\n");
			return NULL;
		}
		return ret;
	}
	return NULL;
}

SECItem *
common_encode_issuer_and_serial_number(struct module_context *mcontext,
				       PLArenaPool *pool,
				       struct issuer_and_serial_number *isn)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, isn,
			       common_issuer_and_serial_number_template) ==
	    &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding issuer and serial number.\n");
	return NULL;
}

SECItem *
common_encode_dh_parameters(struct module_context *mcontext,
			    PLArenaPool *pool, struct dh_parameters *params)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, params,
			       common_dh_parameters_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding DH parameters.\n");
	return NULL;
}

SECItem *
common_encode_domain_parameters(struct module_context *mcontext,
				PLArenaPool *pool,
				struct domain_parameters *params)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, params,
			       common_domain_parameters_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding domain parameters.\n");
	return NULL;
}

SECItem *
common_encode_subject_public_key_info(struct module_context *mcontext,
				      PLArenaPool *pool,
				      struct subject_public_key_info *info)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, info,
			       common_subject_public_key_info_template) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding subject public key info.\n");
	return NULL;
}

SECItem *
common_encode_bitstring(struct module_context *mcontext,
			PLArenaPool *pool, SECItem *bits)
{
	SECItem encoded, bytes;
	bytes = *bits;
	if (SEC_ASN1EncodeItem(pool, &encoded, &bytes,
			       SEC_BitStringTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding bit string.\n");
	return NULL;
}

SECItem *
common_encode_unsigned_integer(struct module_context *mcontext,
			       PLArenaPool *pool, SECItem *bits)
{
	SECItem encoded, *unsigned_raw;
	if (bits->data[0] & 0x80) {
		/* We want an unsigned number, but the high bit is set, so if
		 * we just treat it as an integer, it'll appear to be signed
		 * and negative. */
		unsigned_raw = SECITEM_AllocItem(pool, NULL, bits->len + 1);
		if (unsigned_raw == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory.\n");
		}
		unsigned_raw->data[0] = 0;
		memcpy(unsigned_raw->data + 1, bits->data, bits->len);
	} else {
		unsigned_raw = bits;
	}
	if (SEC_ASN1EncodeItem(pool, &encoded, unsigned_raw,
			       SEC_IntegerTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding integer.\n");
	return NULL;
}

SECItem *
common_encode_octetstring(struct module_context *mcontext,
			  PLArenaPool *pool, SECItem *bits)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, bits,
			       SEC_OctetStringTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding octet string.\n");
	return NULL;
}

SECItem *
common_encode_objectid(struct module_context *mcontext,
		       PLArenaPool *pool, SECItem *oid)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, oid,
			       SEC_ObjectIDTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding object identifier.\n");
	return NULL;
}

SECItem *
common_encode_utctime(struct module_context *mcontext,
		      PLArenaPool *pool, SECItem *time)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, time,
			       SEC_UTCTimeTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding UTC time.\n");
	return NULL;
}

SECItem *
common_encode_generalizedtime(struct module_context *mcontext,
			      PLArenaPool *pool, SECItem *time)
{
	SECItem encoded;
	if (SEC_ASN1EncodeItem(pool, &encoded, time,
			       SEC_GeneralizedTimeTemplate) == &encoded) {
		return SECITEM_ArenaDupItem(pool, &encoded);
	}
	pkinit_debug(mcontext, 0, "Error encoding general time.\n");
	return NULL;
}

struct issuer_and_serial_number *
common_decode_issuer_and_serial_number(struct module_context *mcontext,
				       PLArenaPool *pool, SECItem *item)
{
	struct issuer_and_serial_number *isn;

	isn = PORT_ArenaZAlloc(pool, sizeof(*isn));
	if (isn == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, isn,
			       common_issuer_and_serial_number_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding issuer and serial number.\n");
		return NULL;
	}
	return isn;
}

struct dh_parameters *
common_decode_dh_parameters(struct module_context *mcontext,
			    PLArenaPool *pool, SECItem *item)
{
	struct dh_parameters *dh;

	dh = PORT_ArenaZAlloc(pool, sizeof(*dh));
	if (dh == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, dh,
			       common_dh_parameters_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding DH parameters.\n");
		return NULL;
	}
	return dh;
}

struct domain_parameters *
common_decode_domain_parameters(struct module_context *mcontext,
				PLArenaPool *pool, SECItem *item)
{
	struct domain_parameters *dp;

	dp = PORT_ArenaZAlloc(pool, sizeof(*dp));
	if (dp == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, dp,
			       common_domain_parameters_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding domain-parameters.\n");
		return NULL;
	}
	return dp;
}

struct algorithm_identifier **
common_decode_sequence_of_algorithm_identifier(struct module_context *mcontext,
					       PLArenaPool *pool, SECItem *item)
{
	struct algorithm_identifier **ai;

	ai = NULL;
	if (SEC_ASN1DecodeItem(pool, &ai,
			       common_sequence_of_algorithm_identifier_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding sequence-of-algorithm-identifier.\n");
		return NULL;
	}
	return ai;
}

struct subject_public_key_info *
common_decode_subject_public_key_info(struct module_context *mcontext,
				      PLArenaPool *pool, SECItem *item)
{
	struct subject_public_key_info *spki;

	spki = PORT_ArenaZAlloc(pool, sizeof(*spki));
	if (spki == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, spki,
			       common_subject_public_key_info_template,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding subject-public-key-info.\n");
		return NULL;
	}
	return spki;
}

SECItem *
common_decode_bitstring(struct module_context *mcontext,
			PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_BitStringTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding bit string.\n");
		return NULL;
	}
	return ret;
}

SECItem *
common_decode_integer(struct module_context *mcontext,
		      PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_IntegerTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding integer.\n");
		return NULL;
	}
	return ret;
}

SECItem *
common_decode_octetstring(struct module_context *mcontext,
			  PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_OctetStringTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding octet string identifier.\n");
		return NULL;
	}
	return ret;
}

SECItem *
common_decode_objectid(struct module_context *mcontext,
		       PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_ObjectIDTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1,
			     "Error decoding object identifier.\n");
		return NULL;
	}
	return ret;
}

SECItem *
common_decode_utctime(struct module_context *mcontext,
		      PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_UTCTimeTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding UTC time.\n");
		return NULL;
	}
	return ret;
}

SECItem *
common_decode_generalizedtime(struct module_context *mcontext,
			      PLArenaPool *pool, SECItem *item)
{
	SECItem *ret;

	ret = PORT_ArenaZAlloc(pool, sizeof(SECItem));
	if (ret == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	if (SEC_ASN1DecodeItem(pool, ret,
			       SEC_GeneralizedTimeTemplate,
			       item) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Error decoding general time.\n");
		return NULL;
	}
	return ret;
}

/* Make a copy of the list of identifiers, ensuring that the passed-in value is
 * in there somewhere. */
struct algorithm_identifier **
common_make_algorithm_identifier_list(struct module_context *mcontext,
				      PLArenaPool *pool,
				      struct algorithm_identifier **list,
				      struct algorithm_identifier *entry)
{
	int i;
	PRBool found;
	SECItem *nalgorithm;
	struct algorithm_identifier **nlist, *nentry;

	found = PR_FALSE;
	for (i = 0; (list != NULL) && (list[i] != NULL) && !found; i++) {
		if (SECITEM_ItemsAreEqual(&list[i]->algorithm,
					  &entry->algorithm) &&
		    SECITEM_ItemsAreEqual(list[i]->parameters,
					  entry->parameters)) {
			found = PR_TRUE;
			break;
		}
	}
	if (!found) {
		/* Duplicate the entire entry. */
		nentry = PORT_ArenaZAlloc(pool, sizeof(*nentry));
		if (nentry == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		nalgorithm = SECITEM_ArenaDupItem(pool, &entry->algorithm);
		if (nalgorithm == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		nentry->algorithm = *nalgorithm;
		if (entry->parameters != NULL) {
			nentry->parameters =
				SECITEM_ArenaDupItem(pool, entry->parameters);
			if (nentry->parameters == NULL) {
				pkinit_debug(mcontext, 0, "Out of memory\n");
				return NULL;
			}
		}
		/* i is the length of the list we've already got */
		nlist = PORT_ArenaZAlloc(pool,
					 (i + 2) *
					 sizeof(struct algorithm_identifier*));
		if (nlist == NULL) {
			pkinit_debug(mcontext, 0, "Out of memory\n");
			return NULL;
		}
		/* Add the new algorithm to the end of the list. */
		nlist[i] = nentry;
		/* Copy the rest of the list. */
		while (--i >= 0) {
			nlist[i] = list[i];
		}
	} else {
		/* Save the existing list so that we can assign it later. */
		nlist = list;
	}
	return nlist;
}

/* Generate a new content encryption key for an enveloped-data structure, and
 * store information about the keys and parameters. */
static SECStatus
common_generate_content_encryption_key_rc2(struct module_context *mcontext,
					   PLArenaPool *pool,
					   krb5_context kcontext,
					   struct algorithm_identifier *alg,
					   SECItem *content_enc_key,
					   CK_MECHANISM *mech,
					   int *block_size)
{
	SECOidData *oid;
	SECItem encoded;
	krb5_data octets;
	struct rc2_cbc_parameters params;
	CK_RC2_CBC_PARAMS *pk_params;

	/* Generate a key.  For now at least, just assume 128-bit RC2.  It can
	 * use any string of bytes as a key without additional preparation on
	 * our part. */
	content_enc_key->len = 16;
	content_enc_key->data = PORT_ArenaZAlloc(pool, content_enc_key->len);
	if (content_enc_key->data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = content_enc_key->len;
	octets.data = (char *) content_enc_key->data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Set up some parameters: first, the OID of the algorithm. */
	oid = SECOID_FindOIDByTag(SEC_OID_RC2_CBC);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error: no knowledge of the "
			     "RC2-CBC encryption algorithm.\n");
		return SECFailure;
	}
	alg->algorithm = oid->oid;

	/* Now the "version", which indicates the key length.  0x3a corresponds
	 * to 128 bits. */
	if (SEC_ASN1EncodeInteger(pool, &params.version, 0x3a) !=
	    &params.version) {
		pkinit_debug(mcontext, 0, "Internal error encoding version.\n");
		return SECFailure;
	}

	/* Generate an IV. */
	params.iv.len = 8;
	params.iv.data = PORT_ArenaZAlloc(pool, params.iv.len);
	if (params.iv.data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = params.iv.len;
	octets.data = (char *) params.iv.data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Encode the parameters structure. */
	if (SEC_ASN1EncodeItem(pool, &encoded, &params,
			       common_rc2_cbc_parameters_template) != &encoded) {
		pkinit_debug(mcontext, 0, "Error encoding parameters.\n");
		return SECFailure;
	}
	alg->parameters = SECITEM_ArenaDupItem(pool, &encoded);
	if (alg->parameters == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}

	/* Build the parameter block for the PKCS11 layer. */
	mech->mechanism = CKM_RC2_CBC;
	pk_params = PORT_ArenaZAlloc(pool, sizeof(*pk_params));
	if (pk_params == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	pk_params->ulEffectiveBits = 128;
	memcpy(&pk_params->iv[0], params.iv.data, 8);
	mech->pParameter = pk_params;
	mech->ulParameterLen = sizeof(*pk_params);

	/* The consumer of the key is going to need to know how to pad the
	 * data for encryption. */
	*block_size = 8;

	return SECSuccess;
}

/* Generate a new content encryption key for an enveloped-data structure, and
 * store information about the keys and parameters. */
static SECStatus
common_generate_content_encryption_key_des3(struct module_context *mcontext,
					    PLArenaPool *pool,
					    krb5_context kcontext,
					    struct algorithm_identifier *alg,
					    SECItem *content_enc_key,
					    CK_MECHANISM *mech,
					    int *block_size)
{
	SECOidData *oid;
	SECItem encoded;
	krb5_data octets;
	struct des3_cbc_parameters params;
	char *pk_params;
	unsigned int i, parity;

	/* Generate a key. */
	content_enc_key->len = 24;
	content_enc_key->data = PORT_ArenaZAlloc(pool, content_enc_key->len);
	if (content_enc_key->data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = content_enc_key->len;
	octets.data = (char *) content_enc_key->data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Fix the parity in the key bits -- odd parity, low bit. */
	for (i = 0; i < content_enc_key->len; i++) {
		parity = ((content_enc_key->data[i] & 0x02) >> 1) ^
			 ((content_enc_key->data[i] & 0x04) >> 2) ^
			 ((content_enc_key->data[i] & 0x08) >> 3) ^
			 ((content_enc_key->data[i] & 0x10) >> 4) ^
			 ((content_enc_key->data[i] & 0x20) >> 5) ^
			 ((content_enc_key->data[i] & 0x40) >> 6) ^
			 ((content_enc_key->data[i] & 0x80) >> 7) ^ 1;
		content_enc_key->data[i] &= 0xfe;
		content_enc_key->data[i] |= parity;
	}

	/* Set up some parameters: first, the OID of the algorithm. */
	oid = SECOID_FindOIDByTag(SEC_OID_DES_EDE3_CBC);
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error: no knowledge of the "
			     "DES-EDE3-CBC encryption algorithm.\n");
		return SECFailure;
	}
	alg->algorithm = oid->oid;

	/* Generate an IV. */
	params.iv.len = 8;
	params.iv.data = PORT_ArenaZAlloc(pool, params.iv.len);
	if (params.iv.data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = params.iv.len;
	octets.data = (char *) params.iv.data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Encode the parameters structure. */
	if (SEC_ASN1EncodeItem(pool, &encoded, &params,
			       common_des3_cbc_parameters_template) != &encoded) {
		pkinit_debug(mcontext, 0, "Error encoding parameters.\n");
		return SECFailure;
	}
	alg->parameters = SECITEM_ArenaDupItem(pool, &encoded);
	if (alg->parameters == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}

	/* Build the parameter block for the PKCS11 layer. */
	mech->mechanism = CKM_DES3_CBC;
	pk_params = PORT_ArenaZAlloc(pool, 8);
	if (pk_params == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	memcpy(&pk_params[0], params.iv.data, 8);
	mech->pParameter = pk_params;
	mech->ulParameterLen = 8;

	/* The consumer of the key is going to need to know how to pad the
	 * data for encryption. */
	*block_size = 8;

	return SECSuccess;
}

static SECStatus
common_generate_content_encryption_key_aes(struct module_context *mcontext,
					   PLArenaPool *pool,
					   int key_size_bits,
					   krb5_context kcontext,
					   struct algorithm_identifier *alg,
					   SECItem *content_enc_key,
					   CK_MECHANISM *mech,
					   int *block_size)
{
	SECOidData *oid;
	SECItem encoded;
	krb5_data octets;
	struct aes_cbc_parameters params;
	char *pk_params;

	/* Generate a key. */
	content_enc_key->len = key_size_bits / 8;
	content_enc_key->data = PORT_ArenaZAlloc(pool, content_enc_key->len);
	if (content_enc_key->data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = content_enc_key->len;
	octets.data = (char *) content_enc_key->data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Set up some parameters: first, the OID of the algorithm. */
	switch (key_size_bits) {
	case 128:
		oid = SECOID_FindOIDByTag(SEC_OID_AES_128_CBC);
		break;
	case 192:
		oid = SECOID_FindOIDByTag(SEC_OID_AES_192_CBC);
		break;
	case 256:
		oid = SECOID_FindOIDByTag(SEC_OID_AES_256_CBC);
		break;
	default:
		pkinit_debug(mcontext, 0, "Internal error: no knowledge of the "
			     "AES encryption algorithm (%d bits).\n",
			     key_size_bits);
		return SECFailure;
	}
	if (oid == NULL) {
		pkinit_debug(mcontext, 0, "Internal error: no knowledge of the "
			     "AES-%d encryption algorithm.\n", key_size_bits);
		return SECFailure;
	}
	alg->algorithm = oid->oid;

	/* Generate an IV. */
	params.iv.len = 16;
	params.iv.data = PORT_ArenaZAlloc(pool, params.iv.len);
	if (params.iv.data == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	octets.length = params.iv.len;
	octets.data = (char *) params.iv.data;
	if (krb5_c_random_make_octets(kcontext, &octets) != 0) {
		pkinit_debug(mcontext, 0, "Out of entropy.\n");
		return SECFailure;
	}

	/* Encode the parameters structure. */
	if (SEC_ASN1EncodeItem(pool, &encoded, &params,
			       common_aes_cbc_parameters_template) != &encoded) {
		pkinit_debug(mcontext, 0, "Error encoding parameters.\n");
		return SECFailure;
	}
	alg->parameters = SECITEM_ArenaDupItem(pool, &encoded);
	if (alg->parameters == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}

	/* Build the parameter block for the PKCS11 layer. */
	mech->mechanism = CKM_AES_CBC;
	pk_params = PORT_ArenaZAlloc(pool, 16);
	if (pk_params == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	memcpy(&pk_params[0], params.iv.data, 16);
	mech->pParameter = pk_params;
	mech->ulParameterLen = 16;

	/* The consumer of the key is going to need to know how to pad the
	 * data for encryption. */
	*block_size = 16;

	return SECSuccess;
}

SECStatus
common_generate_content_encryption_key_aes128(struct module_context *mcontext,
					      PLArenaPool *pool,
					      krb5_context kcontext,
					      struct algorithm_identifier *alg,
					      SECItem *content_enc_key,
					      CK_MECHANISM *mech,
					      int *block_size)
{
	return common_generate_content_encryption_key_aes(mcontext, pool, 128,
							  kcontext,
							  alg,
							  content_enc_key,
							  mech, block_size);
}

SECStatus
common_generate_content_encryption_key_aes192(struct module_context *mcontext,
					      PLArenaPool *pool,
					      krb5_context kcontext,
					      struct algorithm_identifier *alg,
					      SECItem *content_enc_key,
					      CK_MECHANISM *mech,
					      int *block_size)
{
	return common_generate_content_encryption_key_aes(mcontext, pool, 192,
							  kcontext,
							  alg,
							  content_enc_key,
							  mech, block_size);
}

SECStatus
common_generate_content_encryption_key_aes256(struct module_context *mcontext,
					      PLArenaPool *pool,
					      krb5_context kcontext,
					      struct algorithm_identifier *alg,
					      SECItem *content_enc_key,
					      CK_MECHANISM *mech,
					      int *block_size)
{
	return common_generate_content_encryption_key_aes(mcontext, pool, 256,
							  kcontext,
							  alg,
							  content_enc_key,
							  mech, block_size);
}

SECStatus
common_generate_content_encryption_key(struct module_context *mcontext,
				       PLArenaPool *pool,
				       krb5_context kcontext,
				       SECOidTag preferred_cek,
				       struct algorithm_identifier *algorithm,
				       SECItem *content_enc_key,
				       CK_MECHANISM *mech,
				       int *block_size)
{
	SECStatus status;
	/* Try to generate a key using the specified algorithm. */
	switch (preferred_cek) {
	case SEC_OID_RC2_CBC:
		status = common_generate_content_encryption_key_rc2(mcontext,
								    pool,
								    kcontext,
								    algorithm,
								    content_enc_key,
								    mech,
								    block_size);
		if (status == SECSuccess) {
			return status;
		}
		break;
	case SEC_OID_DES_EDE3_CBC:
		status = common_generate_content_encryption_key_des3(mcontext,
								     pool,
								     kcontext,
								     algorithm,
								     content_enc_key,
								     mech,
								     block_size);
		if (status == SECSuccess) {
			return status;
		}
		break;
	case SEC_OID_AES_128_CBC:
		status = common_generate_content_encryption_key_aes128(mcontext,
								       pool,
								       kcontext,
								       algorithm,
								       content_enc_key,
								       mech,
								       block_size);
		if (status == SECSuccess) {
			return status;
		}
		break;
	case SEC_OID_AES_192_CBC:
		status = common_generate_content_encryption_key_aes192(mcontext,
								       pool,
								       kcontext,
								       algorithm,
								       content_enc_key,
								       mech,
								       block_size);
		if (status == SECSuccess) {
			return status;
		}
		break;
	case SEC_OID_AES_256_CBC:
		status = common_generate_content_encryption_key_aes256(mcontext,
								       pool,
								       kcontext,
								       algorithm,
								       content_enc_key,
								       mech,
								       block_size);
		if (status == SECSuccess) {
			return status;
		}
		break;
	default:
		break;
	}
	/* Try to generate a key using *our* preferred order. */
	status = common_generate_content_encryption_key_des3(mcontext,
							     pool,
							     kcontext,
							     algorithm,
							     content_enc_key,
							     mech,
							     block_size);
	if (status == SECSuccess) {
		return status;
	}
	status = common_generate_content_encryption_key_rc2(mcontext,
							    pool,
							    kcontext,
							    algorithm,
							    content_enc_key,
							    mech,
							    block_size);
	if (status == SECSuccess) {
		return status;
	}
	status = common_generate_content_encryption_key_aes256(mcontext,
							       pool,
							       kcontext,
							       algorithm,
							       content_enc_key,
							       mech,
							       block_size);
	if (status == SECSuccess) {
		return status;
	}
	return SECSuccess;
}

SECItem *
common_make_random_item(struct module_context *mcontext, PLArenaPool *pool,
			krb5_context kcontext, size_t length)
{
	krb5_data datum;
	SECItem *ret;
	datum.data = malloc(length);
	if (datum.data == NULL) {
		return NULL;
	}
	datum.length = length;
	if (krb5_c_random_make_octets(kcontext, &datum) != 0) {
		free(datum.data);
		return NULL;
	}
	ret = common_secitem_from_krb5_data(mcontext, pool, &datum);
	free(datum.data);
	return ret;
}

void
common_simplify_template(SEC_ASN1Template *simplified_templates,
			 const SEC_ASN1Template *templates,
			 int count)
{
	int i;
	for (i = 0; i < count; i++) {
		simplified_templates[i] = templates[i];
		simplified_templates[i].kind &= (~SEC_ASN1_CONSTRUCTED);
	}
}

PRBool
common_decode_unsigned_long(SECItem *item, unsigned long *ul)
{
	uint32_t u;
	unsigned int i;
	if (item->len > sizeof(unsigned long)) {
		if (!((item->len == sizeof(unsigned long) + 1) &&
		      (item->data[0] == 0))) {
			return PR_FALSE;
		}
	}
	u = 0;
	for (i = 0; i < item->len; i++) {
		u = (u << 8) | item->data[i];
	}
	*ul = u;
	return PR_TRUE;
}
