/*
 * Copyright (C) 2006,2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: map-file.c,v 1.3 2007/05/22 21:19:02 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <krb5/krb5.h>

#include <prmem.h>
#include <nss.h>

#include "pkinit.h"
#include "map-file.h"

struct map_file {
	struct module_context *mcontext;
	PLArenaPool *pool;
	char *pathname;
	struct principal {
		char *principal;
		char **dn;
	} *principal;
	unsigned int n_principal_used, n_principal_allocated;
};

struct map_file *
map_file_open(struct module_context *mcontext, const char *pathname)
{
	struct map_file *ret;
	PLArenaPool *pool;

	ret = NULL;
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		return NULL;
	}

	ret = PORT_ArenaZAlloc(pool, sizeof(*ret));
	if (ret == NULL) {
		PORT_FreeArena(pool, PR_TRUE);
		return NULL;
	}

	ret->mcontext = mcontext;
	ret->pool = pool;
	ret->pathname = PORT_ArenaStrdup(pool, pathname);
	ret->n_principal_used = 0;
	ret->n_principal_allocated = 0;
	ret->principal = NULL;

	return ret;
}

static int
find_principal(struct map_file *map, const char *principal)
{
	unsigned int i;
	for (i = 0; i < map->n_principal_used; i++) {
		if (strcmp(map->principal[i].principal, principal) == 0) {
			break;
		}
	}
	if (i >= map->n_principal_used) {
		return -1;
	}
	return i;
}

static void
add_principal(struct map_file *map, const char *principal_name)
{
	FILE *fp;
	char *buf;
	size_t buflen;
	unsigned int allocated, name_len;
	struct principal *principal;
	struct dn_list {
		char *dn;
		struct dn_list *next;
	} *tmp, *head;
	int count;

	if (map->n_principal_used == map->n_principal_allocated) {
		allocated = map->n_principal_allocated + 128;
		principal = PORT_ArenaZAlloc(map->pool,
					     sizeof(*principal) * allocated);
		if (principal != NULL) {
			memcpy(principal, map->principal,
			       sizeof(*principal) * map->n_principal_allocated);
			map->principal = principal;
			map->n_principal_allocated = allocated;
		} else {
			/* Nothing we can do; degrade. */
			return;
		}
	}
	principal = &map->principal[map->n_principal_used++];
	principal->principal = PORT_ArenaStrdup(map->pool, principal_name);

	fp = fopen(map->pathname, "r");
	if (fp == NULL) {
		pkinit_debug(map->mcontext, 0,
			     "Error reading map file \"%s\".\n", map->pathname);
		return;
	}
	pkinit_debug(map->mcontext, 2,
		     "Scanning for \"%s\" in map file \"%s\".\n",
		     principal_name, map->pathname);

	name_len = strlen(principal_name);
	buflen = name_len * 2 > LINE_MAX ? name_len * 2 : LINE_MAX;
	buf = PR_Malloc(buflen);
	if (buf == NULL) {
		fclose(fp);
		return;
	}

	head = NULL;
	count = 0;
	while (fgets(buf, buflen, fp) != NULL) {
		if (buf[0] == '#') {
			continue;
		}
		if ((strncmp(buf, principal_name, name_len) != 0) ||
		    (buf[name_len] != ':') ||
		    (strcspn(buf + name_len + 1, "\r\n") == 0)) {
			continue;
		}
		tmp = PR_Malloc(sizeof(*tmp));
		if (tmp != NULL) {
			tmp->dn = PORT_ArenaStrdup(map->pool,
						   buf + name_len + 1);
			tmp->dn[strcspn(tmp->dn, "\r\n")] = '\0';
			tmp->next = head;
			head = tmp;
			count++;
		}
	}
	if (count > 0) {
		principal->dn = PORT_ArenaZAlloc(map->pool,
						 sizeof(char *) * (count + 1));
		if (principal->dn != NULL) {
			while ((head != NULL) && (count > 0)) {
				tmp = head->next;
				principal->dn[count - 1] = head->dn;
				PORT_Free(head);
				head = tmp;
				count--;
			}
		}
	}

	PR_Free(buf);
	fclose(fp);
}

void
map_file_read_principal(struct map_file *map, const char *principal_name)
{
	int i;
	if (map == NULL) {
		return;
	}
	i = find_principal(map, principal_name);
	if (i == -1) {
		add_principal(map, principal_name);
	}
}

const char **
map_file_get_list(struct map_file *map, const char *principal_name)
{
	int i;
	if (map == NULL) {
		return NULL;
	}
	map_file_read_principal(map, principal_name);
	i = find_principal(map, principal_name);
	if (i != -1) {
		return (const char **) map->principal[i].dn;
	}
	return NULL;
}

void
map_file_close(struct map_file *map)
{
	if (map == NULL) {
		return;
	}
	PORT_FreeArena(map->pool, PR_TRUE);
}
