/*
 * Copyright (C) 2006,2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: pkinit.c,v 1.63 2007/07/11 21:22:02 nalin Exp $"

#include "../config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include <krb5/krb5.h>
#ifndef HAVE_KRB5_GIC_OPT_PA_DATA
typedef struct _krb5_gic_opt_pa_data krb5_gic_opt_pa_data;
#endif
#include <krb5/preauth_plugin.h>

#include <nss.h>
#include <cert.h>
#include <certdb.h>
#include <ocsp.h>
#include <prerror.h>

#include "commont.h"
#include "bcmst.h"
#include "pkinitt.h"
#include "certs.h"
#include "map-file.h"
#include "pkinit.h"

#ifdef PKINIT_USE_PAL_BACKPORT
#include "backport-errors.h"
#define preauthentication_client_0 preauthentication_client_0_backport_1_6
#define preauthentication_server_0 preauthentication_server_0_backport_1_6
#endif

#ifndef KRB5_PADATA_PK_AS_REQ_OLD
#undef  KRB5_PADATA_PK_AS_REQ
#define KRB5_PADATA_PK_AS_REQ_OLD	14
#define KRB5_PADATA_PK_AS_REQ		16
#endif
#ifndef KRB5_PADATA_PK_AS_REP_OLD
#undef  KRB5_PADATA_PK_AS_REP
#define KRB5_PADATA_PK_AS_REP_OLD	15
#define KRB5_PADATA_PK_AS_REP		17
#endif

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(_x) dgettext(PACKAGE_NAME, _x)
#else
#define _(_x) _x
#endif

static void
derdump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static void
dump(const char *text, SECItem *item)
{
#if DEBUG
	FILE *pp;
	unsigned int i;
	pp = popen("od -t x1c", "w");
	if (pp) {
		if (text) {
			fprintf(stderr, "%s: dumping %d bytes.\n", text,
				item->len);
		} else {
			fprintf(stderr, "Dumping %d bytes.\n", item->len);
		}
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
#endif
}
static krb5_preauthtype supported_client_pa_types[] = {
	KRB5_PADATA_PK_AS_REP_OLD,
	KRB5_PADATA_PK_AS_REQ, KRB5_PADATA_PK_AS_REP,
	0,
};
static krb5_preauthtype supported_server_pa_types[] = {
	KRB5_PADATA_PK_AS_REQ,
	0,
};
static krb5_enctype supported_pk_enc_types[] = {
	11, /* sha-1WithRSAEncryption */
	10, /* md5WithRSAEncryption-CmsOID */
	15, /* des-ede3-cbc-EnvOID */
	12, /* rc2-cbc-EnvOID */
	13, /* rsaEncryption-EnvOID */
	0,
};

struct server_context {
	krb5_data *client_request;
	SECItem *client_cert;
	SECItem *client_public_info;
	SECItem *client_nonce;
	SECItem *client_dh_nonce;
};

struct module_context {
	PLArenaPool *pool;
	struct {
		SECKEYPrivateKey *dh_private_key;
		SECKEYPublicKey *dh_public_key;
		SECItem *dh_nonce;
	} client_request;
	struct {
		struct external_principal_identifier **restricted_cas;
		struct external_principal_identifier **banned_certs;
		struct domain_parameters **allowed_dh_params;
	} server_constraints;
	PRBool allow_pkinit, enable_ocsp, owns_nss, try_dh, is_hw;
	PRBool trust_pkinit_san, trust_upn_san;
	int preferred_group;
	unsigned int minimum_dh_prime_size;
	CERTCertDBHandle *certdb;
	struct aa_bag *bag;
	struct map_file *map;
	char *cert_file, *key_file, *cacert_file, *cert_pool, *cacert_pool;
	int debug_level;
	PRBool debug_syslog, debug_stdout, debug_stderr;
};

void
pkinit_debug(struct module_context *context, int level,
	     const char *fmt, ...)
{
	va_list ap, ap2, ap3;
	if (context != NULL) {
		if (context->debug_level >= level) {
			va_start(ap, fmt);
			va_copy(ap2, ap);
			va_copy(ap3, ap);
			if (context->debug_stdout) {
				if (isatty(fileno(stdout))) {
					vfprintf(stdout, fmt, ap);
					fflush(stdout);
				}
			}
			if (context->debug_stderr) {
				if (isatty(fileno(stderr))) {
					vfprintf(stderr, fmt, ap2);
					fflush(stderr);
				}
			}
			if (context->debug_syslog) {
				vsyslog(LOG_DEBUG, fmt, ap3);
			}
			va_end(ap);
			va_end(ap2);
			va_end(ap3);
		}
	}
}

/* Initialize NSS, read configuration, enable/disable OCSP checking, start up
 * i18n support. */
static krb5_error_code
pkinit_init(krb5_context kcontext, PRBool server, struct module_context **ctx)
{
	int pkinit, ocsp, try_dh, is_hw, preferred_group, minimum_dh_prime_size;
	int trust_pkinit_san, trust_upn_san;
	int debug_level, debug_syslog, debug_stdout, debug_stderr;
	char *p, *q, *mappings_file, *dbdir;
	char *cert_file, *key_file, *cacert_file, *cert_pool, *cacert_pool;
	PRBool owns_nss;
	struct module_context *context;
	krb5_data *default_realm = NULL;
	PLArenaPool *pool;
	SECStatus ret;

	*ctx = NULL;

	/* Check if we're disabled. */
	pkinit_debug(NULL, 2, "Checking if PKINIT is allowed.\n");
	pkinit = TRUE;
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"allow_pkinit", TRUE, &pkinit);
	if (server) {
		krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
					"allow_pkinit_server", pkinit, &pkinit);
	} else {
		krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
					"allow_pkinit_client", pkinit, &pkinit);
	}
	if (pkinit) {
		pkinit_debug(NULL, 2, "PKINIT is allowed.\n");
	} else {
		pkinit_debug(NULL, 2, "PKINIT is disallowed.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Initialize the library, if we need to. */
	PK11_SetPasswordFunc(pkinit_pwcb);
	owns_nss = !NSS_IsInitialized();
	if (owns_nss) {
		if (server) {
			dbdir = DEFAULT_PKINIT_SERVER_DBDIR;
			krb5_appdefault_string(kcontext, "pkinit",
					       default_realm,
					       "server_database",
					       DEFAULT_PKINIT_SERVER_DBDIR,
					       &dbdir);
		} else {
			dbdir = DEFAULT_PKINIT_CLIENT_DBDIR;
			krb5_appdefault_string(kcontext, "pkinit",
					       default_realm,
					       "client_database",
					       DEFAULT_PKINIT_CLIENT_DBDIR,
					       &dbdir);
		}
		pkinit_debug(NULL, 2, "Using database in \"%s\".\n", dbdir);
		ret = NSS_Init(dbdir);
		free(dbdir);
		if (ret != SECSuccess) {
			char *err = NULL;
			PRInt32 errlen = PR_GetErrorTextLength();

			if (errlen > 0) {
				err = malloc(errlen + 1);
				if (err) {
					errlen = PR_GetErrorText(err);
					pkinit_debug(NULL, 1,
						     "NSS_Init returned error "
						     "%d(%s)\n",
						     ret,
						     errlen ? err : "none");
					free(err);
				}
			}
			return EINVAL;
		}
	}

	/* Build the module context. */
	pool = PORT_NewArena(sizeof(double));
	context = PORT_ArenaZAlloc(pool, sizeof(*context));
	if (context == NULL) {
		if (owns_nss) {
			if (NSS_Shutdown() != SECSuccess) {
				pkinit_debug(NULL, 0,
					     "Failed to shut down NSS.\n");
			}
		}
		return ENOMEM;
	}

	context->certdb = CERT_GetDefaultCertDB();
	if (context->certdb == NULL) {
		pkinit_debug(NULL, 1, "No NSS database found.\n");
		if (owns_nss) {
			if (NSS_Shutdown() != SECSuccess) {
				pkinit_debug(NULL, 0,
					     "Failed to shut down NSS.\n");
			}
		}
		return EINVAL;
	}
	context->bag = aa_bag_new(context, context->certdb, pkinit_pwcb);
	if (context->bag == NULL) {
		pkinit_debug(NULL, 1,
			     "Error initializing certicate storage.\n");
		if (owns_nss) {
			if (NSS_Shutdown() != SECSuccess) {
				pkinit_debug(NULL, 0,
					     "Failed to shut down NSS.\n");
			}
		}
		return EINVAL;
	}

	/* Load our private key and certificate. */
	cert_file = NULL;
	key_file = NULL;
	cacert_file = NULL;
	cert_pool = NULL;
	cacert_pool = NULL;
	if (server) {
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "server_certificate", "",
				       &cert_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "server_private_key", "",
				       &key_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "server_ca_certificate", "",
				       &cacert_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "server_certificate_pool", "",
				       &cert_pool);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "server_ca_certificate_pool", "",
				       &cacert_pool);
	} else {
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "client_certificate", "",
				       &cert_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "client_private_key", "",
				       &key_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "client_ca_certificate", "",
				       &cacert_file);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "client_certificate_pool", "",
				       &cert_pool);
		krb5_appdefault_string(kcontext, "pkinit", default_realm,
				       "client_ca_certificate_pool", "",
				       &cacert_pool);
	}

	/* Check the location of our mappings file, if there is one. */
	krb5_appdefault_string(kcontext, "pkinit", default_realm,
			       "mappings_file", "", &mappings_file);

	/* Turn OCSP checking on or off.  Default is on for servers, off for
	 * clients. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"ocsp_checking",
				server ? PR_TRUE : PR_FALSE, &ocsp);
	if (ocsp) {
		CERT_EnableOCSPChecking(context->certdb);
	} else {
		CERT_DisableOCSPChecking(context->certdb);
	}

	/* Control whether we count PKINIT as a hardware-based scheme. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"is_hw", PR_FALSE, &is_hw);

	/* Control whether we try to use DH key agreement or not. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"try_dh", PR_TRUE, &try_dh);

	/* Figure out which DH parameters to give to the KDC first. */
	krb5_appdefault_string(kcontext, "pkinit", default_realm,
			       "preferred_group", "2", &p);
	preferred_group = strtol(p, &q, 0);
	if ((q == NULL) || (*q != '\0')) {
		preferred_group = 0;
	}
	free(p);
	krb5_appdefault_string(kcontext, "pkinit", default_realm,
			       "minimum_dh_prime_size", "1024", &p);
	minimum_dh_prime_size = strtol(p, &q, 0);
	if ((q == NULL) || (*q != '\0')) {
		minimum_dh_prime_size = 1024;
	}
	free(p);

	/* Control whether we trust UPN SAN values or not. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"trust_upn_san", PR_TRUE, &trust_upn_san);
	/* Control whether we trust PKINIT SAN values or not. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"trust_pkinit_san", PR_TRUE, &trust_pkinit_san);
	/* Set debug logging level. */
	krb5_appdefault_string(kcontext, "pkinit", default_realm,
			       "debug_level", "0", &p);
	debug_level = strtol(p, &q, 0);
	if ((q == NULL) || (*q != '\0')) {
		debug_level = 0;
	}
	free(p);
	/* Set debug logging to stderr/syslog. */
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"debug_syslog", PR_TRUE, &debug_syslog);
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"debug_stderr", PR_FALSE, &debug_stderr);
	krb5_appdefault_boolean(kcontext, "pkinit", default_realm,
				"debug_stdout", PR_FALSE, &debug_stdout);

#ifdef HAVE_LIBINTL_H
	bindtextdomain(PACKAGE_NAME, LOCALEDATADIR);
#endif

	/* Fill in the rest. */
	context->pool = pool;
	context->allow_pkinit = pkinit;
	context->owns_nss = owns_nss;
	context->enable_ocsp = ocsp;
	context->is_hw = is_hw;
	context->try_dh = try_dh;
	context->trust_upn_san = trust_upn_san;
	context->trust_pkinit_san = trust_pkinit_san;
	context->preferred_group = preferred_group;
	context->minimum_dh_prime_size = minimum_dh_prime_size;
	context->debug_level = debug_level;
	context->debug_syslog = debug_syslog;
	context->debug_stderr = debug_stderr;
	if ((mappings_file != NULL) && (mappings_file[0] == '/')) {
		context->map = map_file_open(context, mappings_file);
	} else {
		context->map = NULL;
	}
	free(mappings_file);
	if (strlen(cert_file) > 0) {
		context->cert_file = cert_file;
	} else {
		free(cert_file);
	}
	if (strlen(key_file) > 0) {
		context->key_file = key_file;
	} else {
		free(key_file);
	}
	if (strlen(cacert_file) > 0) {
		context->cacert_file = cacert_file;
	} else {
		free(cacert_file);
	}
	if (strlen(cert_pool) > 0) {
		context->cert_pool = cert_pool;
	} else {
		free(cert_pool);
	}
	if (strlen(cacert_pool) > 0) {
		context->cacert_pool = cacert_pool;
	} else {
		free(cacert_pool);
	}

	*ctx = context;

	return 0;
}

static void
pkinit_fini(krb5_context unused_kcontext, struct module_context *context)
{
	PLArenaPool *pool;
	int owns_nss;
	/* Save a copy of the parts of the context we still care about. */
	pool = context->pool;
	owns_nss = context->owns_nss;
	/* Free paths which we kept. */
	free(context->cert_file);
	free(context->key_file);
	free(context->cacert_file);
	free(context->cert_pool);
	free(context->cacert_pool);
	/* Close out our mapping list. */
	if (context->map != NULL) {
		map_file_close(context->map);
	}
	/* Free the certificate/key bag. */
	if (context->bag != NULL) {
		aa_bag_free(context->bag);
	}
	/* Destroy any DH keying data we've still got lying around. */
	if (context->client_request.dh_private_key != NULL) {
		SECKEY_DestroyPrivateKey(context->client_request.dh_private_key);
	}
	if (context->client_request.dh_public_key != NULL) {
		SECKEY_DestroyPublicKey(context->client_request.dh_public_key);
	}
	/* Zero out the structure. */
	memset(context, 0, sizeof(*context));
	/* Free the pool. */
	PORT_FreeArena(pool, PR_TRUE);
	if (owns_nss) {
		if (NSS_Shutdown() != SECSuccess) {
			pkinit_debug(NULL, 0, "Failed to shut down NSS.\n");
		}
	}
}

char *
pkinit_pwcb(PK11SlotInfo *slot, PRBool retry, void *arg)
{
	int ret;
	struct pkinit_pwcb_args *args;
	krb5_prompt prompt;
	krb5_data reply;
	char *name, *text, *answer;
	void *data;

	/* We only want to be called once. */
	if (retry) {
		pkinit_debug(NULL, 2,
			     "Prompter callback called more than once, "
			     "failing\n");
		return NULL;
	}
	/* We need our callback arguments. */
	if (arg == NULL) {
		pkinit_debug(NULL, 2,
			     "Prompter callback called with no args.\n");
		return NULL;
	}
	args = arg;
	if (args->mcontext == NULL) {
		pkinit_debug(NULL, 2,
			     "Prompter callback called with no module "
			     "context.\n");
		return NULL;
	}
	if (args->kcontext == NULL) {
		pkinit_debug(args->mcontext, 2,
			     "Prompter callback called with no Kerberos "
			     "context.\n");
		return NULL;
	}
	if (args->fct == NULL) {
		pkinit_debug(args->mcontext, 2,
			     "Prompter callback called with no prompter.\n");
		return NULL;
	}
	if (args->pool == NULL) {
		pkinit_debug(args->mcontext, 2,
			     "Prompter callback called with no pool.\n");
		return NULL;
	}

	/* Get the name of the token. */
	if (args->name != NULL) {
		name = args->name;
	} else {
		name = PK11_GetTokenName(slot);
	}
	if (name == NULL) {
		pkinit_debug(args->mcontext, 2,
			     "Skipping nameless prompt.\n");
		return NULL;
	}

	/* Set up the prompt. */
	text = PORT_ArenaZAlloc(args->pool, strlen(name) + 100);
	snprintf(text, strlen(name) + 100, _("PIN for %s"), name);
	memset(&prompt, 0, sizeof(prompt));
	prompt.prompt = text;
	prompt.hidden = 1;
	prompt.reply = &reply;
	reply.length = 256;
	data = malloc(reply.length);
	reply.data = data;
	name = NULL;
	answer = NULL;

	if ((args->mcontext != NULL) && (args->mcontext->debug_level > 1)) {
		name = "PKINIT-NSS";
		answer = PACKAGE_STRING;
	}

	/* Call the callback. */
	ret = (*args->fct)(args->kcontext, args->fct_data,
			   name, answer, 1, &prompt);
	answer = NULL;

	if ((ret == 0) && (reply.data != NULL)) {
		answer = PR_Malloc(reply.length + 1);
		memcpy(answer, reply.data, reply.length);
		answer[reply.length] = '\0';
	}

	if (reply.data == data) {
		free(reply.data);
	}

	return answer;
}

static krb5_error_code
client_init(krb5_context kcontext, void **plugin_context)
{
	struct module_context *context;
	krb5_error_code ret;

	*plugin_context = NULL;
	ret = pkinit_init(kcontext, PR_FALSE, &context);
	if (ret == 0) {
		*plugin_context = context;
	}

	return ret;
}

static void
client_fini(krb5_context kcontext, void *plugin_context)
{
	struct module_context *context;

	context = plugin_context;
	if (context != NULL) {
		pkinit_fini(kcontext, context);
	}
}

static krb5_error_code
client_gic_opts(krb5_context unused_kcontext, void *plugin_context,
		krb5_get_init_creds_opt *unused_opt,
		const char *attr, const char *value)
{
	struct module_context *context;
	context = plugin_context;
	pkinit_debug(context, 2, "get_init_creds_opt(\"%s\") = \"%s\"\n",
		     attr, value);
	return 0;
}

static const char *
client_gic_opt(krb5_context kcontext, krb5_get_init_creds_opt *opt,
	       const char *attr)
{
	const char *ret = NULL;
#ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_GET_PA
	int num_preauth_data = 0, i;
	krb5_gic_opt_pa_data *preauth_data = NULL;

	if (krb5_get_init_creds_opt_get_pa(kcontext, opt,
					   &num_preauth_data,
					   &preauth_data) == 0) {
		for (i = 0; i < num_preauth_data; i++) {
			if (strcmp(attr, preauth_data[i].attr) == 0) {
				return preauth_data[i].value;
			}
		}
	}
#endif
	return ret;
}

static void
client_gic_debug_opts(struct module_context *mcontext, krb5_context kcontext,
		      krb5_get_init_creds_opt *opt)
{
	const char *debug, *p, *q;
	/* Parse dynamic debug options. */
	if ((debug = client_gic_opt(kcontext, opt, "debug")) != NULL) {
		p = debug;
		q = debug;
		while (*q != '\0') {
			q = p + strcspn(p, ",");
			if (q - p > 0) {
				if (((q - p) == 6) &&
				    (strncmp(p, "stdout", q - p) == 0)) {
					mcontext->debug_stdout = PR_TRUE;
				} else
				if (((q - p) == 6) &&
				    (strncmp(p, "stderr", q - p) == 0)) {
					mcontext->debug_stderr = PR_TRUE;
				} else
				if (((q - p) == 6) &&
				    (strncmp(p, "syslog", q - p) == 0)) {
					mcontext->debug_syslog = PR_TRUE;
				} else
				if (strspn(p, "0123456789") == (q - p)) {
					mcontext->debug_level = strtol(p,
								       NULL,
								       10);
				}
			}
			p = q + 1;
		}
	}
}

static krb5_error_code
client_process(krb5_context kcontext,
	       void *plugin_context,
	       void *unused_request_context,
#ifndef PKINIT_CLIENT_PROCESS_MISSING_OPT
	       krb5_get_init_creds_opt *opt,
#endif
	       preauth_get_client_data_proc unused_get_data_proc,
	       struct _krb5_preauth_client_rock *unused_rock,
	       krb5_kdc_req *request,
	       krb5_data *encoded_request_body,
	       krb5_data *encoded_previous_request,
	       krb5_pa_data *pa_data,
	       krb5_prompter_fct prompter,
	       void *prompter_data,
	       preauth_get_as_key_proc unused_gak_fct,
	       void *unused_gak_data,
	       krb5_data *unused_salt, krb5_data *unused_s2kparams,
	       krb5_keyblock *as_key,
	       krb5_pa_data **out_pa_data)
{
#ifdef PKINIT_CLIENT_PROCESS_MISSING_OPT
	krb5_get_init_creds_opt *opt = NULL;
#endif
	krb5_data *req, rep;
	krb5_pa_data *pa_req;
	krb5_error_code ret;
	CERTCertificate *cert = NULL;
	PK11SlotInfo *preferred_slot;
	struct pkinit_pwcb_args pwcb_args;
	const char *preferred_slot_name, *tmp;
	const char **additional_client_dn_list, **additional_kdc_dn_list;
	char *unparsed_name;
	struct module_context *context;

	*out_pa_data = NULL;
	preferred_slot = NULL;
	context = plugin_context;
	ret = KRB5KRB_ERR_GENERIC;

	/* Hey! No cheating! Either we have a challenge or a response. */
	if (pa_data == NULL) {
		return EINVAL;
	}

	/* Set up callbacks for the library in case it needs to ask the user
	 * for a PIN. */
	memset(&pwcb_args, 0, sizeof(pwcb_args));
	pwcb_args.kcontext = kcontext;
	pwcb_args.mcontext = context;
	pwcb_args.fct = prompter;
	pwcb_args.fct_data = prompter_data;
	pwcb_args.name = NULL;
	pwcb_args.pool = context->pool;

	/* Allow command-line overrides of debug message settings. */
	client_gic_debug_opts(context, kcontext, opt);

	/* Let the client specify a smaller prime size. */
	tmp = client_gic_opt(kcontext, opt, "minimum_dh_prime_size");
	if (tmp != NULL) {
		if (strspn(tmp, "0123456789") != strlen(tmp)) {
			 pkinit_debug(context, 0,
				      "Ignoring non-number minimum DH prime "
				      "size \"%s\".\n", tmp);
		} else {
			 context->minimum_dh_prime_size = atoi(tmp);
			 pkinit_debug(context, 2,
				      "Set minimum DH prime size to %d.\n",
				      context->minimum_dh_prime_size);
		}
	}

	/* Load our private key and certificate, trusted CA certificates, and
	 * set up search directories for intermediate and CA certificate. */
	tmp = NULL;
	if ((tmp = client_gic_opt(kcontext, opt, "private_key_file")) == NULL) {
		tmp = context->key_file;
	}
	if (tmp != NULL) {
		if (aa_bag_load_key_file(context->bag, tmp,
					 &pwcb_args) == SECSuccess) {
			 pkinit_debug(context, 2,
				      "Loaded private key from \"%s\".\n",
				      tmp);
		} else {
			 pkinit_debug(context, 0,
				      "Failed to load private key from "
			 	      "\"%s\".\n", tmp);
		}
	}
	tmp = NULL;
	if ((tmp = client_gic_opt(kcontext, opt, "certificate_file")) == NULL) {
		tmp = context->cert_file;
	}
	if (tmp != NULL) {
		if (aa_bag_load_cert_file(context->bag, tmp,
					  &pwcb_args) == SECSuccess) {
			 pkinit_debug(context, 2,
				      "Loaded certificate from \"%s\".\n", tmp);
		} else {
			 pkinit_debug(context, 0,
				      "Failed to load certificate from "
			 	      "\"%s\".\n", tmp);
		}
	}
	tmp = NULL;
	if ((tmp = client_gic_opt(kcontext, opt,
				  "ca_certificate_file")) == NULL) {
		tmp = context->cacert_file;
	}
	if (tmp != NULL) {
		if (aa_bag_load_cacert_file(context->bag, tmp,
					    &pwcb_args) == SECSuccess) {
			 pkinit_debug(context, 2,
				      "Loaded CA certificate from \"%s\".\n",
				      tmp);
		} else {
			 pkinit_debug(context, 0,
				      "Failed to load CA certificate from "
			 	      "\"%s\".\n", tmp);
		}
	}
	tmp = NULL;
	if ((tmp = client_gic_opt(kcontext, opt, "certificate_pool")) == NULL) {
		tmp = context->cert_pool;
	}
	if (tmp != NULL) {
		if (aa_bag_add_cert_search_dir(context->bag,
					       tmp) == SECSuccess) {
			 pkinit_debug(context, 2,
				      "Added \"%s\" to certificate "
				      "search pool.\n", tmp);
		} else {
			 pkinit_debug(context, 0,
				      "Failed to add \"%s\" to certificate "
				      "search pool.\n", tmp);
		}
	}
	tmp = NULL;
	if ((tmp = client_gic_opt(kcontext, opt,
				  "ca_certificate_pool")) == NULL) {
		tmp = context->cacert_pool;
	}
	if (tmp != NULL) {
		if (aa_bag_add_cacert_search_dir(context->bag,
						 tmp) == SECSuccess) {
			 pkinit_debug(context, 2,
				      "Added \"%s\" to CA certificate "
				      "search pool.\n", tmp);
		} else {
			 pkinit_debug(context, 0,
				      "Failed to add \"%s\" to CA certificate "
				      "search pool.\n", tmp);
		}
	}

	/* If we have a preferred slot, try to get a handle for it. */
	preferred_slot_name = getenv("PKCS11_LOGIN_TOKEN_NAME");
	if (preferred_slot_name != NULL) {
		preferred_slot = PK11_FindSlotByName(preferred_slot_name);
	}

	/* Try to find the preferred certificate for the user. */
	pkinit_debug(context, 2, "Looking for the user's certificate.\n");
	additional_client_dn_list = NULL;
	if (krb5_unparse_name(kcontext, request->client, &unparsed_name) == 0) {
		additional_client_dn_list = map_file_get_list(context->map,
							      unparsed_name);
		krb5_free_unparsed_name(kcontext, unparsed_name);
	}
	cert = cert_find_preferred_cert(context, context->pool,
					preferred_slot, context->bag,
					&pwcb_args,
					NULL,
					NULL,
					additional_client_dn_list,
					kcontext, request->client,
					CERT_MATCHES_DN_LIST |
					(context->trust_pkinit_san ?
					 CERT_MATCHES_KRB_SAN : 0) |
					(context->trust_upn_san ?
					 CERT_MATCHES_UPN_SAN : 0));
	if (cert == NULL) {
		pkinit_debug(context, 1, "Unable to find certificate.\n");
		if (preferred_slot != NULL) {
			PK11_FreeSlot(preferred_slot);
		}
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}
	pkinit_debug(context, 1, "Using \"%s\".\n", cert->subjectName);

	switch (pa_data->pa_type) {
	case KRB5_PADATA_PK_AS_REP_OLD:
		if (pa_data->length == 0) {
			/* Create a new draft pa_pk_as_req structure, sign it
			 * with a matching certificate, and prepare it for
			 * transmission. */
			pkinit_debug(context, 2,
				     "Called to create a draft request.\n");
			req = pkinit_create_draft_pa_pk_as_req(context,
							       context->pool,
							       kcontext,
							       request,
							       context->certdb,
							       context->bag,
							       cert,
							       preferred_slot,
							       context->preferred_group,
							       &context->client_request.dh_private_key,
							       &context->client_request.dh_public_key,
							       &context->client_request.dh_nonce,
							       pkinit_pwcb,

							       &pwcb_args);
			/* If that succeeded, wrap the result up in a pa_data
			 * structure. */
			if (req != NULL) {
				pa_req = malloc(sizeof(krb5_pa_data));
				/* Aaagh! WHY is this not _REQ? */
				pa_req->pa_type = KRB5_PADATA_PK_AS_REP_OLD;
				pa_req->length = req->length;
				pa_req->contents = malloc(req->length);
				memcpy(pa_req->contents, req->data,
				       req->length);
				*out_pa_data = pa_req;
				/* Note success for our caller. */
				ret = 0;
			}
		} else {
			pkinit_debug(context,
				     2, "Called to handle draft response.\n");
			/* The server supplied data, so assume it's a response
			 * and process it. */
			rep.data = (char *) pa_data->contents;
			rep.length = pa_data->length;
			/* Get the list of mapped subject DNs for the realm's
			 * TGS. */
			additional_kdc_dn_list = NULL;
			unparsed_name = PR_smprintf("%s/%.*s@%.*s",
						    KRB5_TGS_NAME,
						    request->server->realm.length,
						    request->server->realm.data,
						    request->server->realm.length,
						    request->server->realm.data);
			if (unparsed_name != NULL) {
				additional_kdc_dn_list = map_file_get_list(context->map,
									   unparsed_name);
				PR_smprintf_free(unparsed_name);
			}
			ret = pkinit_verify_draft_pa_pk_as_rep(context,
							       context->pool,
							       context->certdb,
							       context->bag,
							       preferred_slot,
							       pkinit_pwcb,
							       &pwcb_args,
							       additional_kdc_dn_list,
							       CERT_MATCHES_DN_LIST |
							       (context->trust_pkinit_san ?
							        CERT_MATCHES_KRB_SAN : 0) |
							       (context->trust_upn_san ?
							        CERT_MATCHES_UPN_SAN : 0) |
							       CERT_MATCHES_TRUSTED_SERVERS,
							       kcontext,
							       request,
							       &rep,
							       as_key);
		}
		break;
	case KRB5_PADATA_PK_AS_REQ:
		/* Create a new pa_pk_as_req structure, sign it
		 * with a matching certificate, and prepare it for
		 * transmission. */
		pkinit_debug(context, 2,
			     "Called to create a RFC4556 request.\n");
		req = pkinit_create_pa_pk_as_req(context,
						 context->pool,
						 kcontext,
						 request,
						 encoded_request_body,
						 context->certdb,
						 context->bag,
						 cert,
						 SEC_OID_SHA1,
						 SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION,
						 preferred_slot,
						 context->preferred_group,
						 NULL,
						 context->minimum_dh_prime_size,
						 context->try_dh ? &context->client_request.dh_private_key : NULL,
						 context->try_dh ? &context->client_request.dh_public_key : NULL,
						 context->try_dh ? &context->client_request.dh_nonce : NULL,
						 pkinit_pwcb,
						 &pwcb_args);
		/* If that succeeded, wrap the result up in a pa_data
		 * structure. */
		if (req != NULL) {
			pa_req = malloc(sizeof(krb5_pa_data));
			pa_req->pa_type = KRB5_PADATA_PK_AS_REQ;
			pa_req->length = req->length;
			pa_req->contents = malloc(req->length);
			memcpy(pa_req->contents, req->data, req->length);
			*out_pa_data = pa_req;
			/* Note success for our caller. */
			ret = 0;
		}
		break;
	case KRB5_PADATA_PK_AS_REP:
		pkinit_debug(context, 2,
			     "Called to handle RFC4556 response.\n");
		/* The server supplied data -> process the response. */
		rep.data = (char *) pa_data->contents;
		rep.length = pa_data->length;
		/* Get the list of mapped subject DNs for the realm's TGS. */
		additional_kdc_dn_list = NULL;
		unparsed_name = PR_smprintf("%s/%.*s@%.*s",
					    KRB5_TGS_NAME,
					    request->server->realm.length,
					    request->server->realm.data,
					    request->server->realm.length,
					    request->server->realm.data);
		if (unparsed_name != NULL) {
			additional_kdc_dn_list = map_file_get_list(context->map,
								   unparsed_name);
			PR_smprintf_free(unparsed_name);
		}
		/* Verify the result. */
		ret = pkinit_verify_pa_pk_as_rep(context,
						 context->pool,
						 context->certdb,
						 context->bag,
						 preferred_slot,
						 pkinit_pwcb,
						 &pwcb_args,
						 additional_kdc_dn_list,
						 CERT_MATCHES_DN_LIST |
						 (context->trust_pkinit_san ?
						  CERT_MATCHES_KRB_SAN : 0) |
						 (context->trust_upn_san ?
						  CERT_MATCHES_UPN_SAN : 0) |
						 CERT_MATCHES_TRUSTED_SERVERS,
						 kcontext,
						 request,
						 encoded_previous_request,
						 &rep,
						 context->client_request.dh_private_key,
						 context->client_request.dh_public_key,
						 context->client_request.dh_nonce,
						 as_key);
		break;
	default:
		break;
	}

	CERT_DestroyCertificate(cert);
	if (preferred_slot != NULL) {
		PK11_FreeSlot(preferred_slot);
	}

	if (ret == 0) {
		pkinit_debug(context, 2, "Returning success to libkrb5.\n");
	} else {
		pkinit_debug(context, 1, "Returning failure (%s) to libkrb5.\n",
			     error_message(ret));
	}
	return ret;
}

static krb5_error_code
client_try_again(krb5_context kcontext,
	         void *plugin_context,
	         void *unused_request_context,
#ifndef PKINIT_CLIENT_PROCESS_MISSING_OPT
		 krb5_get_init_creds_opt *unused_opt,
#endif
	         preauth_get_client_data_proc unused_get_data_proc,
	         struct _krb5_preauth_client_rock *unused_rock,
	         krb5_kdc_req *request,
	         krb5_data *encoded_request_body,
	         krb5_data *unused_encoded_previous_request,
	         krb5_pa_data *pa_data,
	         krb5_error *error,
	         krb5_prompter_fct prompter,
	         void *prompter_data,
	         preauth_get_as_key_proc unused_gak_fct,
	         void *unused_gak_data,
	         krb5_data *unused_salt, krb5_data *unused_s2kparams,
	         krb5_keyblock *unused_as_key,
	         krb5_pa_data **out_pa_data)
{
#ifdef PKINIT_CLIENT_PROCESS_MISSING_OPT
	krb5_get_init_creds_opt *unused_opt = NULL;
#endif
	krb5_data *req;
	krb5_pa_data *pa_req;
	krb5_error_code ret;
	SECItem e_data, *td_data;
	SECOidData *oid;
	SECStatus status;
	CERTCertificate *cert = NULL;
	PK11SlotInfo *preferred_slot;
	struct pkinit_pwcb_args pwcb_args;
	const char *preferred_slot_name;
	const char **additional_client_dn_list, **additional_kdc_dn_list;
	struct algorithm_identifier **allowed_algorithms;
	struct domain_parameters *dh_params;
	struct typed_datum **typed_errors;
	char *unparsed_name;
	struct module_context *context;
	unsigned long td_type;
	int i, j, k;

	*out_pa_data = NULL;
	preferred_slot = NULL;
	context = plugin_context;
	ret = KRB5KRB_ERR_GENERIC;

	/* Set up callbacks for the library in case it needs to ask the user
	 * for a PIN. */
	pwcb_args.kcontext = kcontext;
	pwcb_args.mcontext = context;
	pwcb_args.fct = prompter;
	pwcb_args.fct_data = prompter_data;
	pwcb_args.pool = context->pool;

	/* Parse the error data. */
	memset(&e_data, 0, sizeof(e_data));
	e_data.data = (unsigned char *) error->e_data.data;
	e_data.len = error->e_data.length;
	typed_errors = pkinit_decode_typed_data(context,
						context->pool,
						&e_data);
	if (typed_errors == NULL) {
		pkinit_debug(context, 1, "Error parsing typed-data from the "
			     "KDC, not retrying.\n");
		derdump("e-data", &e_data);
		return error->error;
	}
	for (i = 0; (typed_errors[i] != NULL); i++) {
		status = SEC_ASN1DecodeInteger(&typed_errors[i]->td_type,
					       &td_type);
		if (status != SECSuccess) {
			pkinit_debug(context, 1,
				     "Error parsing typed data type.\n");
			return error->error;
		}
		pkinit_debug(context, 2, "Received typed-data %d.\n", td_type);
		td_data = &typed_errors[i]->td_data;
		switch (td_type) {
		case 104:
			context->server_constraints.restricted_cas = bcms_decode_sequence_of_external_principal_identifier(context, context->pool, td_data);
			derdump("Trusted certifiers", td_data);
			for (j = 0;
			     (context->server_constraints.restricted_cas != NULL) &&
			     (context->server_constraints.restricted_cas[j] != NULL);
			     j++) {
				continue;
			}
			pkinit_debug(context, 2,
				     "KDC sent list of %d trusted "
				     "certifiers.\n", j);
			break;
		case 105:
			context->server_constraints.banned_certs = bcms_decode_sequence_of_external_principal_identifier(context, context->pool, td_data);
			derdump("Invalid certificates", td_data);
			for (j = 0;
			     (context->server_constraints.banned_certs != NULL) &&
			     (context->server_constraints.banned_certs[j] != NULL);
			     j++) {
				continue;
			}
			pkinit_debug(context, 2,
				     "KDC sent list of %d invalid "
				     "certificates.\n", j);
			break;
		case 109:
			allowed_algorithms = common_decode_sequence_of_algorithm_identifier(context, context->pool, td_data);
			derdump("Sequence of allowed algorithms", td_data);
			if (allowed_algorithms != NULL) {
				for (j = 0; allowed_algorithms[j] != NULL; j++) {
					continue;
				}
				context->server_constraints.allowed_dh_params =
					PORT_ArenaZAlloc(context->pool,
							 sizeof(struct domain_parameters*) * (i + 1));
				if (context->server_constraints.allowed_dh_params == NULL) {
					pkinit_debug(context, 0,
						     "Out of memory.\n");
					break;
				}
				for (j = k = 0;
				     allowed_algorithms[j] != NULL;
				     j++) {
					oid = SECOID_FindOID(&allowed_algorithms[j]->algorithm);
					if (oid == NULL) {
						pkinit_debug(context, 1,
							     "DH parameter is not a DH parameter?\n");
						continue;
					}
					if (oid->offset != SEC_OID_X942_DIFFIE_HELMAN_KEY) {
						pkinit_debug(context, 1,
							     "Expected DH parameters, got %s.\n", oid->desc);
						continue;
					}
					dh_params = common_decode_domain_parameters(context, context->pool, allowed_algorithms[j]->parameters);
					if (dh_params == NULL) {
						pkinit_debug(context, 1,
							     "Error parsing domain parameters.\n");
						continue;
					}
					context->server_constraints.allowed_dh_params[k++] = dh_params;
					dump("Allowed DH prime", &dh_params->q);
					dump("Allowed DH base", &dh_params->g);
				}
			}
			break;
		default:
			break;
		}
	}

	/* If we have a preferred slot, try to get a handle for it. */
	preferred_slot_name = getenv("PKCS11_LOGIN_TOKEN_NAME");
	if (preferred_slot_name != NULL) {
		preferred_slot = PK11_FindSlotByName(preferred_slot_name);
	}

	/* Try to find the preferred certificate for the user, incorporating
	 * restrictions received from the KDC. */
	pkinit_debug(context, 2, "Looking for the user's certificate.\n");
	additional_client_dn_list = NULL;
	if (krb5_unparse_name(kcontext, request->client, &unparsed_name) == 0) {
		additional_client_dn_list = map_file_get_list(context->map,
							      unparsed_name);
		krb5_free_unparsed_name(kcontext, unparsed_name);
	}
	cert = cert_find_preferred_cert(context, context->pool,
					preferred_slot, context->bag,
					&pwcb_args,
					context->server_constraints.restricted_cas,
					context->server_constraints.banned_certs,
					additional_client_dn_list,
					kcontext, request->client,
					CERT_MATCHES_DN_LIST |
					(context->trust_pkinit_san ?
					 CERT_MATCHES_KRB_SAN : 0 ) |
					(context->trust_upn_san ?
					 CERT_MATCHES_UPN_SAN : 0));
	if (cert == NULL) {
		pkinit_debug(context, 1, "Unable to find certificate.\n");
		if (preferred_slot != NULL) {
			PK11_FreeSlot(preferred_slot);
		}
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}
	pkinit_debug(context, 1, "Using \"%s\".\n", cert->subjectName);

	switch (pa_data->pa_type) {
	case KRB5_PADATA_PK_AS_REP_OLD:
		if (pa_data->length == 0) {
			/* Create a new draft pa_pk_as_req structure, sign it
			 * with a matching certificate, and prepare it for
			 * transmission. */
			pkinit_debug(context, 2,
				     "Called to create a draft request.\n");
			req = pkinit_create_draft_pa_pk_as_req(context,
							       context->pool,
							       kcontext,
							       request,
							       context->certdb,
							       context->bag,
							       cert,
							       preferred_slot,
							       context->preferred_group,
							       &context->client_request.dh_private_key,
							       &context->client_request.dh_public_key,
							       &context->client_request.dh_nonce,
							       pkinit_pwcb,
							       &pwcb_args);
			/* If that succeeded, wrap the result up in a pa_data
			 * structure. */
			if (req != NULL) {
				pa_req = malloc(sizeof(krb5_pa_data));
				/* Aaagh! WHY is this not _REQ? */
				pa_req->pa_type = KRB5_PADATA_PK_AS_REP_OLD;
				pa_req->length = req->length;
				pa_req->contents = malloc(req->length);
				memcpy(pa_req->contents, req->data,
				       req->length);
				*out_pa_data = pa_req;
				/* Note success for our caller. */
				ret = 0;
			}
		}
		break;
	case KRB5_PADATA_PK_AS_REQ:
		/* Create a new pa_pk_as_req structure, sign it
		 * with a matching certificate, and prepare it for
		 * transmission. */
		pkinit_debug(context, 2,
			     "Called to create a RFC4556 request.\n");
		req = pkinit_create_pa_pk_as_req(context,
						 context->pool,
						 kcontext,
						 request,
						 encoded_request_body,
						 context->certdb,
						 context->bag,
						 cert,
						 SEC_OID_SHA1,
						 SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION,
						 preferred_slot,
						 context->preferred_group,
						 context->server_constraints.allowed_dh_params,
						 context->minimum_dh_prime_size,
						 context->try_dh ? &context->client_request.dh_private_key : NULL,
						 context->try_dh ? &context->client_request.dh_public_key : NULL,
						 context->try_dh ? &context->client_request.dh_nonce : NULL,
						 pkinit_pwcb,
						 &pwcb_args);
		/* If that succeeded, wrap the result up in a pa_data
		 * structure. */
		if (req != NULL) {
			pa_req = malloc(sizeof(krb5_pa_data));
			pa_req->pa_type = KRB5_PADATA_PK_AS_REQ;
			pa_req->length = req->length;
			pa_req->contents = malloc(req->length);
			memcpy(pa_req->contents, req->data, req->length);
			*out_pa_data = pa_req;
			/* Note success for our caller. */
			ret = 0;
		}
		break;
	default:
		break;
	}

	CERT_DestroyCertificate(cert);
	if (preferred_slot != NULL) {
		PK11_FreeSlot(preferred_slot);
	}

	if (ret == 0) {
		pkinit_debug(context, 2, "Returning success to libkrb5.\n");
	} else {
		pkinit_debug(context, 1, "Returning failure (%s) to libkrb5.\n",
			     error_message(ret));
	}
	return ret;
}

static int
client_get_flags(krb5_context unused_kcontext, krb5_preauthtype unused_patype)
{
	return PA_REAL;
}

static krb5_error_code
server_init(krb5_context kcontext, void **plugin_context)
{
	struct module_context *context;
	krb5_error_code ret;

	*plugin_context = NULL;
	ret = pkinit_init(kcontext, PR_TRUE, &context);
	if (ret == 0) {
		*plugin_context = context;
	}

	return ret;
}

static void
server_fini(krb5_context kcontext, void *plugin_context)
{
	struct module_context *context;

	context = plugin_context;
	if (context != NULL) {
		pkinit_fini(kcontext, context);
	}
}

/* Free state. */
static krb5_error_code
server_free_pa_context(krb5_context kcontext,
		       void *unused_plugin_context, void **pa_context)
{
	struct server_context *context;

	if (pa_context != NULL) {
		context = *pa_context;
		if (context != NULL) {
			if (context->client_nonce != NULL) {
				SECITEM_FreeItem(context->client_nonce,
						 PR_TRUE);
				context->client_nonce = NULL;
			}
			if (context->client_dh_nonce != NULL) {
				SECITEM_FreeItem(context->client_dh_nonce,
						 PR_TRUE);
				context->client_dh_nonce = NULL;
			}
			if (context->client_public_info != NULL) {
				SECITEM_FreeItem(context->client_public_info,
						 PR_TRUE);
				context->client_public_info = NULL;
			}
			if (context->client_cert != NULL) {
				SECITEM_FreeItem(context->client_cert, PR_TRUE);
				context->client_cert = NULL;
			}
			if (context->client_request != NULL) {
				krb5_free_data(kcontext,
					       context->client_request);
				context->client_request = NULL;
			}
			free(context);
		}
		*pa_context = NULL;
	}

	return 0;
}

static void
load_certs(struct module_context *context)
{
	/* Load up any certificates and keys which we may need.  We only
	 * need to do this once, though. */
	if (context->key_file != NULL) {
		if (aa_bag_load_key_file(context->bag,
					 context->key_file,
					 NULL) == SECSuccess) {
			pkinit_debug(context, 2, "Loaded "
				     "private key file \"%s\".\n",
				     context->key_file);
		} else {
			pkinit_debug(context, 0, "Error loading "
				     "private key file \"%s\".\n",
				     context->key_file);
		}
		context->key_file = NULL;
	}
	if (context->cert_file != NULL) {
		if (aa_bag_load_cert_file(context->bag,
					  context->cert_file,
					  NULL) == SECSuccess) {
			pkinit_debug(context, 2, "Loaded "
				     "certificate file \"%s\".\n",
				     context->cert_file);
		} else {
			pkinit_debug(context, 0, "Error loading "
				     "certificate file \"%s\".\n",
				     context->cert_file);
		}
		context->cert_file = NULL;
	}
	if (context->cert_pool != NULL) {
		if (aa_bag_add_cert_search_dir(context->bag,
					       context->cert_pool) == SECSuccess) {
			pkinit_debug(context, 2, "Set "
				     "certificate pool \"%s\".\n",
				     context->cert_pool);
		} else {
			pkinit_debug(context, 0, "Error setting "
				     "certificate pool \"%s\".\n",
				     context->cert_pool);
		}
		context->cert_pool = NULL;
	}
	if (context->cacert_file != NULL) {
		if (aa_bag_load_cacert_file(context->bag,
					    context->cacert_file,
					    NULL) == SECSuccess) {
			pkinit_debug(context, 2, "Loaded "
				     "CA certificate file \"%s\".\n",
				     context->cacert_file);
		} else {
			pkinit_debug(context, 0, "Error loading "
				     "CA certificate file \"%s\".\n",
				     context->cacert_file);
		}
		context->cacert_file = NULL;
	}
	if (context->cacert_pool != NULL) {
		if (aa_bag_add_cacert_search_dir(context->bag,
						 context->cacert_pool) == SECSuccess) {
			pkinit_debug(context, 2, "Set "
				     "certificate pool \"%s\".\n",
				     context->cacert_pool);
		} else {
			pkinit_debug(context, 0, "Error setting "
				     "certificate pool \"%s\".\n",
				     context->cacert_pool);
		}
		context->cacert_pool = NULL;
	}
}

/* Obtain and return any preauthentication data (which is destined for the
 * client) which matches type data->pa_type. */
static krb5_error_code
server_get_edata(krb5_context kcontext,
		 krb5_kdc_req *request,
		 struct _krb5_db_entry_new *unused_client,
		 struct _krb5_db_entry_new *unused_server,
	         preauth_get_entry_data_proc unused_server_get_entry_data,
		 void *plugin_context,
		 krb5_pa_data *data)
{
	struct module_context *context;
	CERTCertificate *cert;
	PLArenaPool *pool;
	krb5_principal tgs_principal;
	const char **additional_kdc_dn_list;
	char *unparsed_name;

	context = plugin_context;

	/* No data, but double-check that we do in fact support this type of
	 * preauthentication. */
	if (data->pa_type != KRB5_PADATA_PK_AS_REQ) {
		/* Any non-zero value is treated as an error, so this works. */
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		pkinit_debug(context, 0, "Out of memory.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Build the KDC's principal name. */
	tgs_principal = NULL;
	if (pkinit_build_tgs_name(context, kcontext, request->server,
				  &tgs_principal) != 0) {
		pkinit_debug(context, 0,
			     "Error building TGS principal name for %.*s.\n",
			     request->server->realm.length,
			     request->server->realm.data);
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}

	/* Get the list of mapped subject DNs for the realm's TGS. */
	additional_kdc_dn_list = NULL;
	unparsed_name = PR_smprintf("%s/%.*s@%.*s",
				    KRB5_TGS_NAME,
				    request->server->realm.length,
				    request->server->realm.data,
				    request->server->realm.length,
				    request->server->realm.data);
	if (unparsed_name != NULL) {
		additional_kdc_dn_list = map_file_get_list(context->map,
							   unparsed_name);
		PR_smprintf_free(unparsed_name);
	}

	/* Load certificates and keys, if needed. */
	load_certs(context);

	/* Check if we have a KDC certificate for the server, and error out if
	 * one isn't found, to prevent the KDC from listing us as a supported
	 * preauth type when we can't really do anything. */
	pkinit_debug(context, 2, "Looking for the KDC's certificate.\n");
	cert = cert_find_preferred_cert(context, pool, NULL, context->bag, NULL,
					NULL, NULL, additional_kdc_dn_list,
					kcontext, tgs_principal,
					CERT_MATCHES_DN_LIST |
					(context->trust_pkinit_san ?
					 CERT_MATCHES_KRB_SAN : 0 ) |
					(context->trust_upn_san ?
					 CERT_MATCHES_UPN_SAN : 0) |
					CERT_MATCHES_TRUSTED_SERVERS);

	/* Clean up. */
	krb5_free_principal(kcontext, tgs_principal);
	if (cert != NULL) {
		CERT_DestroyCertificate(cert);
	}
	PORT_FreeArena(pool, PR_TRUE);

	/* We're not really concerned with the contents of the certificate at
	 * this point, merely that we found one which we can use. */
	if (cert == NULL) {
		pkinit_debug(context, 0,
			     "No suitable KDC certificate found.\n");
		return KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
	}
	pkinit_debug(context, 2, "Found a suitable certificate.\n");

	return 0;
}

/* Add a member to a list of pointers.  Note that unlike other instances of
 * similar functions in this tree, we do not scan for duplicates. */
static void **
make_pointer_list(void **current_list, void *new_member)
{
	int i;
	void **ret;
	if (current_list == NULL) {
		ret = malloc(2 * sizeof(void*));
		if (ret != NULL) {
			ret[0] = new_member;
			ret[1] = NULL;
		}
		return ret;
	}
	for (i = 0; current_list[i] != NULL; i++) {
		continue;
	}
	ret = malloc((i + 2) * sizeof(void*));
	if (ret != NULL) {
		for (i = 0; current_list[i] != NULL; i++) {
			ret[i] = current_list[i];
		}
		ret[i++] = new_member;
		ret[i++] = NULL;
		free(current_list);
	}
	return ret;
}

/* Verify a request from a client. */
static krb5_error_code
server_verify(krb5_context kcontext,
	      struct _krb5_db_entry_new *client_db_entry,
	      krb5_data *req_pkt,
	      krb5_kdc_req *request,
	      krb5_enc_tkt_part *enc_tkt_reply,
	      krb5_pa_data *data,
	      preauth_get_entry_data_proc server_get_entry_data,
	      void *pa_plugin_context,
	      void **pa_request_context,
	      krb5_data **e_data)
{
	PLArenaPool *pool;
	krb5_error_code ret;
	krb5_data *client_entry_cert, *client_max_skew, preauth_data;
	krb5_authdata *authdata;
	krb5_deltat max_skew;
	SECItem client_cert_item, *mapped_certs, *item;
	int n_mapped_certs;
	CERTCertificate *signer;
	struct subject_public_key_info *client_public_info;
	SECItem *client_nonce, *client_dh_nonce;
	struct module_context *context;
	struct server_context *client;
	const char **additional_client_dn_list;
	char *unparsed_name;

	context = pa_plugin_context;
	pkinit_debug(context, 2, "Called to handle a client request.\n");

	/* Load certificates and keys, if needed. */
	load_certs(context);

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		return ENOMEM;
	}

	/* Allocate the server per-client context, but do it using the heap. */
	client = malloc(sizeof(*client));
	if (client == NULL) {
		PORT_FreeArena(pool, PR_TRUE);
		return ENOMEM;
	}
	*pa_request_context = client;
	memset(client, 0, sizeof(*client));

	/* Determine the maximum clock skew. */
	client_max_skew = NULL;
	ret = (*server_get_entry_data)(kcontext, request, client_db_entry,
				       krb5plugin_preauth_entry_max_time_skew,
				       &client_max_skew);
	if ((ret != 0) || (client_max_skew->length != sizeof(max_skew))) {
		pkinit_debug(context, 0,
			     "Unable to determine max permitted clock skew.\n");
		PORT_FreeArena(pool, PR_TRUE);
		return ret;
	}
	max_skew = *(krb5_deltat *)(client_max_skew->data);

	/* Check if the client has an authorized certificate in the database. */
	client_entry_cert = NULL;
	memset(&client_cert_item, 0, sizeof(client_cert_item));
	ret = (*server_get_entry_data)(kcontext, request, client_db_entry,
				       krb5plugin_preauth_entry_max_time_skew,
				       &client_entry_cert);
	if (client_entry_cert != NULL) {
		client_cert_item.data =
			(unsigned char *) client_entry_cert->data;
		client_cert_item.len = client_entry_cert->length;
		mapped_certs = &client_cert_item;
		n_mapped_certs = 1;
	} else {
		mapped_certs = NULL;
		n_mapped_certs = 0;
	}

	/* Verify the request. */
	preauth_data.data = (char *) data->contents;
	preauth_data.length = data->length;
	signer = NULL;
	client_public_info = NULL;
	client_nonce = NULL;
	client_dh_nonce = NULL;
	authdata = NULL;
	additional_client_dn_list = NULL;
	if (krb5_unparse_name(kcontext, request->client, &unparsed_name) == 0) {
		additional_client_dn_list = map_file_get_list(context->map,
							      unparsed_name);
		krb5_free_unparsed_name(kcontext, unparsed_name);
	}
	ret = pkinit_verify_pa_pk_as_req(context, pool,
					 context->certdb, context->bag,
					 kcontext, request, req_pkt,
					 &preauth_data, max_skew,
					 context->minimum_dh_prime_size,
					 &mapped_certs,
					 n_mapped_certs,
					 additional_client_dn_list,
					 CERT_MATCHES_DN_LIST |
					 (context->trust_pkinit_san ?
					  CERT_MATCHES_KRB_SAN : 0) |
					 (context->trust_upn_san ?
					  CERT_MATCHES_UPN_SAN : 0) |
					 CERT_MATCHES_TRUSTED_SERVERS,
					 &signer,
					 &client_public_info,
					 &client_nonce,
					 &client_dh_nonce,
					 &authdata,
					 e_data);
	if (ret == 0) {
		/* Success!  Note that we count as preauthentication. */
		enc_tkt_reply->flags |= TKT_FLG_PRE_AUTH;

		/* We may consider ourselves hardware preauthentication, too. */
		if (context->is_hw) {
			enc_tkt_reply->flags |= TKT_FLG_HW_AUTH;
		}

		/* Cache a copy of the client's certificate. */
		client->client_cert = &signer->derCert;

		/* Add supplied authdata to the list. */
		if (authdata != NULL) {
			enc_tkt_reply->authorization_data = (krb5_authdata**)make_pointer_list((void**) enc_tkt_reply->authorization_data, authdata);
		}

		/* The certificate is "owned" by the pool, but we need
		 * a copy which will survive the pool's destruction. */
		if (client->client_cert != NULL) {
			client->client_cert =
				SECITEM_DupItem(client->client_cert);
			krb5_copy_data(kcontext, req_pkt,
				       &client->client_request);
		}

		/* Save the client's DH info. */
		if (client_public_info != NULL) {
			item = common_encode_subject_public_key_info(context,
								     pool,
								     client_public_info);
			client->client_public_info = SECITEM_DupItem(item);
			pkinit_debug(context, 2,
				     "Saved client DH key information.\n");
		}
		if (client_dh_nonce != NULL) {
			client->client_dh_nonce =
				SECITEM_DupItem(client_dh_nonce);
			pkinit_debug(context, 2, "Saved client DH nonce.\n");
		}

		/* Save the client's nonce. */
		if (client_nonce != NULL) {
			client->client_nonce = SECITEM_DupItem(client_nonce);
			pkinit_debug(context, 2, "Saved client nonce.\n");
		}
	}
	if (signer != NULL) {
		CERT_DestroyCertificate(signer);
	}

	/* Clean up. */
	PORT_FreeArena(pool, PR_TRUE);

	return ret;
}

/* Create the response for a client. */
static krb5_error_code
server_return(krb5_context kcontext,
	      krb5_pa_data *unused_padata,
	      struct _krb5_db_entry_new *unused_client_db_entry,
	      krb5_data *unused_req_pkt,
	      krb5_kdc_req *request,
	      krb5_kdc_rep *unused_reply,
	      struct _krb5_key_data *unused_client_key,
	      krb5_keyblock *encrypting_key,
	      krb5_pa_data **send_pa,
	      preauth_get_entry_data_proc unused_server_get_entry_data,
	      void *pa_plugin_context,
	      void **pa_request_context)
{
	PLArenaPool *pool;
	CERTCertificate *kdc_cert;
	krb5_data *rep;
	krb5_pa_data *pa_data;
	krb5_error_code ret, usage_ret;
	krb5_principal tgt_princ;
	const char **additional_kdc_dn_list;
	char *unparsed_name;
	struct subject_public_key_info *client_public_info;
	struct module_context *context;
	struct server_context *client;

	ret = ENOMEM;

	/* We can't return data if we didn't process a PKINIT preauth request.
	 */
	if ((pa_request_context == NULL) || (*pa_request_context == NULL)) {
		return 0;
	}
	client = *pa_request_context;
	context = pa_plugin_context;

	/* If we can't create a new memory arena, bail. */
	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		return ENOMEM;
	}

	/* Get the list of mapped subject DNs for the server realm's TGS. */
	additional_kdc_dn_list = NULL;
	unparsed_name = PR_smprintf("%s/%.*s@%.*s",
				    KRB5_TGS_NAME,
				    request->server->realm.length,
				    request->server->realm.data,
				    request->server->realm.length,
				    request->server->realm.data);
	if (unparsed_name != NULL) {
		additional_kdc_dn_list = map_file_get_list(context->map,
							   unparsed_name);
		PR_smprintf_free(unparsed_name);
	}

	/* Find the KDC's certificate. */
	tgt_princ = NULL;
	if (pkinit_build_tgs_name(context, kcontext,
				  request->server, &tgt_princ) != 0) {
		pkinit_debug(context, 0,
			     "Error building KDC principal name for %.*s.\n",
			     request->server->realm.length,
			     request->server->realm.data);
		return ENOMEM;
	}

	pkinit_debug(context, 2, "Looking for the KDC's certificate.\n");
	kdc_cert = cert_find_preferred_cert(context,
					    pool,
					    NULL,
					    context->bag,
					    NULL,
					    NULL,
					    NULL,
					    additional_kdc_dn_list,
					    kcontext,
					    tgt_princ,
					    CERT_MATCHES_DN_LIST |
					    (context->trust_pkinit_san ?
					     CERT_MATCHES_KRB_SAN : 0 ) |
					    (context->trust_upn_san ?
					     CERT_MATCHES_UPN_SAN : 0));
	krb5_free_principal(kcontext, tgt_princ);
	if (kdc_cert == NULL) {
		pkinit_debug(context, 0, "Unable to locate KDC certificate.\n");
		return KRB5KDC_ERR_KDC_NAME_MISMATCH;
	}

	/* Build a new reply structure. */
	if ((client->client_cert != NULL) &&
	    (client->client_request != NULL)) {
		client_public_info = NULL;
		if (client->client_public_info != NULL) {
			client_public_info = common_decode_subject_public_key_info(context, pool,
										   client->client_public_info);
		}
		usage_ret = cert_verify_cert_for_encryption(context,
							    context->certdb,
							    context->bag,
							    client->client_cert);
		rep = NULL;
		if ((client_public_info != NULL) ||
		    (usage_ret == 0)) {
			rep = pkinit_create_pa_pk_as_rep(context,
							 pool,
							 context->certdb,
							 context->bag,
							 kdc_cert,
							 kcontext,
							 encrypting_key,
							 client->client_request,
							 client->client_cert,
							 SEC_OID_SHA1,
							 SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION,
							 client_public_info,
							 client->client_nonce,
							 client->client_dh_nonce);
		} else {
			ret = usage_ret;
		}
		if (rep != NULL) {
			pa_data = malloc(sizeof(*pa_data));
			if (pa_data == NULL) {
				pkinit_debug(context, 0, "Out of memory.\n");
				ret = ENOMEM;
			} else {
				pa_data->pa_type = KRB5_PADATA_PK_AS_REP;
				pa_data->length = rep->length;
				pa_data->contents = malloc(rep->length);
				if (pa_data->contents == NULL) {
					pkinit_debug(context, 0,
						     "Out of memory.\n");
					free(pa_data);
					ret = ENOMEM;
				} else {
					memcpy(pa_data->contents,
					       rep->data, rep->length);
					*send_pa = pa_data;
					pkinit_debug(context, 2,
						     "Generated response.\n");
					ret = 0;
				}
			}
		}
	}

	/* Clean up. */
	CERT_DestroyCertificate(kdc_cert);
	PORT_FreeArena(pool, PR_TRUE);

	/* Clear the context. */
	server_free_pa_context(kcontext, pa_plugin_context, pa_request_context);

	pkinit_debug(context, ret ? 1 : 2, "Returning %d (%s).\n",
		     ret, error_message(ret));
	return ret;
}

static int
server_get_flags(krb5_context kcontext, krb5_preauthtype unused_patype)
{
	int is_hw;
	krb5_appdefault_boolean(kcontext, "pkinit", NULL,
				"is_hw", PR_FALSE, &is_hw);
	return PA_SUFFICIENT | PA_REPLACES_KEY | (is_hw ? PA_HARDWARE : 0);
}

struct krb5plugin_preauth_client_ftable_v0 preauthentication_client_0 = {
	.name = "pkinit-nss",
	.pa_type_list = &supported_client_pa_types[0],
	.enctype_list = &supported_pk_enc_types[0],
	.init = client_init,
	.fini = client_fini,
	.process = client_process,
	.tryagain = client_try_again,
#ifndef PKINIT_CLIENT_MISSING_GIC_OPTS
	.gic_opts = client_gic_opts,
#endif
	.flags = client_get_flags,
};
struct krb5plugin_preauth_server_ftable_v0 preauthentication_server_0 = {
	.name = "pkinit-nss",
	.pa_type_list = &supported_server_pa_types[0],
	.init_proc = server_init,
	.fini_proc = server_fini,
	.flags_proc = server_get_flags,
	.edata_proc = server_get_edata,
	.verify_proc = server_verify,
	.return_proc = server_return,
	.freepa_reqcontext_proc = server_free_pa_context,
};
